//
// Date.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Date.h"

#include "Time.h"
#include "String.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	//
	// Tables defining the number of days in a month.
	//
	const int kDaysInMonth[] = 
	{
	//  J   F   M   A   M   J   J   A   S   O   N   D
		31,	28,	31,	30,	31,	30,	31,	31,	30,	31,	30,	31
	};

	const int kLeapYearDaysInMonth[] = 
	{
	//  J   F   M   A   M   J   J   A   S   O   N   D
		31,	29,	31,	30,	31,	30,	31,	31,	30,	31,	30,	31
	};


	/**************************************************************************
	
	>	Date::Date()
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	Date::Date() : m_Day(1),
				   m_Month(January),
				   m_Year(-1)
	{
		// Default to illegal date - client must set date values
	}

	/**************************************************************************
	
	>	Date::Date(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - Date to copy into this object
	
		Purpose:	Copy constructor
	
	**************************************************************************/
	
	Date::Date(const Date& rhs)
	{
		// Check for self-copy
		if (this == &rhs)
			return;

		m_Day = rhs.m_Day;
		m_Month = rhs.m_Month;
		m_Year = rhs.m_Year;
	}

	/**************************************************************************
	
	>	bool Date::IsValid(int Year, int Month, int Day)
	
		Author:		Tim Browse
		Date:		11/9/2002
		Inputs:		Year, Month, Day - the date to test
		Returns:	true => date is valid
					false => date is not valid (Day is invalid given Year
							 and Month values)
	
		Purpose:	Test to see if a given date is valid - i.e. the Day
					is legal, given the month and year (e.g. year may be
					a leap year)
	
	**************************************************************************/
	
	bool Date::IsValid(int Year, int Month, int Day)
	{
		Date TestDate;
		TestDate.m_Year = Year;
		TestDate.m_Month = (MonthName) (Month - 1);

		// Validate days
		return ((Day >= 1) && (Day <= TestDate.GetDaysInMonth()));
	}

	/**************************************************************************
	
	>	Date& Date::operator=(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - the Date object to assign to this object
	
		Purpose:	Assignment operator
	
	**************************************************************************/
	
	Date& Date::operator=(const Date& rhs)
	{
		// Check for self-copy
		if (this == &rhs)
			return *this;

		m_Day = rhs.m_Day;
		m_Month = rhs.m_Month;
		m_Year = rhs.m_Year;
		return *this;
	}

	/**************************************************************************

	>	bool Date::Set(int Day,
					   MonthName Month,
					   int Year)

		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		Day - the day of the month to use (first day = 1)
					Month - the name of the month to use
					Year - the year to use (in full, not just the last 2 digits)

		Purpose:	Set a date object to a particular date.

	**************************************************************************/

	bool Date::Set(int Day,
				   MonthName Month,
				   int Year)
	{
		m_Year = Year;
		m_Month = Month;

		// Validate days
		if ((Day < 1) || (Day > GetDaysInMonth()))
		{
			// Illegal date
			vosERROR();
			return false;
		}

		m_Day = Day;

		return true;
	}

	/**************************************************************************

	>	bool Date::Set(int Day,
					   int Month,
					   int Year)


		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		Day - the day of the month to use (first day = 1)
					Month - the index of the month to use (January = 1)
					Year - the year to use (in full, not just the last 2 digits)

		Purpose:	Set a date object to a particular date.

	**************************************************************************/

	bool Date::Set(int Day,
				   int Month,
				   int Year)
	{
		m_Year = Year;

		// Validate months and days
		if ((Month < 1) || (Month > 12) ||
			(Day < 1) || (Day > GetDaysInMonth()))
		{
			// Illegal date
			vosERROR();
			return false;
		}

		m_Month = (MonthName) (Month - 1);
		m_Day = Day;

		return true;
	}

	/**************************************************************************
	
	>	bool Date::IsLeapYear()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	true => date falls within a leap year
					false => otherwise
	
		Purpose:	Test to see if this date falls within a leap year.
	
	**************************************************************************/
	
	bool Date::IsLeapYear()
	{
		// Leap year - rules are as follows:
		//
		// 1. If the year is divisible by 4    => leap year.
		//
		// 2. If the year is divisible by 100  => not a leap year.
		//
		// 3. If the year is divisible by 400  => leap year.
		//
		// These rules must be applied in this order.
		//
		bool bIsLeapYear = false;

		if ((m_Year % 4) == 0)
		{
			bIsLeapYear = true;
		}

		if ((m_Year % 100) == 0)
		{
			bIsLeapYear = false;
		}

		if ((m_Year % 400) == 0)
		{
			bIsLeapYear = true;
		}

		return bIsLeapYear;
	}

	/**************************************************************************
	
	>	int Date::GetDaysInMonth()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	Number of days in month, i.e. 28, 29, 30 or 31
	
		Purpose:	Find out how many days are in the month that this date
					falls in.  Leap years are handled correctly.
	
	**************************************************************************/
	
	int Date::GetDaysInMonth()
	{
		if (IsLeapYear())
			return kLeapYearDaysInMonth[m_Month];
		else
			return kDaysInMonth[m_Month];
	}

	/**************************************************************************
	
	>	void Date::AddDays(int nDays)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		nDays - the number of days to add to this date
	
		Purpose:	Add a number of days to this date.
	
	**************************************************************************/
	
	void Date::AddDays(int nDays)
	{
		m_Day += nDays;

		// Keep looping around, moving to the correct month
		while (m_Day < 1)
		{
			// Move to previous month
			if (m_Month == January)
			{
				// Move to previous year
				m_Month = December;
				m_Year--;
			}
			else
			{
				m_Month = (MonthName) ((int) m_Month - 1);
			}

			// Adjust days into previous month.
			m_Day += GetDaysInMonth();
		}

		while (m_Day > GetDaysInMonth())
		{
			// Move to next month
			if (m_Month == December)
			{
				// Move to next year
				m_Month = January;
				m_Year++;
			}
			else
			{
				m_Month = (MonthName) ((int) m_Month + 1);
			}

			// Adjust days into next month.
			m_Day -= GetDaysInMonth();
		}
	}

	/**************************************************************************
	
	>	Date& Date::operator+=(int nDays)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		nDays - the number of days to add to this date
	
		Purpose:	Add a number of days to this date.
	
	**************************************************************************/
	
	Date& Date::operator+=(int nDays)
	{
		AddDays(nDays);
		return *this;
	}

	/**************************************************************************
	
	>	Date& Date::operator-=(int nDays)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		nDays - the number of days to subtract from this date
	
		Purpose:	Subtract a number of days from this date.
	
	**************************************************************************/
	
	Date& Date::operator-=(int nDays)
	{
		AddDays(-nDays);
		return *this;
	}

	/**************************************************************************
	
	>	bool Date::operator==(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - Date to compare this object against
	
		Purpose:	Equality operator
	
	**************************************************************************/
	
	bool Date::operator==(const Date& rhs)
	{
		return (m_Year == rhs.m_Year) &&
			   (m_Month == rhs.m_Month) &&
			   (m_Day == rhs.m_Day);
	}

	/**************************************************************************
	
	>	bool Date::operator!=(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - Date to compare this object against
	
		Purpose:	Inequality operator
	
	**************************************************************************/
	
	bool Date::operator!=(const Date& rhs)
	{
		return (m_Year != rhs.m_Year) ||
			   (m_Month != rhs.m_Month) ||
			   (m_Day != rhs.m_Day);
	}

	/**************************************************************************
	
	>	bool Date::operator>(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - Date to compare this object against
	
		Purpose:	> inequality operator
	
	**************************************************************************/
	
	bool Date::operator>(const Date& rhs)
	{
		if (m_Year > rhs.m_Year)
			return true;

		if (m_Year == rhs.m_Year)
		{
			if (m_Month > rhs.m_Month)
				return true;

			if ((m_Month == rhs.m_Month) && (m_Day > rhs.m_Day))
				return true;
		}

		// Otherwise, must be less than or equal to
		return false;
	}

	/**************************************************************************
	
	>	bool Date::operator>=(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - Date to compare this object against
	
		Purpose:	>= inequality operator
	
	**************************************************************************/
	
	bool Date::operator>=(const Date& rhs)
	{
		if (m_Year > rhs.m_Year)
			return true;

		if (m_Year == rhs.m_Year)
		{
			if (m_Month > rhs.m_Month)
				return true;

			if ((m_Month == rhs.m_Month) && (m_Day >= rhs.m_Day))
				return true;
		}

		// Otherwise, must be less than
		return false;
	}

	/**************************************************************************
	
	>	bool Date::operator<=(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - Date to compare this object against
	
		Purpose:	<= inequality operator
	
	**************************************************************************/
	
	bool Date::operator<=(const Date& rhs)
	{
		if (m_Year < rhs.m_Year)
			return true;

		if (m_Year == rhs.m_Year)
		{
			if (m_Month < rhs.m_Month)
				return true;

			if ((m_Month == rhs.m_Month) && (m_Day <= rhs.m_Day))
				return true;
		}

		// Otherwise, must be greater than
		return false;
	}

	/**************************************************************************
	
	>	bool Date::operator<(const Date& rhs)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		rhs - Date to compare this object against
	
		Purpose:	< inequality operator
	
	**************************************************************************/
	
	bool Date::operator<(const Date& rhs)
	{
		if (m_Year < rhs.m_Year)
			return true;

		if (m_Year == rhs.m_Year)
		{
			if (m_Month < rhs.m_Month)
				return true;

			if ((m_Month == rhs.m_Month) && (m_Day < rhs.m_Day))
				return true;
		}

		// Otherwise, must be greater than or equal to
		return false;
	}

	/**************************************************************************
	
	>	Date Date::GetToday()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	Today's date
	
		Purpose:	Get today's date.
	
	**************************************************************************/
	
	Date Date::GetToday()
	{
		// Get current time/date
		Time Now = Time::GetCurrentTime();

		// Decode into fields
		TimeInfo Info(Now);

		// Construct date representing today and return to caller
		Date Today;
		Today.Set(Info.m_DayOfMonth, Info.m_Month, Info.m_Year);
		return Today;
	}

	/**************************************************************************
	
	>	String Date::MakeString()
	
		Author:		Tim Browse
		Date:		13/4/2002
		Returns:	String representation of Data, in DD/MM/YYYY format.
	
		Purpose:	Construct canonical string for this date - locale date
					orderings do not affect the layout.

					Hence this is a reliable way of saving dates in text format
					that will work across locales.

					Uses standard british DD/MM/YYYY format (arbitrary decision)
	
	**************************************************************************/
	
	String Date::MakeString()
	{
		String strDate = String(m_Day);
		strDate += "/";
		strDate += GetCardinalMonth();
		strDate += "/";
		strDate += String(m_Year);

		return strDate;
	}

};
