//
// DiskFile.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "DiskFile.h"

#include <cerrno>

#include "Errors.h"
#include "Msgs.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/**************************************************************************

	>	DiskFile::DiskFile()

		Author:		Tim Browse
		Date:		23/08/99

		Purpose:	Constructor

	**************************************************************************/

	DiskFile::DiskFile() : m_hFile(NULL)
	{
		// m_FileSpec
	}

	/**************************************************************************

	>	DiskFile::~DiskFile()

		Author:		Tim Browse
		Date:		23/08/99

		Purpose:	Destructor - closes file if it is open.

	**************************************************************************/

	DiskFile::~DiskFile()
	{
		// Make sure open files are closed.
		Close();
	}



	/**************************************************************************

	>	Error DiskFile::Create(const FileSpec& FS)

		Author:		Jason Williams
		Date:		19/1/2001

		Inputs:		FS - file spec of the file to access
		Returns:	Standard error

		Purpose:	Create a disk file. This associates the DiskFile object
					with a given FileSpec, which can subsequently be Open()ed.

	**************************************************************************/

	Error DiskFile::Create(const FileSpec& FS)
	{
		if (IsOpen())
		{
			COMPLAIN("File already open in DiskFile::Create().");
			return ERROR(Err_File_OpenFailed, msgDiskFile_FileAlreadyOpen);
		}

		// Ignore empty filenames
		if (FS.IsEmpty())
		{
			COMPLAIN("Empty FileSpec passed to DiskFile::Open().");
			return ERROR(Err_File_OpenFailed, msgDiskFile_OpenFailedNoFS);
		}

		// Remember the FileSpec so we can open the file later
		m_FileSpec = FS;

		return(VNOERROR);
	}


	
	/**************************************************************************

	>	Error DiskFile::Open(const FileSpec& FS, FileMode Mode)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		FS - file spec of the file to open
					Mode - the access mode to open the file with.
		Returns:	Standard error

		Purpose:	Create & Open a disk file.

	**************************************************************************/

	Error DiskFile::Open(const FileSpec& FS, FileMode Mode)
	{
		RETURN_ON_FAILURE(Create(FS));		// Create the file
		RETURN_ON_FAILURE(Open(Mode));		// And open it

		return(VNOERROR);
	}



	/**************************************************************************

	>	Error DiskFile::Open(FileMode Mode)

		Author:		Tim Browse
		Date:		23/08/99

		Inputs:		FS - file spec of the file to open
					Mode - the access mode to open the file with.
		Returns:	Standard error

		Purpose:	Open a disk file

		Notes:		You must call Create() first to associate this object with
					the file to be opened (or see alternative Open() methods)

					You may Open() and Close() the file as many times as required
					without having to call Create() again.

		SeeAlso:	DiskFile::Create(); DiskFile::Open();

	**************************************************************************/

	Error DiskFile::Open(FileMode Mode)
	{
		if (IsOpen())
		{
			COMPLAIN("File already open in DiskFile::Open().");
			return ERROR(Err_File_OpenFailed, msgDiskFile_FileAlreadyOpen);
		}

		// Ignore empty filenames
		if (m_FileSpec.IsEmpty())
		{
			COMPLAIN("Call DiskFile::Create() before DiskFile::Open().");
			return ERROR(Err_File_OpenFailed, msgDiskFile_OpenFailedNoFS);
		}

		// [Oscar added - 20/02/2001]
		// Construct the file access mode string...
		// this is being engineered to be suitable for fopen... see the 
		// relevant documentation if you're curious.

		// the cases we'll consider legal are
		// "a+" = append mode (text), "r" = read mode (text), "w" = write mode (text), 
		// "ab+" = append mode (binary), "rb" = read mode (binary), "wb" = write mode (binary)

		// bear in mind that write mode zaps the contents of the file before continuing.

		// also you can't choose more than one from read/write(new)/append which is good
		// but append has higher priority than read which is higher than write.
		// this should probably involve a warning somewhere.
	
		// I think this covers the desired range of usage for DiskFile...
		// this change shouldn't change existing behaviour apart from making append 
		// work predictably.

		char szMode[16] = "";

		// We use "a+" for append, because "a" does not remove the EOF before
		// appending, or restore it when the file is closed, which "a+" does.
		// If we did not use this, appending to text files would result in
		// text files with EOF markers in the middle of the file.

		if (Mode & fam_Append)
		{
			strcat(szMode, "a");
		}
		else
		if (Mode & fam_Read)
		{
			strcat(szMode, "r");
		}
		else
		if (Mode & fam_Write)
		{
			strcat(szMode, "w");
		}
		else
		{
			// complain about this using a 
			COMPLAIN("Wacky file access mode used in DiskFile::Open().");
			return ERROR(Err_File_OpenFailed, msgDiskFile_OpenFailed);
		}

		if (!(Mode & fam_Text))
		{
			// if text is not specified then we put the binary chap in.
			strcat(szMode, "b");
		}

		if (Mode & fam_Append)
		{
			// the + symbol supposedly refers to "opening the file for an update".
			// we only want to do that the append case.
			strcat(szMode, "+");
		}

		// Open the file
		m_hFile = fopen(m_FileSpec.m_FileName.c_str(), szMode);

		if (m_hFile != NULL)
		{
			// File opened successfully
			return VNOERROR;
		}

		// Unable to open the file - make informative error message.
		return RegisterError(Err_File_OpenFailed, 
							 msgDiskFile_OpenFailed, 
							 m_FileSpec.GetString().c_str(),
							 strerror(errno));
	}



	/**************************************************************************

	>	void DiskFile::Close()

		Author:		Tim Browse
		Date:		23/08/99

		Purpose:	Close a disk-based file.

	**************************************************************************/

	void DiskFile::Close()
	{
		if (m_hFile != NULL)
		{
			fclose(m_hFile);
			m_hFile = NULL;
		}
	}

	/**************************************************************************

	>	virtual bool DiskFile::IsOpen() const

		Author:		Tim Browse
		Date:		23/08/99
		Returns:	true => file is open
					false => file is not open

		Purpose:	Test to see if the file is currently open.

	**************************************************************************/

	bool DiskFile::IsOpen() const
	{
		return (m_hFile != NULL);
	}


	/**************************************************************************

	>	virtual Error DiskFile::GetSize(uint* pSize)

		Author:		Jason
		Date:		15/1/2001

		Outputs:	pSize	- The size, in bytes, of this file

		Purpose:	Determine the size (in bytes) of this file

	**************************************************************************/
/*
	Error DiskFile::GetSize(uint* pSize)
	{
		COMPLAIN_IF(pSize == NULL, "Illegal NULL param");

		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, msgDiskFile_NotOpen);
		}

		if (pSize != NULL)
		{
			// TODO - determine the file size. For now, the base class does a
			// reasonable (and portable) job by seeking to the end of the file
			// and reading the file position, but this really ought to override
			// it to do something sensible. Of course this file class has no idea
			// what file it actually has open, so it might not be able to do a
			// heck of a lot on some platforms...
		}

		return(VNOERROR);
	}

*/


	/**************************************************************************

	>	Error DiskFile::Read(void* pData, uint cBytes, uint* pcBytes)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		cBytes - the number of bytes to read
		Outputs:	pData - the data read from the file.
					pcBytes - the number of bytes actually read from the file.
							  (pcBytes may be NULL fi not needed)
		Returns:	Standard error

		Purpose:	Read the specified number of bytes from the file into the 
					specified area of memory.

	**************************************************************************/

	Error DiskFile::Read(void* pData, uint cBytes, uint* pcBytes)
	{
		if (m_hFile == NULL)
		{
			return ERROR(Err_File_NotOpen, msgDiskFile_Read_NotOpen);
		}
		else
		{
			// Read the data
			uint cBytesRead = (uint) fread(pData, 1, cBytes, m_hFile);

			// Tell caller how many bytes we read, if they care.
			if (pcBytes != NULL)
			{
				*pcBytes = cBytesRead;
			}

			// Did the read work?
			if (cBytesRead == cBytes)
			{
				// Yes.
				return VNOERROR;
			}

			// Read failed - was it eof, or something else?
			if (feof(m_hFile))
			{
				// EOF occurred
				return ERROR(Err_File_EOF, msgDiskFile_Read_EOF);
			}
			else
			{
				// Generic read failure
				return RegisterError(Err_File_ReadFail, 
									 msgDiskFile_Read_Fail, 
									 strerror(errno));
			}
		}
	}

	/**************************************************************************

	>	Error DiskFile::Write(const void* pData, uint cBytes, uint* pcBytes)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		pData - the data to write to the file.
					cBytes - the number of bytes to write to the file.
		Outputs:	pcBytes - the number of bytes actually written to the file.
							  (may be NULL if not needed)
		Returns:	Standard error

		Purpose:	Write the specified number of bytes from the specified
					area of memory to this file.

	**************************************************************************/

	Error DiskFile::Write(const void* pData, uint cBytes, uint* pcBytes)
	{
		if (m_hFile == NULL)
		{
			return ERROR(Err_File_NotOpen, msgDiskFile_Write_NotOpen);
		}
		else
		{
			uint cBytesWritten = (uint) fwrite(pData, 1, cBytes, m_hFile);

			if (pcBytes != NULL)
			{
				*pcBytes = cBytesWritten;
			}

			// Did the write work?
			if (cBytesWritten == cBytes)
			{
				// Yes.
				return VNOERROR;
			}

			// Write failed - work out what sort of write error occurred
			switch (errno)
			{
#ifdef VOS_MSWINDOWS
				case ENOSPC:
					// Report that there is no room to write.
					return ERROR(Err_File_NoSpaceAvailable, 
								 msgDiskFile_Write_NoSpaceLeft);

				case EACCES:
					// Permission denied for write.
					return ERROR(Err_File_AccessDenied, 
								 msgDiskFile_Write_AccessDenied);
#endif
				default:
					// Generic write failure.
					return RegisterError(Err_File_WriteFail, 
										 msgDiskFile_Write_Fail,
										 strerror(errno));
			}
		}
	}


	/**************************************************************************

	>	Error DiskFile::SeekFromStart(uint Offset)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		Offset - the number of bytes from the start of the file
							 to position the file read/write pointer at.
		Returns:	Standard error

		Purpose:	Seek to a position in the file specified as a number of 
					bytes from the start of the file.

	**************************************************************************/

	Error DiskFile::SeekFromStart(uint Offset)
	{
		if (m_hFile == NULL)
		{
			COMPLAIN("DiskFile::SeekFromStart() called when file is not open");

			return ERROR(Err_File_NotOpen, msgDiskFile_NotOpen);
		}

		if (fseek(m_hFile, Offset, SEEK_SET) == 0)
		{
			// Seek worked
			return VNOERROR;
		}

		// Seek failed
		return RegisterError(Err_File_SeekFail, 
							 msgDiskFile_Access_Fail, 
							 strerror(errno));
	}

	/**************************************************************************

	>	Error DiskFile::SeekFromEnd(uint Offset)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		Offset - the number of bytes from the end of the file
							 to position the file read/write pointer at.
		Returns:	Standard error

		Purpose:	Seek to a position in the file specified as a number of 
					bytes from the end of the file.

	**************************************************************************/

	Error DiskFile::SeekFromEnd(uint Offset)
	{
		if (m_hFile == NULL)
		{
			COMPLAIN("DiskFile::SeekFromEnd() called when file is not open");

			return ERROR(Err_File_NotOpen, msgDiskFile_NotOpen);
		}

		if (fseek(m_hFile, Offset, SEEK_END) == 0)
		{
			// Seek worked
			return VNOERROR;
		}

		// Seek failed
		return RegisterError(Err_File_SeekFail, 
							 msgDiskFile_Access_Fail, 
							 strerror(errno));
	}

	/**************************************************************************

	>	Error DiskFile::SeekRelative(int Offset)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		Offset - the offset, in bytes, to seek to, relative to the
							 current file read/write position.

							 Positive values seek forward, negative values
							 seek backwards.

		Returns:	Standard error

		Purpose:	Seek to a position in the file, relative to the current
					read/write position of the file.

	**************************************************************************/

	Error DiskFile::SeekRelative(int Offset)
	{
		if (m_hFile == NULL)
		{
			COMPLAIN("DiskFile::SeekRelative() called when file is not open");

			return ERROR(Err_File_NotOpen, msgDiskFile_NotOpen);
		}

		if (fseek(m_hFile, Offset, SEEK_CUR) == 0)
		{
			// Seek worked
			return VNOERROR;
		}

		// Seek failed
		return RegisterError(Err_File_SeekFail, 
							 msgDiskFile_Access_Fail, 
							 strerror(errno));
	}

	/**************************************************************************

	>	Error DiskFile::GetPos(uint* pPos) const

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pPos - the current read/write position of the file.
		Returns:	Standard error

		Purpose:	Get the current read/write position of the file.

	**************************************************************************/

	Error DiskFile::GetPos(uint* pPos) const
	{
		if (m_hFile == NULL)
		{
			COMPLAIN("DiskFile::GetPos() called when file is not open");

			return ERROR(Err_File_NotOpen, msgDiskFile_NotOpen);
		}

		long lResult = ftell(m_hFile);

		if (lResult < 0)
		{
			// Error
			return RegisterError(Err_File_GetPosFail, 
								 msgDiskFile_Access_Fail, 
								 strerror(errno));
		}

		// Function successful
		*pPos = (uint) lResult;
		return VNOERROR;
	}

	/**************************************************************************

	>	bool DiskFile::eof() const

		Author:		Tim Browse
		Date:		23/08/99
		Returns:	true => eof encountered
					false => eof not encountered

		Purpose:	Find out if the end-of-file has been reached for this
					file yet.

	**************************************************************************/

	bool DiskFile::eof() const
	{
		// Check for null file
		if (m_hFile == NULL)
		{
			COMPLAIN("DiskFile::eof() called when file is not open");
			return false;
		}

		return (feof(m_hFile) != 0);
	}


	/**************************************************************************

	>	Error DiskFile::ReadLine(String* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the line of text read from the file.
		Returns:	Standard error

		Purpose:	Read a line of text from the file, terminating at any
					type of line-end combination or EOF.  If a newline
					character is read, it is included in the string.

					There is no limit to the length of line that can be
					read from a text file with this function.

	**************************************************************************/

	Error DiskFile::ReadLine(String* pData)
	{
		if (m_hFile == NULL)
		{
			COMPLAIN("DiskFile::ReadLine() called when file is not open");

			return ERROR(Err_File_NotOpen, msgDiskFile_NotOpen);
		}

		// Before we start, are we already at EOF?
		if (feof(m_hFile))
		{
			// Error occurred.
			return ERROR(Err_File_EOF, msgDiskFile_Read_EOF);
		}

		// Clear the string first
		pData->Clear();

		// Buffer to hold line fragments
		
		// Mike: Buffer size temporarily changed from 128 to 256 by me. See bug23 in testtrack
		const int kLineBufSize = 256;
		char szBuffer[kLineBufSize+1];
		szBuffer[0] = 0;

		for(;;)
		{	

#ifdef VOS_MACOS
			// Read up to an EOL or a newline
			// NOTE: This is at least as efficient as MWKS's and MSFT's ANSI libs...
			int 	CharCount = 	kLineBufSize; 
			char *	pCurrentChar =	szBuffer;
			char 	c;
			do
			{
				// Read the next char and append it
				c = ::getc(m_hFile);
				if (c == EOF)
				{
					// EOF or Error?
					if (!::feof(m_hFile))
					{
						// Error occurred.
						return RegisterError(Err_File_ReadFail, 
											 msgDiskFile_Read_Fail,
											 strerror(errno));
					}
					// Exit this inner reading loop
					break;
				}
				*pCurrentChar++ = c;
			}
			while ((c != 0x0A) && (c != 0x0D) && (--CharCount > 0));

			// Null terminate the buffer to make a C string so strlen works
			*pCurrentChar = 0;

#else
			// Read as much as we can into the buffer.
			char *Result = fgets(szBuffer, kLineBufSize, m_hFile);

			if (Result == NULL)
			{
				// EOF or Error?
				if (!feof(m_hFile))
				{
					// Error occurred.
					return RegisterError(Err_File_ReadFail, 
										 msgDiskFile_Read_Fail,
										 strerror(errno));

				}
			}
#endif

			// Add newly read text to string
			*pData += szBuffer;

			// We keep going until we get a CR or LF character at the end of the string.
			int nLength = strlen(szBuffer);

			if (nLength < (kLineBufSize - 1))
			{
				// we're done
				break;
			}
			else
			{
				// Check for CR or LF at the end
				if ((szBuffer[nLength-1] == 0x0D) ||
					(szBuffer[nLength-1] == 0x0A))
				{
					// We're finished
					break;
				}
			}
		}

		// Remove CRLF characters from end of string
		int iLastChar = pData->GetLength() - 1;

		while ((iLastChar >= 0) && 
			   ((pData->GetAt(iLastChar) == 0x0D) ||
			    (pData->GetAt(iLastChar) == 0x0A)))
		{
			// Remove this character
			pData->SetAt(iLastChar, 0);
			iLastChar--;
		}

		return VNOERROR;
	}

	/**************************************************************************
	
	>	Error DiskFile::RegisterError(int Code, 
									  const char *pszFormatString, ...) const
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		Code - the error code to register
					pszFormatString - the error message to register
									  (May contain printf specifiers, which will
									   use the following parameters in the usual
									   va_list way).
		Returns:	Error (failure) - so that this can be returned to the
									  caller of the caller easily.
	
		Purpose:	Register the specified error with the vOS error manager.
	
	**************************************************************************/
	
	Error DiskFile::RegisterError(int Code, const char *pszFormatString, ...) const
	{
		String ErrMsg;

		// Pass the varargs list to printf
		va_list ArgList;
		va_start(ArgList, pszFormatString);

		ErrMsg.Printf_va(pszFormatString, ArgList);

		va_end(ArgList);

		// Report error to error manager
		return ERROR(Code, ErrMsg.c_str());
	}
};
