//
// Endian.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_Endian_h
#define vos_Endian_h

#include "vos.h"

namespace vos
{
	/**************************************************************************
	
	>	inline int16 EndianSwap_int16(int16 n)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		n - value to endian swap
		Returns:	Endian swapped value.
	
		Purpose:	Perform Endian swap on a signed 16 bit value.
	
	**************************************************************************/
	
	inline int16 EndianSwap_int16(int16 n)
	{
		return ((n & 0xFF) << 8) | 
				((n >> 8) & 0xFF);
	}

	/**************************************************************************
	
	>	inline int32 EndianSwap_int32(int32 n)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		n - value to endian swap
		Returns:	Endian swapped value.
	
		Purpose:	Perform Endian swap on a signed 32 bit value.
	
	**************************************************************************/
	
	inline int32 EndianSwap_int32(int32 n)
	{
		return (((n & 0xFF) << 24) | 
				((n & 0xFF00) << 8) |
				((n & 0xFF0000) >> 8) |
				 ((n >> 24) & 0xFF));
	}

	/**************************************************************************
	
	>	inline int64 EndianSwap_int64(int64 n)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		n - value to endian swap
		Returns:	Endian swapped value.
	
		Purpose:	Perform Endian swap on a signed 64 bit value.
	
	**************************************************************************/
	
	inline int64 EndianSwap_int64(int64 n)
	{
        return   (((((int64)n)<<56) & 0xFF00000000000000)  |
                 ((((int64)n)<<40) & 0xFF000000000000)  |
                 ((((int64)n)<<24) & 0xFF0000000000)  |
                 ((((int64)n)<< 8) & 0xFF00000000)  |
                 ((((int64)n)>> 8) & 0xFF000000)  |
                 ((((int64)n)>>24) & 0xFF0000)  |
                 ((((int64)n)>>40) & 0xFF00)  |
                 ((((int64)n)>>56) & 0xFF));
	}

	/**************************************************************************
	
	>	inline uint16 EndianSwap_uint16(uint16 n)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		n - value to endian swap
		Returns:	Endian swapped value.
	
		Purpose:	Perform Endian swap on an unsigned 16 bit value.
	
	**************************************************************************/
	
	inline uint16 EndianSwap_uint16(uint16 n)
	{
		return ((n & 0xFF) << 8) | (n >> 8);
	}

	/**************************************************************************
	
	>	inline uint32 EndianSwap_uint32(uint32 n)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		n - value to endian swap
		Returns:	Endian swapped value.
	
		Purpose:	Perform Endian swap on an unsigned 32 bit value.
	
	**************************************************************************/
	
	inline uint32 EndianSwap_uint32(uint32 n)
	{
		return (((n & 0xFF) << 24) | 
				((n & 0xFF00) << 8) |
				((n & 0xFF0000) >> 8) |
				 (n >> 24));
	}

	/**************************************************************************
	
	>	inline uint64 EndianSwap_uint64(uint64 n)
	
		Author:		Tim Browse
		Date:		13/4/2002
		Inputs:		n - value to endian swap
		Returns:	Endian swapped value.
	
		Purpose:	Perform Endian swap on an unsigned 64 bit value.
	
	**************************************************************************/
	
	inline uint64 EndianSwap_uint64(uint64 n)
	{
        return   (((((uint64)n)<<56) & 0xFF00000000000000)  |
                 ((((uint64)n)<<40) & 0xFF000000000000)  |
                 ((((uint64)n)<<24) & 0xFF0000000000)  |
                 ((((uint64)n)<< 8) & 0xFF00000000)  |
                 ((((uint64)n)>> 8) & 0xFF000000)  |
                 ((((uint64)n)>>24) & 0xFF0000)  |
                 ((((uint64)n)>>40) & 0xFF00)  |
                 ((((uint64)n)>>56) & 0xFF));
	}

#if ! VOS_LITTLE_ENDIAN

	//
	// Big-endian is native.
	//

	//
	// Native to Little-Endian 
	//
	inline int16 EndianSwap_NtoL_int16(int16 n)
	{
		return EndianSwap_int16(n);
	}

	inline int32 EndianSwap_NtoL_int32(int32 n)
	{
		return EndianSwap_int32(n);
	}

	inline int64 EndianSwap_NtoL_int64(int64 n)
	{
		return EndianSwap_int64(n);
	}

	inline uint16 EndianSwap_NtoL_uint16(uint16 n)
	{
		return EndianSwap_uint16(n);
	}

	inline uint32 EndianSwap_NtoL_uint32(uint32 n)
	{
		return EndianSwap_uint32(n);
	}

	inline uint64 EndianSwap_NtoL_uint64(uint64 n)
	{
		return EndianSwap_uint64(n);
	}

	//
	// Little-endian to Native
	//
	inline int16 EndianSwap_LtoN_int16(int16 n)
	{
		return EndianSwap_int16(n);
	}

	inline int32 EndianSwap_LtoN_int32(int32 n)
	{
		return EndianSwap_int32(n);
	}

	inline int64 EndianSwap_LtoN_int64(int64 n)
	{
		return EndianSwap_int64(n);
	}

	inline uint16 EndianSwap_LtoN_uint16(uint16 n)
	{
		return EndianSwap_uint16(n);
	}

	inline uint32 EndianSwap_LtoN_uint32(uint32 n)
	{
		return EndianSwap_uint32(n);
	}

	inline uint64 EndianSwap_LtoN_uint64(uint64 n)
	{
		return EndianSwap_uint64(n);
	}

	//
	// Native to Big-Endian
	//
	inline int16 EndianSwap_NtoB_int16(int16 n)
	{
		return n;
	}

	inline int32 EndianSwap_NtoB_int32(int32 n)
	{
		return n;
	}

	inline int64 EndianSwap_NtoB_int64(int64 n)
	{
		return n;
	}

	inline uint16 EndianSwap_NtoB_uint16(uint16 n)
	{
		return n;
	}

	inline uint32 EndianSwap_NtoB_uint32(uint32 n)
	{
		return n;
	}

	inline uint64 EndianSwap_NtoB_uint64(uint64 n)
	{
		return n;
	}

	//
	// Big-Endian to Native
	//
	inline int16 EndianSwap_BtoN_int16(int16 n)
	{
		return n;
	}

	inline int32 EndianSwap_BtoN_int32(int32 n)
	{
		return n;
	}

	inline int64 EndianSwap_BtoN_int64(int64 n)
	{
		return n;
	}

	inline uint16 EndianSwap_BtoN_uint16(uint16 n)
	{
		return n;
	}

	inline uint32 EndianSwap_BtoN_uint32(uint32 n)
	{
		return n;
	}

	inline uint64 EndianSwap_BtoN_uint64(uint64 n)
	{
		return n;
	}

#else

	// 
	// Little-endian is native.
	//

	//
	// Native to Little-Endian 
	//
	inline int16 EndianSwap_NtoL_int16(int16 n)
	{
		return n;
	}

	inline int32 EndianSwap_NtoL_int32(int32 n)
	{
		return n;
	}

	inline int64 EndianSwap_NtoL_int64(int64 n)
	{
		return n;
	}

	inline uint16 EndianSwap_NtoL_uint16(uint16 n)
	{
		return n;
	}

	inline uint32 EndianSwap_NtoL_uint32(uint32 n)
	{
		return n;
	}

	inline uint64 EndianSwap_NtoL_uint64(uint64 n)
	{
		return n;
	}

	//
	// Little-endian to Native
	//
	inline int16 EndianSwap_LtoN_int16(int16 n)
	{
		return n;
	}

	inline int32 EndianSwap_LtoN_int32(int32 n)
	{
		return n;
	}

	inline int64 EndianSwap_LtoN_int64(int64 n)
	{
		return n;
	}

	inline uint16 EndianSwap_LtoN_uint16(uint16 n)
	{
		return n;
	}

	inline uint32 EndianSwap_LtoN_uint32(uint32 n)
	{
		return n;
	}

	inline uint64 EndianSwap_LtoN_uint64(uint64 n)
	{
		return n;
	}

	//
	// Native to Big-Endian
	//
	inline int16 EndianSwap_NtoB_int16(int16 n)
	{
		return EndianSwap_int16(n);
	}

	inline int32 EndianSwap_NtoB_int32(int32 n)
	{
		return EndianSwap_int32(n);
	}

	inline int64 EndianSwap_NtoB_int64(int64 n)
	{
		return EndianSwap_int64(n);
	}

	inline uint16 EndianSwap_NtoB_uint16(uint16 n)
	{
		return EndianSwap_uint16(n);
	}

	inline uint32 EndianSwap_NtoB_uint32(uint32 n)
	{
		return EndianSwap_uint32(n);
	}

	inline uint64 EndianSwap_NtoB_uint64(uint64 n)
	{
		return EndianSwap_uint64(n);
	}

	//
	// Big-Endian to Native
	//
	inline int16 EndianSwap_BtoN_int16(int16 n)
	{
		return EndianSwap_int16(n);
	}

	inline int32 EndianSwap_BtoN_int32(int32 n)
	{
		return EndianSwap_int32(n);
	}

	inline int64 EndianSwap_BtoN_int64(int64 n)
	{
		return EndianSwap_int64(n);
	}

	inline uint16 EndianSwap_BtoN_uint16(uint16 n)
	{
		return EndianSwap_uint16(n);
	}

	inline uint32 EndianSwap_BtoN_uint32(uint32 n)
	{
		return EndianSwap_uint32(n);
	}

	inline uint64 EndianSwap_BtoN_uint64(uint64 n)
	{
		return EndianSwap_uint64(n);
	}
#endif
};

#endif // !vos_Endian_h
