//
// ErrorManager.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_ErrMgr_h
#define vos_ErrMgr_h


#include "String.h"
#include "Lists.h"

namespace vos
{
#ifdef _DEBUG
#define VOS_ERROR_TYPESAFE (1)

	/**************************************************************************

		IMPORTANT!

		This file (ErrorManager.h) contains the nuts and bolts of the error
		management system.

		It is highly recommended that you use the include file vOS\Errors.h,
		which will automatically include this file, and then provide numerous
		convenient macros, which you should use whenever possible for
		error-related code, as it allows the error system to be changed with
		minimal knock-on effects in your client code.
  
	**************************************************************************/

	
	/**************************************************************************
	
	>	class Error
	
		Author:		Jason Williams
	
		Purpose:	For debug (development) builds, we use an error class which 
					can be compared to see if an error occurred, but which will 
					not accept anything being assigned to it except for another 
					Error. This makes it much harder to intersperse vOS error
					handling with any old error handler (e.g. when returning 
					error values etc).

					NB. In release builds, Errors are simple and efficient bools.
  
	**************************************************************************/
	
	class Error
	{
		public:
			enum vos_Error_Constructor
			{
				Error_NoError		= 123456789,
				Error_RaiseError	= 987654321
			};

		public:
			Error(vos_Error_Constructor Init)	{ State = (Init == Error_NoError); };

			Error()							{ State = false; };
			Error(const Error &Source)		{ State = Source.State; };

			operator=(const Error &Source)	{ State = Source.State; };

			// Allow people to do  "if (!MyFunction()) ..."
			operator const bool() const		{ return(State); };

		private:
			// Generate errors if people try to mix vos::Errors with other return values
			Error(const bool Source)		{ ASSERT(0); };
			Error(const int Source)			{ ASSERT(0); };
			Error(const double Source)		{ ASSERT(0); };
			Error(const void *pSource)		{ ASSERT(0); };
			operator=(const bool Source)	{ ASSERT(0); return(*this); };
			operator=(const int Source)		{ ASSERT(0); return(*this); };
			operator=(const double Source)	{ ASSERT(0); return(*this); };
			operator=(const void *pSource)	{ ASSERT(0); return(*this); };

		private:
			bool State;
	};
#else
	// In release builds, Errors are simple and efficient bools
#define VOS_ERROR_TYPESAFE (0)

	typedef bool Error;
#endif


	typedef enum
	{
		Error_None = 0,			// No error has been reported.
		Error_Complaint,
		Error_Normal,
		Error_Critical,
		Error_Terminal,
	}
	TErrorType;

	/**************************************************************************
	
	>	class ErrorInfo : public ListItem
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Hold information about a registered Error.  This class
					is used by the ErrorManager to hold details of any error
					that has been registered with it.
	
	**************************************************************************/
	
	class ErrorInfo : public ListItem
	{
		public:
			/// Constructor
			ErrorInfo(int Code, 
					  TErrorType Type,
					  char* szSourceFile = NULL,
					  int LineNumber = 0);

			/// Constructor
			ErrorInfo(int Code, 
					  TErrorType Type, 
					  const String& Text,
					  char* szSourceFile = NULL,
					  int LineNumber = 0);

		public:
			//
			// Access to error details
			//

			// Get the error code from an error
			int			GetCode() const { return m_Code; }

			// Get the error type from an error
			TErrorType	GetType() const { return m_Type; }

			// Get the textual error message from an error
			String		GetText() const { return m_Text; }

			// Get the source file where the error was registered (usually only in debug builds)
			String		GetSourceFile() const { return m_SourceFile; }

			// Get the source file line number where the error was registered (usually only in debug builds)
			int			GetSourceLine() const { return m_SourceLine; }

			// Previous/next error to this one (if there is one)
			bool GetPrevError(const ErrorInfo** ppErrorInfo = NULL) const;
			bool GetNextError(const ErrorInfo** ppErrorInfo = NULL) const;

		private:
			// Details of the error
			int		   m_Code;
			TErrorType m_Type;
			String	   m_Text;

			// Where the error occured in the source code
			String	   m_SourceFile;
			int		   m_SourceLine;
	};

	/**************************************************************************
	
	>	class ErrorStack : public ListItem, public List
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Holds a stack of errors - this allows errors to be
					made more generic as control passes up through a chain
					of error handlers.

					For example, if a disk read error occurs, the DiskFile class
					will register this error, but the caller might then register
					another error to say, e.g. "data file could not be read", 
					which might in turn be made more generic by another error
					handler, such as "document could not be loaded."

					Thus a (possibly) more useful error message can be given
					to the user.  Note that this is completely optional, and
					depends on the code that calls the function that registers
					the original error.  It is perfectly acceptable not to 
					register any errors in addition to the original cause of
					the problem.

					This class provides the data structure for maintaining this
					'stack' of errors.
	
	**************************************************************************/
	
	class ErrorStack : public ListItem, public List
	{
		public:
			ErrorStack() {}
			~ErrorStack();

		public:
			// Standard push/pop access.
			void Push(ErrorInfo*);
			ErrorInfo* Pop();
			
			// Get the top error on the stack, but do not remove it
			ErrorInfo* Peek();

			// Returns true if there are any errors outstanding
			bool ErrorsOutstanding();

			// Remove all errors from the stack.
			void Clear();
	};

	/**************************************************************************
	
	>	class ErrorStateStack : public List
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Allows the current error stack/state to be preserved and
					then restored.

					This is mainly used when an error handling routine wishes
					to preserve the error state for its calling routine, but
					it wishes to call some functions which may fail.

					For example, when dealing with files, an error handler
					may decide to try to close the problem file, but calling
					Close() may itself cause an error, thus corrupting the
					details of the original problem.  Therefore the error
					routine should preserve the error state - see the
					ErrorManager::PushState() and ErrorManager::PopState()
					routines for details.  (These functions use the
					ErrorStateStack class to implement their behaviour.)
	
	**************************************************************************/
	
	class ErrorStateStack : public List
	{
		public:
			ErrorStateStack() {}

		public:
			void Push(ErrorStack* pNewState);
			ErrorStack* Pop();
	};

	/**************************************************************************
	
	>	class ErrorReporter
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	Interface for a class that can report errors.  An object
					derived from this class can be installed as an error
					handler with the ErrorManager object.

					Implement this to provide custom error reporting.

					For example, the vGUI library provides a derivation of
					this class for reporting errors in a dialog box.
	
	**************************************************************************/
	
	class ErrorReporter
	{
		public:
			virtual ~ErrorReporter() {}

		public:
			// Implement this function to report an error
			virtual void ReportError(ErrorStack* pErrors) = 0;
	};

	/**************************************************************************
	
	>	class ErrorManager
	
		Author:		Tim Browse
		Date:		13/4/2002
	
		Purpose:	The hub of the vOS error system.  All clients should
					use this class to report/manage errors.

					The ErrorManager provides the following functionality:

						- standard structure for an error
						- centralised error registration/reporting
						- nested error registration
						- a stack of multiple error states

					IMPORTANT:

					You should look at the vos\Errors.h file, and use the
					macros in there whenever possible, as they provide 
					convenient forms and abstractions that make changing your
					program's behaviour much easier than if you use the 
					Error Manager functions directly.
					
					An application should call RegisterError() when it wants
					to register an error. Registering an error is not the same
					as reporting an error. Registering an error simply informs
					the ErrorManager that an error has occurred.  An application's
					code can either recover from this state, and call ClearErrors(),
					or it can fail and decide to report the error to the user,
					using ReportError().

					It is up to the application to install an appropriate
					ErrorReporter object - if none is installed, then the
					Debugger::Interrupt() function is called in the debug 
					build.  In release builds, nothing happens - the error
					state is just cleared by calling ClearErrors().

					To re-emphasise: if you do not install an error reporter,
					then in release builds errors will be silently ignored.

					To get information about the current error, call GetErrorInfo().

					Be careful of corrupting the error state when writing error
					handling code - you may call functions that register errors,
					and thus corrupt the error state.  Use the PushState() and
					PopState() functions to avoid this.

					NB. See the vOS\Errors.h include file for convenient macros
						for accessing the ErrorManager's functionality.

	**************************************************************************/
	
	class ErrorManager
	{
		public:
			ErrorManager();
			~ErrorManager();

		public:
			/// Register a new error that has just occurred.
			Error RegisterError(int Code, 
								TErrorType Type, 
								const char *szReason,
								char* szSourceFile = NULL,
								int LineNumber = 0);

			/// Report current error using installed error reporter object.
			void ReportError();

			/// Clear all errors - the caller has recovered from the outstanding error(s).
			void ClearErrors();

			/// Get details of last error reported.
			const ErrorInfo* GetLastError() const;

			/// Used to pass on an error to a caller.
			Error LastError() const;

			/// Used to check that there are no outstanding errors.
			Error NoError() const;

			/// Install an error reporter
			void SetErrorReporter(ErrorReporter *pReporter);

			/// Push (save) current error state.
			void PushState();

			/// Pop (restore) error state previously saved by PushState().
			void PopState();

		private:
			/// The object returned by GetLastError() when no errors are outstanding.
			static ErrorInfo s_NoError;

			/// The current outstanding errors (if there are any).
			ErrorStack* m_pCurrentErrors;

			/// List of error stacks for use when pushing/popping error manager state.
			ErrorStateStack m_StateStack;

			/// The installed error reporter (if there is one).
			ErrorReporter *m_pReporter;
	};

	/// Singleton error manager object.
	ErrorManager& TheErrorManager();

	//
	// Convenient global functions.
	//

	/// Global function for accessing the information about the last registered error.
	const ErrorInfo* GetLastError();

	/// Global function for acessing the error code of the last registered error.
	int				 GetLastErrorCode();

	/// Global function for accessing the type of the last registered error.
	TErrorType		 GetLastErrorType();
};

#endif // !vos_ErrMgr_h
