//
// Errors.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_Errors_h
#define vos_Errors_h

#include "ErrorManager.h"
#include "ErrorCodes.h"

/** \file 
	This include file contains macros to simplify using the vOS error system.

	The vOS ErrorManager class (and associated classes) can be used directly,
	but the macros contained in this header file are intended to make it
	easier to write code to use the error system, and to write code that
	is less verbose and hence easier to understand.

	Use of these macros is completely optional - you can use the ErrorManager
	class directly, or write your own set of macros for error handling if
	you so wish - the vOS Error Manager will stil work properly (as long
	as you call its member functions correctly, of course).
*/

/*!
	Macro: VNOERROR

    Use this macro to return a value (of type Error) that indicates no 
	error occurred.

	For example:

		\code
		
		  // It all worked!
		  return VNOERROR;

		\endcode

*/

/*!
	Macro: LASTERROR

	Use this macro when you have called a function that returns an Error
	indicating that the process failed, and you wish to pass this error
	back to your caller, e.g.

	\code
		
		if (!MyFunctionThatReturnsError())
			// Error occurred - return to caller
			return LASTERROR;

		// No error - carry on normal processing...
		Blah();

	\endcode

	NB. LASTERROR evaluates to a value of type Error, so your function must
		obviously be prototyped to return this type of value for this to work.
*/


/*!
	Macro: ERROR_HANDLED()

	Call this macro when you have handled an error, and want to clear
	the error state, e.g.

	\code

		if (!MyFunction())
		{
			// Error occurred...handle it
			MyErrorHandlingCode();

			// Ok, we handled the error...
			ERROR_HANDLED();
		}

		// Normal processing
		Blah();

	\endcode
*/


/*!
	Macro: ERRMGR()

	Use this macro as the preferred way to get hold of the vOS ErrorManager
	object, e.g.

	\code

		ERRMGR().PushState();

	\endcode
*/


/*!
	Macro: VSUCCESS

	Use this macro when you have stored an Error value, and wish to compare
	it against something later to see if it was successful, e.g.

	\code

		// Allocate buffer and call our function (that may fail)
		byte pBuffer = new byte[20];

		Error Result = MyFunction(pBuffer);

		// Delete temporary buffer
		delete [] pBuffer;

		// Now, did it work?
		if (Result == VSUCCESS)
		{
			// Yes...
		}
		else
		{
			// No...error occurred
			return LASTERROR;
		}

	\endcode

	NB. If you arrange the conditional the other way around, you don't need this 
		macro, as you can just do this:

	\code

		if (!Result)
		{
			// it failed...
			return LASTERROR;
		}
		else
		{
			// it worked!
		}

	\endcode

	But this is not always the most 'natural' way of arranging code, hence
	the VSUCCESS macro is available.
*/


/*!
	Macro: REPORT_ERROR()

	Use this macro to cause any outstanding errors to be reported to
	the user.  This is done using whatever ErrorReporter object you have
	installed.

	NB. If you have not installed an error reporter, then calling this
		macro in release builds does nothing.

	\code

		if (!ProcessData())
		{
			// Oops - something went wrong; tell the user
			REPORT_ERROR();
		}

	\endcode
*/


/*!
	Macro: NORMAL_ERROR(ErrCode, szReason)

	Use this macro to report an error of a normal kind, e.g. file not
	found, that the program may well be able to recover from on its own,
	without need user intervention (although the user may be informed about
	the error).

	e.g.

	\code

		vos::Error LoadDocument()
		{
			if (!FindDocument())
			{
				return NORMAL_ERROR(Err_MyApp_DocumentNotFound,
									"Unable to load your document!");
			}

			// Ok, load the document...
			Blah();

			// Worked ok
			return VNOERROR;
		}

	\endcode

*/


/*!
	Macro: CRITICAL_ERROR(ErrCode, szReason)

	Use this macro to report a critical error.  The meaning of a 'critical'
	error is largely up to you, but it generally means that the program
	needs user intervention of some kind before continuing.

	e.g.

	\code

		if (!CheckForCD())
		{
			return CRITICAL_ERROR(Err_MyApp_CDMissing,
								  "Please insert the program CD.");
		}

	\endcode

	As mentioned, the vOS error system places no special meaning on
	critical errors, except that they have the type 'Error_Critical'.
	It is up to you to decide what this means.  (If you want, you can
	report all errors as normal errors if it makes it easier.)
*/


/*!
	Macro: TERMINAL_ERROR(ErrCode, szReason)

	Use this macro to report a 'terminal' error.  As with critical errors,
	it is up to you to decide what a terminal error means, but it is
	intended to register errors that cannot be recovered from, and so 
	the program must end, e.g.

	\code

		if (RAMIsMissing())
		{
			TERMINAL_ERROR(Err_MyApp_RAMMissing,
						   "Can't find any RAM!");
			REPORT_ERROR();
			abort();
		}

	\endcode

	NB. Notice in this example that although the macro returns an Error
		value you may pass back to your caller, you don't have to use it.
		Using the macro in this way has the effect of registering the
		error, but allowing you to handle it yourself.
*/


/*!
	Macro: COMPLAIN(szReason)

	\arg szReason a textual description of what	went wrong (see code example)

	This macro is the equivalent of assert-style functionality.

	i.e. you should use this to report errors in debug builds, and
		 to perform consistency checks, etc.

	Basically, use this the way you would an assert macro - like
	assert, the COMPLAIN() macro has no effect in non-debug builds.

	Notice this macro always causes (and reports) an error. If you
	want a conditional form like assert(), then use COMPLAIN_IF().

	\code

		if (!MyFunction())
		{
			// This is not good
			COMPLAIN("The wheels have come off!");
			DoCleanUp();
			return;
		}

	\endcode

	NB. You can see from this example that sometimes COMPLAIN_IF() is not
		appropriate, which is why COMPLAIN() is available. To see what
		I mean, try rewriting the above code using COMPLAIN_IF() instead 
		of COMPLAIN().
*/


/*!
	Macro: COMPLAIN_IF(cond, szReason)

	\arg cond is the condition to check - if it evaluates to true, an
	error is reported.

	\arg szReason is the textual description of what went wrong (see code example)

	Use this macro to check a condition in debug builds, and report an
	error immediately if the condition is true.

	This macro is similar to COMPLAIN(), but it has a condition, which
	makes it more like assert - BUT: the condition is inverted if you
	compare it to assert, e.g. these two are equivalent, but notice that
	the condition is inverted:

	\code

		assert(cBytes > 0);

	\endcode

	and:

	\code
		
		COMPLAIN_IF(cBytes == 0, "No bytes were read!");

	\endcode

	COMPLAIN_IF() disappears in non-debug builds - in debug builds, it
	registers an error and causes it to be reported to the user immediately.
*/


/*!
	Macro: ERROR_NOMEMORY()

	A handy macro to report the standard error state of a memory allocation
	failing.  The standard vOS error code Err_OutOfMemory is used.
*/


/*!
	Macro: RETURN_ON_FAILURE(stmnt)

	Use this macro when calling functions (that return an Error value) that 
	must succeed, otherwise	you will just return the error to the caller.

	i.e. instead of this:

	\code

		if (!MyFunction())
			return LASTERROR;

	\endcode

	you can do this:

	\code

		RETURN_ON_FAILURE(MyFunction());

	\endcode

	NB. This is only useful when you don't have to do any error clean-up;
		i.e. you really just need to return LASTERROR.
*/

  
//////////////////////////////////////////////////////////////////
//  DEPRECATED - do not use:									//
//////////////////////////////////////////////////////////////////

/*!
	Macro: NOERROR

	Deprecated!
	
	The NOERROR macro has been replaced by VNOERROR to avoid clashes
	with the Win32 header files - do not use it.
*/

/*!
	Macro: ERROR()

	Deprecated!
	
	The ERROR() macro has been replaced by NORMAL_ERROR() to avoid
	clashes with the Win32 header files - do not use it.
*/


//////////////////////////////////////////////////////////////////


// Unset this to disable error logging (to C:\vOSErrorLog.txt) in release builds
//#define VOS_LOG_ERRORS (1)
#undef VOS_LOG_ERRORS

#ifdef VOS_LOG_ERRORS
	// See Errors.h - this reports all VNOERROR, LASTERROR, REPORTERROR,
	// COMPLAIN, and COMPLAIN_IFs that go off. This gives us a chance to
	// actually find out what the hell happened, because otherwise the
	// error manager is totally useless in release builds.
	// See ErrorManager.cpp for implementation
	extern bool LogReport(char *Report, int Line, char *File, bool ReturnCode);
#endif


//
// Some macros to interface to the error manager.
//
// NB. This is one possible interface to the error manager - you
//     don't have to use it.
//

#if defined WIN32

// Lose annoying Win32 macros
#undef ERROR
#undef NOERROR

#elif defined MACOS

// Lose annoying Open Transport macro
#undef NOERROR

#endif

#define ERROR_HANDLED() \
		ERRMGR().ClearErrors()


#if VOS_ERROR_TYPESAFE
#define SUCCESS  vos::Error(vos::Error::Error_NoError)
#define VSUCCESS vos::Error(vos::Error::Error_NoError)
#else
#define SUCCESS  (true)
#define VSUCCESS (true)
#endif

//
// Access to the error manager - normally, you call TheErrorManager(),
// but for DLLs (on whatever platform) to use the same error manager as the 
// parent process (which is usually desirable), we call a different function,
// called dll::TheErrorManager().
//
// However, you must #define VOS_USE_DLL_ERRORMGR when compiling DLLs for 
// this to work, and provide an implementation of dll::TheErrorManager() in 
// your DLL, which returns a reference to the correct error manager object.
//
// NB. The term DLL is used to mean any kind of dynamic link library,
//     not just Windows DLLs.


#ifdef VOS_USE_DLL_ERRORMGR

//
// DLL Error manager access - you should provide the implementation
// of this function in your DLL.
//
namespace dll
{
	vos::ErrorManager& TheErrorManager();
}

#define ERRMGR() dll::TheErrorManager()

#else

//
// Normal Error manager access
//
#define ERRMGR() vos::TheErrorManager()

#endif


//
// Error registering macros etc.
//
#ifdef VOS_LOG_ERRORS
#define REPORT_ERROR()													\
		{																\
			LogReport("ReportError called", __LINE__, __FILE__, false);	\
			ERRMGR().ReportError();										\
		}
#else
#define REPORT_ERROR() \
		ERRMGR().ReportError()
#endif

#ifdef _DEBUG

#define VNOERROR ERRMGR().NoError()

#define NORMAL_ERROR(ErrCode, szReason)							\
		ERRMGR().RegisterError(ErrCode,						\
							   vos::Error_Normal,			\
							   szReason,					\
							   __FILE__,					\
							   __LINE__)

#define ERROR(ErrCode, szReason)							\
		ERRMGR().RegisterError(ErrCode,						\
							   vos::Error_Normal,			\
							   szReason,					\
							   __FILE__,					\
							   __LINE__)

#define CRITICAL_ERROR(ErrCode, szReason)				\
		ERRMGR().RegisterError(ErrCode,					\
							   vos::Error_Critical,		\
							   szReason,				\
							   __FILE__,				\
							   __LINE__)

#define TERMINAL_ERROR(ErrCode, szReason)				\
		ERRMGR().RegisterError(ErrCode,					\
							   vos::Error_Terminal,		\
							   szReason,				\
							   __FILE__,				\
							   __LINE__)

#define LASTERROR \
		ERRMGR().LastError()

#define COMPLAIN(szReason)																	\
		{																					\
			ERRMGR().RegisterError(0, vos::Error_Complaint, szReason, __FILE__, __LINE__);	\
			ERRMGR().ReportError();															\
		}

#define COMPLAIN_IF(cond, szReason)		\
		{								\
			if (cond)					\
			{							\
				COMPLAIN(szReason);		\
			}							\
		}

#else

#ifdef VOS_LOG_ERRORS
#define VNOERROR LogReport("VNOERROR returned", __LINE__, __FILE__, true)
#else
#if VOS_ERROR_TYPESAFE
#define VNOERROR vos::Error(vos::Error::Error_NoError)
#else
#define VNOERROR (true)
#endif
#endif

#define ERROR(ErrCode, szReason) \
		ERRMGR().RegisterError(ErrCode, vos::Error_Normal, szReason)

#define NORMAL_ERROR(ErrCode, szReason) \
		ERRMGR().RegisterError(ErrCode, vos::Error_Normal, szReason)

#define CRITICAL_ERROR(ErrCode, szReason) \
		ERRMGR().RegisterError(ErrCode, vos::Error_Critical, szReason)

#define TERMINAL_ERROR(ErrCode, szReason) \
		ERRMGR().RegisterError(ErrCode, vos::Error_Terminal, szReason)

#ifdef VOS_LOG_ERRORS
#define LASTERROR LogReport("LASTERROR returned", __LINE__, __FILE__, false)

#define COMPLAIN(szReason) LogReport("COMPLAIN failed", __LINE__, __FILE__, false)

#define COMPLAIN_IF(cond, szReason) { if (cond) LogReport("COMPLAIN_IF failed", __LINE__, __FILE__, false); }
#else
#if VOS_ERROR_TYPESAFE
#define LASTERROR vos::Error(vos::Error::Error_RaiseError)
#else
#define LASTERROR (false)
#endif

#define COMPLAIN(szReason)	((void)0)

#define COMPLAIN_IF(cond, szReason)	((void)0)
#endif


#endif

//
// Easy way to report out of memory
//
#define ERROR_NOMEMORY() \
		ERROR(Err_OutOfMemory, msgvos_NoMemory)


// Call a function. If it succeeds, continue running. If it fails, return the
// error immediately.
#define RETURN_ON_FAILURE(x)	{ if (!(x)) return(LASTERROR); }

#endif // !vos_Errors_h
