//
// File.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "File.h"

#include "Endian.h"

#include "Errors.h"


//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/**************************************************************************

	>	File::File()

		Author:		Tim Browse
		Date:		23/08/99

		Purpose:	Constructor

	**************************************************************************/

	File::File()
	{
	}

	/**************************************************************************

	>	File::~File()

		Author:		Tim Browse
		Date:		23/08/99

		Purpose:	Destructor.

	**************************************************************************/

	File::~File()
	{
	}


	/**************************************************************************

	>	virtual bool File::IsOpen() const

		Author:		Jason
		Date:		15/1/2001

		Returns:	true if the file is open, false if it is not

		Purpose:	Determine if this file is open.

		Notes:		The base class always returns false.

	**************************************************************************/

	bool File::IsOpen() const
	{
		return(false);
	}


	/**************************************************************************

	>	virtual Error File::GetSize(uint* pSize)

		Author:		Jason
		Date:		15/1/2001

		Outputs:	pSize	- The size, in bytes, of this file

		Purpose:	Determine the size (in bytes) of this file

		Notes:		The base-class fudges this result by seeking to the end of
					the file and reading the file position. While this works in
					a portable manner, it may cause grief with some filing
					systems, like flushing read buffers, etc. Derived classes
					should override this to determine the size in a less
					dodgy manner.

	**************************************************************************/

	Error File::GetSize(uint* pSize)
	{
		COMPLAIN_IF(pSize == NULL, "Illegal NULL param");

		if (!IsOpen())
		{
			return ERROR(Err_File_NotOpen, "File::GetSize called for non-open file");
		}

		if (pSize != NULL)
		{
			uint OldPos = 0;
			RETURN_ON_FAILURE(GetPos(&OldPos));
			RETURN_ON_FAILURE(SeekFromEnd(0));
			RETURN_ON_FAILURE(GetPos(pSize));
			RETURN_ON_FAILURE(SeekFromStart(OldPos));
		}

		return(VNOERROR);
	}


	/**************************************************************************

	>	Error File::Read_byte(byte* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the byte read from the file.
		Returns:	Standard error

		Purpose:	Read a byte from the file.

	**************************************************************************/

	Error File::Read_byte(byte* pData)
	{
		return Read(pData, 1, NULL);
	}

	/**************************************************************************

	>	Error File::Read_bool(bool* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the bool read from the file.
		Returns:	Standard error

		Purpose:	Read a bool value from the file.

	**************************************************************************/

	Error File::Read_bool(bool* pData)
	{
		uint8 Data;

		if (Read(&Data, 1, NULL))
		{
			*pData = (Data != 0);
			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Read_int8(int8* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the int8 value read from the file.
		Returns:	Standard error

		Purpose:	Read an int8 value from the file.

	**************************************************************************/

	Error File::Read_int8(int8* pData)
	{
		return Read(pData, 1, NULL);
	}

	/**************************************************************************

	>	Error File::Read_int16(int16* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the int16 read from the file.
		Returns:	Standard error

		Purpose:	Read an int16 value from the file.

					NB. Performs endian swap correction - it assumes that
						data in the file is in little-endian format.

	**************************************************************************/

	Error File::Read_int16(int16* pData)
	{
		int16 Data;

		if (Read(&Data, 2, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			*pData = EndianSwap_LtoN_int16(Data);

			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Read_int32(int32* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the in32 read from the file.
		Returns:	Standard error

		Purpose:	Read an int32 value from the file.

					NB. Performs endian swap correction - it assumes that
						data in the file is in little-endian format.

	**************************************************************************/

	Error File::Read_int32(int32* pData)
	{
		int32 Data;
		if (Read(&Data, 4, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			*pData = EndianSwap_LtoN_int32(Data);

			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Read_int64(int64* pData)

		Author:		Rob Myers
		Date:		26/10/01
		Outputs:	pData - the int64 value read from the file.
		Returns:	Standard error

		Purpose:	Read an int64 value from the file.

	**************************************************************************/

	Error File::Read_int64(int64* pData)
	{
		int64 Data;

		if (Read(&Data, 8, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			*pData = EndianSwap_LtoN_int64(Data);

			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Read_uint8(uint8* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the uint8 value read from the file.
		Returns:	Standard error

		Purpose:	Read a uint8 value from the file.

	**************************************************************************/

	Error File::Read_uint8(uint8* pData)
	{
		return (Read(pData, 1, NULL));
	}

	/**************************************************************************

	>	Error File::Read_uint16(uint16* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the uint16 value read from the file.
		Returns:	Standard error

		Purpose:	Read a uint16 value from the file.

					NB. Performs endian swap correction - it assumes that
						data in the file is in little-endian format.

	**************************************************************************/

	Error File::Read_uint16(uint16* pData)
	{
		uint16 Data;
		if (Read(&Data, 2, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			*pData = EndianSwap_LtoN_uint16(Data);

			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Read_uint32(uint32* pData)

		Author:		Tim Browse
		Date:		23/08/99
		Outputs:	pData - the uint32 value read from the file.
		Returns:	Standard error

		Purpose:	Read an uint32 value from the file.

	**************************************************************************/

	Error File::Read_uint32(uint32* pData)
	{
		uint32 Data;

		if (Read(&Data, 4, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			*pData = EndianSwap_LtoN_uint32(Data);

			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Read_uint64(uint64* pData)

		Author:		Rob Myers
		Date:		26/10/01
		Outputs:	pData - the uint64 value read from the file.
		Returns:	Standard error

		Purpose:	Read an uint64 value from the file.

	**************************************************************************/

	Error File::Read_uint64(uint64* pData)
	{
		uint64 Data;

		if (Read(&Data, 8, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			*pData = EndianSwap_LtoN_uint64(Data);

			return VNOERROR;
		}

		return LASTERROR;
	}


	/**************************************************************************

	>	Error File::Read_float32(float32* pData)

		Author:		Rob Myers
		Date:		24/10/01
		Outputs:	pData - the float32 value read from the file.
		Returns:	Standard error

		Purpose:	Read an float32 value from the file.

	**************************************************************************/

	Error File::Read_float32(float32* pData)
	{
		int32 Data;

		if (Read(&Data, 4, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			Data = EndianSwap_LtoN_uint32(Data);
			*pData = *(float32*)&Data;

			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Read_float64(float64* pData)

		Author:		Rob Myers
		Date:		24/10/01
		Outputs:	pData - the float64 value read from the file.
		Returns:	Standard error

		Purpose:	Read an float64 value from the file.

	**************************************************************************/

	Error File::Read_float64(float64* pData)
	{
		int64 Data;

		if (Read(&Data, 8, NULL))
		{
			// Little endian is the format used in files, so we
			// convert back to native format after loading the data.
			Data = EndianSwap_LtoN_uint64(Data);
			*pData = *(float64*)&Data;

			return VNOERROR;
		}

		return LASTERROR;
	}

	/**************************************************************************

	>	Error File::Write_byte(byte Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The byte to write to the file
		Returns:	Standard error

		Purpose:	Write a byte to the file.

	**************************************************************************/

	Error File::Write_byte(byte Data)
	{
		return Write(&Data, 1, NULL);
	}

	/**************************************************************************

	>	Error File::Write_bool(bool Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The bool to write to the file.
		Returns:	Standard error

		Purpose:	Write a bool value to the file.

	**************************************************************************/

	Error File::Write_bool(bool Data)
	{
		return Write_byte((byte) Data);
	}

	/**************************************************************************

	>	Error File::Write_int8(int8 Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The int8 value to write to the file.
		Returns:	Standard error

		Purpose:	Write an int8 value to the file.

	**************************************************************************/

	Error File::Write_int8(int8 Data)
	{
		return Write_byte((byte) Data);
	}

	/**************************************************************************

	>	Error File::Write_int16(int16 Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The in16 value to write to the file.
		Returns:	Standard error

		Purpose:	Write an int16 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_int16(int16 Data)
	{
		// Little endian is the format used in files, so we
		// convert to it before writing data.
		int16 TheData = EndianSwap_NtoL_int16(Data);

		return Write(&TheData, 2, NULL);
	}

	/**************************************************************************

	>	Error File::Write_int32(int32 Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The int32 value to write to the file.
		Returns:	Standard error

		Purpose:	Write an int32 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_int32(int32 Data)
	{
		// Little endian is the format used in files, so we
		// convert to it before writing data.
		int32 TheData = EndianSwap_NtoL_int32(Data);

		return Write(&TheData, 4, NULL);
	}
	
	/**************************************************************************

	>	Error File::Write_int64(int64 Data)

		Author:		Rob Myers
		Date:		26/10/01
		Inputs:		The int64 value to write to the file.
		Returns:	Standard error

		Purpose:	Write a int64 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_int64(int64 Data)
	{
		// Little endian is the native format used in files, so we
		// convert to it before writing data.
		int64 TheData = EndianSwap_NtoL_int64(Data);

		return (Write(&TheData, 8, NULL));
	}

	/**************************************************************************

	>	Error File::Write_uint8(uint8 Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The uint8 value to write to the file.
		Returns:	Standard error

		Purpose:	Write a uint8 value to the file.

	**************************************************************************/

	Error File::Write_uint8(uint8 Data)
	{
		return Write_byte((byte) Data);
	}

	/**************************************************************************

	>	Error File::Write_uint16(uint16 Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The uint16 value to write to the file.
		Returns:	Standard error

		Purpose:	Write a uint16 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_uint16(uint16 Data)
	{
		// Little endian is the format used in files, so we
		// convert to it before writing data.
		uint16 TheData = EndianSwap_NtoL_uint16(Data);

		return Write(&TheData, 2, NULL);
	}

	/**************************************************************************

	>	Error File::Write_uint32(uint32 Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		The uint32 value to write to the file.
		Returns:	Standard error

		Purpose:	Write a uint32 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_uint32(uint32 Data)
	{
		// Little endian is the native format used in files, so we
		// convert to it before writing data.
		uint32 TheData = EndianSwap_NtoL_uint32(Data);

		return (Write(&TheData, 4, NULL));
	}
	
	/**************************************************************************

	>	Error File::Write_uint64(uint64 Data)

		Author:		Rob Myers
		Date:		26/10/01
		Inputs:		The uint64 value to write to the file.
		Returns:	Standard error

		Purpose:	Write a uint64 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_uint64(uint64 Data)
	{
		// Little endian is the native format used in files, so we
		// convert to it before writing data.
		uint64 TheData = EndianSwap_NtoL_uint64(Data);

		return (Write(&TheData, 8, NULL));
	}

	/**************************************************************************

	>	Error File::Write_float32(float32 Data)

		Author:		Rob Myers
		Date:		24/10/01
		Inputs:		The float32 value to write to the file.
		Returns:	Standard error

		Purpose:	Write a float32 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_float32(float32 Data)
	{
		// Little endian is the native format used in files, so we
		// convert to it before writing data.
		uint32 TheData = EndianSwap_NtoL_uint32(*(uint32*)&Data);

		return (Write(&TheData, 4, NULL));
	}

	/**************************************************************************

	>	Error File::Write_float64(float64 Data)

		Author:		Rob Myers
		Date:		24/10/01
		Inputs:		The float64 value to write to the file.
		Returns:	Standard error

		Purpose:	Write a float64 value to the file.

					NB. Performs endian swap correction - it assumes that
						data in the file should be in little-endian format.

	**************************************************************************/

	Error File::Write_float64(float64 Data)
	{
		// Little endian is the native format used in files, so we
		// convert to it before writing data.
		uint64 TheData = EndianSwap_NtoL_uint64(*(uint64*)&Data);

		return (Write(&TheData, 8, NULL));
	}

	/**************************************************************************

	>	Error File::WriteLine(const String& Data)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		Data - the string to write to the file.
		Returns:	Standard error

		Purpose:	Write a line of text to the file, terminated by a 
					line end character sequence.

	**************************************************************************/

	Error File::WriteLine(const String& Data)
	{
		int cBytes = Data.GetLength();

		if (!Write(Data.c_str(), cBytes, NULL))
		{
			// Error writing string
			return LASTERROR;
		}

		// Write line ending
		return WriteLineEnd();
	}

	/**************************************************************************

	>	Error File::WriteLine(const char* szData)

		Author:		Tim Browse
		Date:		23/08/99
		Inputs:		szData - the string to write to the file.
		Returns:	Standard error

		Purpose:	Write a line of text to the file, terminated by a 
					line end character sequence.

	**************************************************************************/

	Error File::WriteLine(const char* szData)
	{
		if (!Write(szData, strlen(szData), NULL))
		{
			// Error writing string
			return LASTERROR;
		}

		// Write line ending
		return WriteLineEnd();
	}

	/**************************************************************************

	>	Error File::WriteLineEnd()

		Author:		Tim Browse
		Date:		23/08/99
		Returns:	Standard error

		Purpose:	Write the line-end character sequence to the file.

	**************************************************************************/

	Error File::WriteLineEnd()
	{
		// Line ending depends on platforms...

		//
		// NB. These are from memory, so don't be surprised if they're wrong! :)
		//

#if VOS_MSWINDOWS
		// CRLF for Windows
		const char LineEnd[] = "\r\n";
		const int cBytes = 2;
#elif VOS_MACOS
		// CR for Mac OS
		const char LineEnd[] = "\r";
		const int cBytes = 1;
#else
		// LF for Unix derivatives/anything else.
		const char LineEnd[] = "\n";
		const int cBytes = 1;
#endif

		return Write(LineEnd, cBytes, NULL);
	}

	/**************************************************************************

	>	Error File::WriteString(const String& Data)

		Author:		Rob Myers
		Date:		15/10/01
		Inputs:		Data - The String to write
		Returns:	Standard error

		Purpose:	Write the String to the file.

	**************************************************************************/
	
	Error File::WriteString(const String& Data)
	{
		long cBytes = Data.GetLength();
		if (!Write_int32(cBytes))
		{
			// Error writing string length
			return LASTERROR;
		}

		// Write the bytes, remembering that a String is an array of tchars
		if (!Write(Data.c_str(), cBytes * sizeof(tchar), NULL))
		{
			// Error writing string
			return LASTERROR;
		}

		return VNOERROR;		
	}
	
	/**************************************************************************

	>	Error File::ReadString(const String& Data)

		Author:		Rob Myers
		Date:		15/10/01
		Inputs:		pData - The String to read into
		Returns:	Standard error

		Purpose:	Read the String from the file.
		Note:		pData must not be locked.

	**************************************************************************/
	
	Error File::ReadString(String * pData)
	{
		int32 cBytes;
		if(!Read_int32(&cBytes))
		{
			// Error reading string length
			return LASTERROR;
		}
		
		// Make sure the string's buffer is the right size
		char * pBuffer = pData->LockBuffer(cBytes + 1);
		
		// Read in the string, remembering that a String is an array of tchars
		Error readError = Read(pBuffer, cBytes * sizeof(tchar), NULL);

		// Null terminate the string so UnlockBuffer() gets the length right
		pBuffer[cBytes] = NULL;
		
		// Unlock the string buffer, allowing the String to ruffle its feathers
		pData->UnlockBuffer();
		
		// Return the last file-based error
		return readError;
	}
};


