//
// File.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_File_h
#define vos_File_h

#include "vos.h"

#include "FileSpec.h"
#include "ErrorManager.h"
#include "ErrorCodes.h"

namespace vos
{
	//
	// File error codes
	//

	enum
	{
		Err_File_NotFound			= Err_File_Errors,
		Err_File_NotOpen,
		Err_File_OpenFailed,
		Err_File_EOF,
		Err_File_SeekFail,
		Err_File_GetPosFail,
		Err_File_ReadFail,
		Err_File_WriteFail,
		Err_File_DeleteFail,
		Err_File_AccessDenied,
		Err_File_NoSpaceAvailable,
		Err_File_AlreadyExists,
		Err_File_CreateFailed,

		Err_File_InvalidFileFormat,

		Err_FileLock_Fail,
	};

	//
	// File access modes (used when opening files)
	//

	typedef enum
	{
		fam_Read = 1,
		fam_Write = 2,
		fam_Full = fam_Read|fam_Write,
		fam_Text = 4,
		fam_Append = 8
	}
	FileMode;

	//
	// Allow bitwise ops on FileMode values
	//
	inline FileMode operator|(FileMode lhs, FileMode rhs)
	{
		return (FileMode) (((int) lhs) | ((int) rhs));
	}

	inline FileMode operator|=(FileMode lhs, FileMode rhs)
	{
		return (FileMode) (((int) lhs) | ((int) rhs));
	}

	inline FileMode operator&(FileMode lhs, FileMode rhs)
	{
		return (FileMode) (((int) lhs) & ((int) rhs));
	}

	/**************************************************************************

	>	class File

		Author:		Tim Browse

		Date:		23/08/99

		Purpose:	Provide simple and efficient file class without a 
					'candy-machine' interface, that encourages proper 
					error-handling by the client.

					This class is deliberately designed as a 'raw' file class;
					no clever object persistence or text formatting/number
					precision is included.  These services can be built on
					top of this class if required (or you can use iostream
					objects), but this class is designed as a 'no-surprises' 
					file class - never again will I waste a day because I've 
					forgotten to disable text mode on a file (this File class 
					assumes binary mode is the default).

					The only data conversion that occurs is in the following
					functions:

						- Read_int16()

						- Read_int32()

						- Read_uint16()

						- Read_uint32()

						- Write_int16()

						- Write_int32()

						- Write_uint16()

						- Write_uint32()

  
					The read functions all convert data to little-endian format
					before saving to the file, and the write functions all
					convert from little-endian to native format after reading
					the data from the file (obviously, on little-endian
					platforms, this means the conversion is a no-op).

					This means that you can save data structures to a file 
					using these functions, and the data file will be 
					compatible between big-endian and little-endian platforms.

					NB. The compatibility is obviously compromised if you also
					    save some data to the file using the raw read/write
						functions, and that data is endian-sensitive.

		How to use:

					Create a file with something like:

					\code

						DiskFile Infile;
						Infile.Create("c:\\MyFile.txt");
					
					\endcode

					Once created, you can read/write it. You may open/close the
					file more than once without having to re-create it.

					\code

						if (!Infile.Open(fam_Read))
							return(LASTERROR);

						int8 MyByte = 0;
						if (!Infile.Read_int8(&MyByte))
							return(LASTERROR)

						Infile.Close();

					\endcode

	**************************************************************************/

	class File
	{
		public:
			// Constructors/Destructor
			File();
			virtual ~File();

			//
			// Pure interface
			//

			// Create() Methods to initialise files are defined by the concrete classes...

			// Opening & Closing files
			virtual Error Open(FileMode Mode = fam_Full) = 0;
			virtual void Close() = 0;

			virtual bool IsOpen() const;

			// Determine the size of the file, in bytes
			virtual Error GetSize(uint* pSize);

			// Binary read/write  NB. does no endian conversions
			virtual Error Read(void* pData, uint cBytes, uint* pcBytes =  NULL) = 0;
			virtual Error Write(const void* pData, uint cBytes, uint* pcBytes = NULL) = 0;

			// Seeking
			virtual Error SeekFromStart(uint Offset) = 0;
			virtual Error SeekFromEnd(uint Offset) = 0;
			virtual Error SeekRelative(int Offset) = 0;

			// Get file read/write pointer position
			virtual Error GetPos(uint* pPos) const = 0;

			// Test for end of file.
			virtual bool eof() const = 0;

			// Read line of text
			virtual Error ReadLine(String* pData) = 0;

			//
			// Read/write specialisations (concrete functions)
			//

			/// Read/write of scalars
			Error Read_byte(byte* pData);
			Error Read_bool(bool* pData);

			Error Read_int8(int8* pData);
			Error Read_int16(int16* pData);
			Error Read_int32(int32* pData);
			Error Read_int64(int64* Data);

			Error Read_uint8(uint8* pData);
			Error Read_uint16(uint16* pData);
			Error Read_uint32(uint32* pData);
			Error Read_uint64(uint64* Data);

			Error Read_float32(float* Data);
			Error Read_float64(double* Data);

			Error Write_byte(byte Data);
			Error Write_bool(bool Data);

			Error Write_int8(int8 Data);
			Error Write_int16(int16 Data);
			Error Write_int32(int32 Data);
			Error Write_int64(int64 Data);

			Error Write_uint8(uint8 Data);
			Error Write_uint16(uint16 Data);
			Error Write_uint32(uint32 Data);
			Error Write_uint64(uint64 Data);

			Error Write_float32(float Data);
			Error Write_float64(double Data);

			// Text write
			Error WriteLine(const String& Data);
			Error WriteLine(const char* szData);
			Error WriteLineEnd();
			
			// Strings
			Error WriteString(const String& Data);
			Error ReadString(String * pData);

			//
			// For file copy/delete etc, see the FileSpec class.
			//
	};

};


#endif // !vos_File_h
