//
// FileInAFile.h - a class to support a virtual file within another file
//

#ifndef vos_FileInAFile_h
#define vos_FileInAFile_h

#include "File.h"

namespace vos
{
	
	/**************************************************************************

	>	class FileInAFile : public vos::File

		Author:		Tim Browse

		Date:		02/06/2001

		Purpose:	This class supports using a section of an existing vos::File
					object's data as a separate, independent vos::File object.

					(Obviously there are limits - if one piece of code is writing
					to the section being used as a sub-file, then you need
					to synchronise access to the file...as you might expect,
					there's no magic here).

					For example, given a 500k file on disk, you can set up
					a FileInAFile object to simulate a file located somewhere
					in this 500k file - for example, you might specify that
					starting at byte offset 23254 in the file, there is a 
					file that is 83235 bytes in length.  You can then open
					this file and access it using the standard vos::File
					interface (because FileInAFile is derived from vos::File).

					This is useful for image filing systems which store many
					files in one file.  The class allows you to create a veneer
					to an embedded file, and then pass this vos::File-derived
					object to any function that knows how to work with vos::File
					objects.

					The class vetos any attempts to access data outside the
					area marked by the Create() function, and returns an error
					code if this occurs.

					Usage
					-----

					To set up a FileInAFile, you simply create a instance
					of a vos::File-derived object, initialise it so that
					vos::File::Open() can be called on it, and pass it
					to the FileInAFile::Create() function.

					For example...
					
					\code

						// Create a diskfile object that uses the
						// file c:\myfile.dat.

						vos::DiskFile* pMainFile = new vos::DiskFile;
						vos::FileSpec FS("c:\\myfile.dat");
						pMainFile->Create(FS);

						// Create a FileInAFile that uses this file to
						// simulate a 500-byte file that starts 2000 bytes
						// into c:\myfile.dat.

						FileInAFile SubFile;
						SubFile.Create(pMainFile, 2000, 500);

						// Use as a normal file object...
						MyFileProcessingFn(&SubFile);

						// etc...

					\endcode

					NB. At present, write operations are not permitted on
						FileInAFile objects.

					IMPORTANT!

						Once you pass in the vos::File object to Create(),
						the FileInAFile object OWNS that object, and you should
						NOT use it anywhere else.  You should also NOT destroy
						it - the FileInAFile object will 'delete' it when it has
						finished with it.  
						It is therefore very important that you create all File 
						objects you pass to FileInAFile objects dynamically - if 
						they are auto variables, they will obviously be destroyed 
						when they go out of scope - this would be bad!

	**************************************************************************/

	class FileInAFile : public vos::File
	{
		public:
			FileInAFile();
			~FileInAFile();

			// FileInAFile creation/init (default to read-only access)
			void Create(vos::File* pParentFile, uint FileStart, int FileLength,
						vos::FileMode ModesPermitted = vos::fam_Read);

			// Opening & Closing files
			virtual vos::Error Open(vos::FileMode Mode = vos::fam_Read);
			virtual void Close();

			virtual bool IsOpen() const;

			// Determine the size of the file, in bytes
			virtual vos::Error GetSize(uint* pSize);

			// Binary read/write  NB. does no endian conversions
			virtual vos::Error Read(void* pData, uint cBytes, uint* pcBytes =  NULL);
			virtual vos::Error Write(const void* pData, uint cBytes, uint* pcBytes = NULL);

			// Seeking
			virtual vos::Error SeekFromStart(uint Offset);
			virtual vos::Error SeekFromEnd(uint Offset);
			virtual vos::Error SeekRelative(int Offset);

			// Get file read/write pointer position
			virtual vos::Error GetPos(uint* pPos) const;

			// Test for end of file.
			virtual bool eof() const;

			// Read line of text
			virtual vos::Error ReadLine(vos::String* pData);

		private:
			// Move past a line-ending in the file.
			vos::Error ConsumeLineEnding();

		private:
			// Parent file object
			vos::File* m_pParentFile;

			// Segment of parent file that we are using
			uint m_FileStart;
			uint m_FileLength;

			// Maximum permissions allowed for underlying file.
			// (Only fam_Read and fam_Write are used to limit access)
			vos::FileMode m_Modes;
	};
}

#endif // !vos_FileInAFile_h
