//
// FilePath.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "vOS.h"

#include "FilePath.h"
#include "File.h"

#ifdef VOS_MSWINDOWS
#include "Win32\Win32_Headers.h"
#else
#include "DiskFile.h"
#endif


#include "Errors.h"



//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/////////////////////////////////////////////////////////////////////////
	// FileSpecList class (internal helper class)


	/***************************************************************************
	
	>	FileSpecList::FileSpecList(const FileSpec &FileSpec)

		Author:		Jason
		Date:		6/9/99

		Inputs:		pFileSpec	- The FileSpec that this list header represents

		Purpose:	Constructor

	 ***************************************************************************/

	FileSpecList::FileSpecList(const FileSpec &FileSpec)
				 : m_FileSpec(FileSpec),
				   m_pNext(NULL)
	{
	}


	/***************************************************************************

	>	FileSpecList::~FileSpecList()

		Author:		Jason
		Date:		6/9/99

		Purpose:	Destructor
		Notes:		Recursively destroys the entire FileSpecList

	 ***************************************************************************/

	FileSpecList::~FileSpecList()
	{
		delete(m_pNext);
		m_pNext = NULL;
	}



	/***************************************************************************

	>	void FileSpecList::Append(const FileSpec &File)

		Author:		Jason
		Date:		6/9/99

		Purpose:	Appends the given FileSpec to this list

	 ***************************************************************************/

	void FileSpecList::Append(const FileSpec &File)
	{
		// Find the tail of the list
		FileSpecList *pTail = this;
		while (pTail->GetNext() != NULL)
			pTail = pTail->GetNext();

		// And append the item
		pTail->m_pNext = new FileSpecList(File);
	}





	/////////////////////////////////////////////////////////////////////////
	// FilePath class


	/***************************************************************************

	>	FilePath::FilePath(const FilePath &Source)

		Author:		Jason
		Date:		31/8/2000

		Inputs:		Source	- The object to copy

		Purpose:	Copy constructor

	 ***************************************************************************/

	FilePath::FilePath(const FilePath &Source)
	{
		// Initialise out folder list to be empty
		m_pPaths = NULL;

		// Copy the folder list from the Source FilePath
		FileSpecList *pFolder = Source.m_pPaths;
		while (pFolder != NULL)
		{
			AppendFolder(pFolder->GetFileSpec());
			pFolder = pFolder->GetNext();
		}
	}

	/***************************************************************************

	>	FilePath &FilePath::operator=(const FilePath &Source)

		Author:		Jason
		Date:		31/8/2000

		Inputs:		Source	- The object to copy

		Purpose:	Assignment operator

	 ***************************************************************************/

	FilePath &FilePath::operator=(const FilePath &Source)
	{
		// Clear our own folders
		delete(m_pPaths);
		m_pPaths = NULL;

		// Copy the folders in the other path
		FileSpecList *pFolder = Source.m_pPaths;
		while (pFolder != NULL)
		{
			AppendFolder(pFolder->GetFileSpec());
			pFolder = pFolder->GetNext();
		}

		return(*this);
	}



	/***************************************************************************

	>	void FilePath::AppendFolder(const String &FolderName)

		Author:		Jason
		Date:		6/9/99

		Inputs:		FolderName	- A Folder pathname
		Purpose:	Appends the given FileSpec to this FilePath. 

	 ***************************************************************************/

	void FilePath::AppendFolder(const String &FolderName)
	{
		FileSpec FS;
		FS.SetFromString(FolderName);

		AppendFolder(FS);
	}



	/***************************************************************************

	>	void FilePath::AppendFolder(const FileSpec &Folder)

		Author:		Jason
		Date:		6/9/99

		Inputs:		Folder		- A Folder FileSpec
		Purpose:	Appends the given FileSpec to this FilePath.

	 ***************************************************************************/

	void FilePath::AppendFolder(const FileSpec &Folder)
	{
		if (m_pPaths == NULL)
			m_pPaths = new FileSpecList(Folder);
		else
			m_pPaths->Append(Folder);
	}



	/***************************************************************************

	>	Error FilePath::FindFile(const FileSpec &fsFilename, |* to *| FileSpec *pFileSpec) const

		Author:		Jason
		Date:		6/9/99

		Inputs:		fsFilename  - name of the file to search for.
								  It can either be a leafname, or contain folders,
								  such as "MyFolder\MyFile.dat"

		Outputs:	pFileSpec	- If successful, will contain a full FileSpec for
								  the file.

		Returns:	VNOERROR if the file is found,
					else  an error code indicating the problem (Err_File_NotFound)

		Purpose:	Searches the file paths registered with this object for a file
					with the given leafname. If a match is found, the full FileSpec
					of the file is returned.

	 ***************************************************************************/

	Error FilePath::FindFile(const FileSpec &fsFilename, /* to */ FileSpec *pFileSpec) const
	{
		ASSERT(pFileSpec != NULL);
		if (pFileSpec != NULL)
		{
			FileSpecList *pFolder = m_pPaths;

			while (pFolder != NULL)
			{
				FileSpec FullSpec(pFolder->GetFileSpec());
				FullSpec.AppendRelativeFileSpec(fsFilename);

				//---
				#ifdef VOS_MSWINDOWS
					WIN32_FIND_DATA FindData;

					HANDLE hFile = ::FindFirstFile(FullSpec.GetString().c_str(), &FindData);
					if (hFile != INVALID_HANDLE_VALUE)
					{
						// We have found a matching file - return the FileSpec
						*pFileSpec = FullSpec;

						::FindClose(hFile);
						hFile = NULL;

						return(VNOERROR);
					}

				#elif (defined VOS_MACOS)
					// Generic code that should do the trick...
					DiskFile Test;
					if (Test.Open(FullSpec, fam_Read))
					{
						// We have found a matching file - return the FileSpec
						*pFileSpec = FullSpec;

						Test.Close();

						return(VNOERROR);
					}
					else
					{
						// Clear the error
						ERROR_HANDLED();
					}
				#else
					#error "FilePath::Findfile requires porting (or at least testing)"

					// Generic code that should do the trick...
					DiskFile Test;
					if (Test.Open(FullSpec, fam_Read))
					{
						// We have found a matching file - return the FileSpec
						*pFileSpec = FullSpec;

						Test.Close();

						return(VNOERROR);
					}

				#endif
				//---

				pFolder = pFolder->GetNext();

			}
		}

		String ErrMsg = "FilePath::FindFile(\"";
		ErrMsg += fsFilename.GetString();
		ErrMsg += "\") failed (File not found)";

		return ERROR(Err_File_NotFound, ErrMsg.c_str());
	}

};

