//
// FileSpecCommon.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "FileSpec.h"

#include "DiskFile.h"

#include "FileMatches.h"
#include "Errors.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif


namespace vos
{

	/////////////////////////////////////////////////////////////////////////
	// Statics

	//
	// The default character to use as the folder separator in strings
	// (See SetFromString() and GetString() functions).
	//
	char FileSpec::kStringFolderSeparator = kFolderSeparator[0];



	/**************************************************************************

		String FileSpec::GetString(char FolderSeparator = kStringFolderSeparator) const

		Author:		Tim Browse
		Date:		20/08/99
		Inputs:		FolderSeparator - the character to be used in the string to
					denote separation of folders/pathname components.

		Returns:	String representing the FileSpec object.

		Purpose:	Convert a FileSpec object into a String object.

	**************************************************************************/

	String FileSpec::GetString(char FolderSeparator) const
	{
		// Copy string
		String S = m_FileName;

		// Convert folder separators
		if (FolderSeparator != kStringFolderSeparator)
		{
			S.Replace(kStringFolderSeparator, FolderSeparator);
		}

		return S;
	}

	/**************************************************************************

		bool FileSpec::operator==(const FileSpec& rhs) const

		Author:		Tim Browse
		Date:		20/08/99

		Purpose:	Standard equality test

	**************************************************************************/

	bool FileSpec::operator==(const FileSpec& rhs) const
	{
#ifdef VOS_UNIX
		// Unix filenames are case-sensitive
		return (m_FileName == rhs.m_FileName);
#else
		// Almost all others are not, so we ignore case when comparing
		return m_FileName.CompareText(rhs.m_FileName);
#endif
	}

	/**************************************************************************

		bool FileSpec::operator!=(const FileSpec& rhs) const

		Author:		Tim Browse
		Date:		20/08/99

		Purpose:	Standard non-equality test

	**************************************************************************/

	bool FileSpec::operator!=(const FileSpec& rhs) const
	{
#ifdef VOS_UNIX
		// Unix filenames are case-sensitive
		return (m_FileName != rhs.m_FileName);
#else
		// Almost all others are not, so we ignore case when comparing
		return !m_FileName.CompareText(rhs.m_FileName);
#endif
	}




	/**************************************************************************

		bool FileSpec::GetRelativeFileSpec(const FileSpec& Prefix, 
										   FileSpec* pRelativeFileSpec) const

		Author:		Tim Browse
		Date:		20/08/99
		Inputs:		Prefix - the path prefix to remove
		Outputs:	pRelativeFileSpec - the resulting relative FileSpec
		Returns:	true => successful
					false => Prefix did not match the initial segment of
							 this FileSpec.  pRelativeFileSpec is not
							 updated.

		Purpose:	Strip the specified prefix from a FileSpec to obtain a 
					relative FileSpec.

					e.g.

						FileSpec Prefix;
						Prefix.SetFromString("c:\apps");

						FileSpec FullPath;
						FullPath.SetFromString("c:\apps\excel\myfile.xls");

						FileSpec Relative;
						FullPath.GetRelativeFileSpec(Prefix, &Relative);

						String S = Relative.GetString();

						// S now equals "excel\myfile.xls"

	**************************************************************************/

	bool FileSpec::GetRelativeFileSpec(const FileSpec& Prefix, 
									   FileSpec* pRelativeFileSpec) const
	{
		// Cope with simple case first (null prefix)
		String PrefixString = Prefix.GetString();
		uint PrefixLength = PrefixString.GetLength();

		if (PrefixLength == 0)
		{
			*pRelativeFileSpec = *this;
			return true;
		}

		// OK: first, confirm that the prefix is a partial match
		if (PrefixString.CompareText(m_FileName.Left(PrefixLength)))
		{
			// it's a match - increase the length if we need to skip the preceding folder
			// separator
			if (m_FileName.GetAt(PrefixLength) == kStringFolderSeparator)
			{
				PrefixLength++;
			}

			// And copy the trailing pathname
			String Remainder = m_FileName.Right(m_FileName.GetLength() - PrefixLength);
			pRelativeFileSpec->SetFromString(Remainder);

			return true;
		}
		else
		{
			// not a match - complain
			COMPLAIN("GetRelativeFileSpec(): prefix failed to match");
			return false;
		}
	}


	/**************************************************************************

		String FileSpec::GetLeafname() const

		Author:		Tim Browse
		Date:		20/08/99
		Returns:	The leafname of the FileSpec.

		Purpose:	Obtain the leafname of this FileSpec.

	**************************************************************************/

	String FileSpec::GetLeafname() const
	{
		// Look for the last folder separation character in the name
		uint iStart;
		if (m_FileName.ReverseFind(kStringFolderSeparator, &iStart))
		{
			// Return the characters following the last backslash
			return m_FileName.Right(m_FileName.GetLength() - (iStart + 1));
		}
		else
		{
			// No folder separators - just return whole filename
			return m_FileName;
		}
	}

	/**************************************************************************

		Error FileSpec::CreateFolderTree(CreateFolderHandler *pHandler = NULL) const

		Author:		Tim Browse
		Date:		12/05/00
		Returns:	Error

		Purpose:	Create a folder with the name specified by this object.

					If the parent folder of the specified folder does not 
					exist, then that folder	is created too, and so on.

	**************************************************************************/

	Error FileSpec::CreateFolderTree(CreateFolderHandler *pHandler) const
	{
		// Get parent folder, to see if it exists
		FileSpec Parent;
		if (GetParent(&Parent))
		{
			// Does the parent folder exist?
			switch (Parent.GetFileType())
			{
				case FileType_Folder:
				{
					// Parent folder exists - no action required
					break;
				}

				case FileType_NotFound:
				{
					// Parent folder does not exist - try to create it
					if (!Parent.CreateFolderTree(pHandler))
					{
						// Error occured - pass on to caller
						return LASTERROR;
					}
					else
					{
						// Parent folder created ok - no further action needed.
					}
					break;
				}

				case FileType_File:
				{
					// Parent folder 'exists' but it's a file object, and so we can't
					// change it.
					String ErrMsg = "Unable to create the folder \"";
					ErrMsg += Parent.GetString();
					ErrMsg += "\" because it already exists as a file.";

					return ERROR(Err_File_AccessDenied, ErrMsg.c_str());
				}
			}
		}

		// If we get here, then the parent folder exists (or the parent folder cannot
		// exist because it is a device root), so we try to create the original folder 
		// we were asked to create.
		return CreateFolder(pHandler);
	}

	/**************************************************************************

		Error FileSpec::Copy(const FileSpec& DestinationFile,
							 FileCreation CreationType = FileCreation_Normal,
							 CopyFileHandler* pCopyFileHandler = NULL) const

		Author:		Tim Browse
		Date:		20/08/99
		Inputs:		DestinationFile - the file to copy to.

					CreationType - controls how to cope with the parent folder
								   of the destination file not existing.
								   If set to ForceParentCreation, then the
								   parent folder is created to contain the
								   destination file if it doesn't already 
								   exist (this copes with arbitrarily nested 
								   parent folders).

					pCopyFileHandler - the handler to monitor the copy
									   operation, or NULL if no handler needed.
		Returns:	Error

		Purpose:	Copy the file specified by this object to another file.

					If the destination file already exists, it will be 
					overwritten.  If DestinationFile describes a folder,
					the function will fail.

					A progress monitor can be used by passing in an object
					of class CopyFileHandler, which will be periodically
					polled during the copy.  The handler will be told how many
					bytes have been copied so far, and can optionally terminate 
					the copy operation prematurely, in which case the Copy() 
					function will return the Error code that the handler 
					returns.

	**************************************************************************/

	Error FileSpec::Copy(const FileSpec& DestinationFile,
						 FileCreation CreationType,
						 CopyFileHandler* pCopyFileHandler) const
	{
		// If the destination file exists, delete it.
		switch (DestinationFile.GetFileType())
		{
			case FileType_File:
			{
				// Try to delete this file
				if (!DestinationFile.Delete())
				{
					// Unable to delete destination file - can't continue
					return LASTERROR;
				}
				break;

			}

			case FileType_Folder:
			{
				// Can't copy the file to a folder object - doesn't make sense.
				String ErrMsg = "Unable to copy to \"";
				ErrMsg += DestinationFile.GetString();
				ErrMsg += "\" because it is a folder.";

				return ERROR(Err_File_AccessDenied, ErrMsg.c_str());
			}

			case FileType_NotFound:
			{
				// Ok - do nothing
				break;
			}

			default:
			{
				// Unknown file type
				COMPLAIN("FileSpec::Copy() - logic error.");
				return ERROR(Err_GenericError, "FileSpec::Copy() - internal logic error.");
			}
		}

		// Try to open source file
		DiskFile SourceDiskFile;
		if (!SourceDiskFile.Open(*this, fam_Read))
		{
			// Unable to open source file - return error.
			return LASTERROR;
		}

		if (CreationType == FileCreation_ForceParentCreation)
		{
			// Make sure parent folder of the file exists...
			FileSpec ParentFolder;
			
			if (DestinationFile.GetParent(&ParentFolder))
			{
				if (ParentFolder.GetFileType() != FileType_Folder)
				{
					// Parent folder does not exist - try to create it
					if (!ParentFolder.CreateFolderTree(pCopyFileHandler))
					{
						// Unable to create parent folder - return error back.
						return LASTERROR;
					}
					else
					{
						// Created folder successfully 
						//
						// => no further action.
					}
				}
			}
			else
			{
				// Could not find parent folder - so fall through and try to create file,
				// as the parent might be a server share or drive root or similar.
				//
				// => no further action.
			}
		}

		// Try to open destination file for writing
		DiskFile DestinationDiskFile;
		if (!DestinationDiskFile.Open(DestinationFile, fam_Write))
		{
			// Clean up source file object.
			SourceDiskFile.Close();

			// Unable to open destination file - return error.
			return LASTERROR;
		}

		// Ok, try to copy in 64k chunks
		const int kChunkSize = 64 * 1024;

		// Allocate buffer to copy file with
		byte *pData = new byte[kChunkSize];
		uint cbBytesRead;

		// Initialise error values to signify no errors yet.
		int ReadError = Err_NoError;
		Error ReadResult = SUCCESS;

		// Running total of how much we have copied
		uint cbBytesCopied = 0;

		// keep copying chunks until we run out.
		for(;;)
		{
			// Read a chunk in
			ReadResult = SourceDiskFile.Read(pData, kChunkSize, &cbBytesRead);

			if (ReadResult != SUCCESS)
			{
				// Error occured - was it just EOF?
				ReadError = GetLastErrorCode();

				if (ReadError != Err_File_EOF)
				{
					// No, so shutdown the operation and return the error.
					goto error;
				}
				else
				{
					// We can cope with EOF errors
					ERROR_HANDLED();
				}
			}

			// Chunk read ok - write it to the destination file.
			if (!DestinationDiskFile.Write(pData, cbBytesRead))
			{
				// Error when writing - clean up and return error
				goto error;
			}

			cbBytesCopied += cbBytesRead;

			// If that was the last chunk, then we're done.
			if ((ReadResult != SUCCESS) && (ReadError == Err_File_EOF))
			{
				break;
			}

			// Otherwise, call handler object...
			if (pCopyFileHandler != NULL)
			{
				if (!pCopyFileHandler->CopyFileCallback(cbBytesCopied))
				{
					goto error;
				}
			}

			// ...and then do the next chunk
		}

		// If we've got here, it all worked.
		delete [] pData;
		SourceDiskFile.Close();
		DestinationDiskFile.Close();

		return VNOERROR;

error:
		// Error handling
		delete [] pData;
		SourceDiskFile.Close();
		DestinationDiskFile.Close();

		// Remove destination file (taking care to preserve error
		// status so we can return it to the caller).
		TheErrorManager().PushState();
			if (!DestinationFile.Delete())
			{
				ERROR_HANDLED();
			}
		TheErrorManager().PopState();

		return LASTERROR;

	}
	
};	// namespace vos

