//
// Preferences.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_Preferences_h
#define vos_Preferences_h

#include "vos.h"

#include "TreeNode.h"
#include "String.h"

namespace vos
{
	class PreferencesImp;

	/**************************************************************************
	
	>	class Preferences : public TreeNode
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Provide a way of storing/retrieving user preferences for
					the application.

					You must call the static Init() function to set up
					the preferences system before using any of the features.

					The preference system allows storage of the following
					types as preferences:

						- String
						- int
						- bool
						- FileSpec

					Preferences are conceptually arranged as a nested tree of
					named sections, in which you can store values.

					The preferences will be stored in whatever way is appropriate
					to the host platform.

					You can use the global function GetPreferences() for easy
					access to the root Preferences object.

					The best way to provide default values for Preferences is
					to initialise the value before reading it, e.g.

					\code

						vos::String Username("noname");
						vos::GetPreferences().Read("username", &Username);

					\endcode

					In this way, if the preference is not found, the Username
					string will be left unchanged, and will have the default
					value of 'noname' in it.  If the preference does exist,
					Username will hold the value of the preference.

	**************************************************************************/
	
	class Preferences : public TreeNode
	{
		public:
			// Singleton object
			static Preferences& GetPreferences();

			// Set up preference details
			static void Init(const String& AuthorName,
							 const String& ProgramName);

			// Remove all preferences for this application
			static void DeleteAll();

			// Clean up preferences on shutdown
			// (Not strictly necessary, but will make memory leak checkers happy)
			void CleanUp();

			~Preferences();

			// Find out if the preferences system has been initialised by a call
			// to Init();
			static bool IsInitialised();

		public:
			// Get a sub-section
			// (You do not need to delete this object when you are done with it)
			Preferences* FindSection(const String& SectionName);

			// Read/write of preferences
			bool Read(const String& BooleanPrefName, bool bDefaultValue);

			bool Read(const String& PrefName, String*	pPrefValue);
			bool Read(const String& PrefName, int*		pPrefValue);
			bool Read(const String& PrefName, bool*		pPrefValue);
			bool Read(const String& PrefName, FileSpec* pPrefValue);

			bool Write(const String& PrefName, const String& PrefValue);
			bool Write(const String& PrefName, const char *szPrefValue);
			bool Write(const String& PrefName, int PrefValue);
			bool Write(const String& PrefName, bool PrefValue);
			bool Write(const String& PrefName, const FileSpec& PrefValue);

			// Read/write of preference lists
			int ReadListCount(const String& ListPrefName);
			String ReadListItem(const String& ListPrefName, int iItem);

			bool WriteListCount(const String& ListPrefName, int cItems);
			bool WriteListItem(const String& ListPrefName, 
							   int iItem, const String& strItem);

			// Access the preference strings
			static String GetAuthorName() { return s_AuthorName; }
			static String GetProgramName() { return s_ProgramName; }
			String GetSectionName() { return m_SectionName; }

			// Not for general use.
			PreferencesImp* GetImp() { return m_pImp; }

		private:
			// Prohibit construction except for members of this class.
			Preferences(const String& SectionName,
						Preferences* pParent);

			// Prohibit copying
			Preferences(const Preferences& rhs);
			Preferences& operator=(const Preferences& rhs);

			// Pathname manipulation
			bool SplitPreferencePathname(const String& Pathname,
										 String* pHead,
										 String* pTail);

			bool SplitPathnameAtLeaf(const String& PrefName, 
									 String* pParentName,
									 String* pLeafname);

			Preferences* FindParentSection(const String& PrefName,
										   String* pLeafname = NULL);

		private:
			// Implementation.
			PreferencesImp* m_pImp;

			// Name of this preference section
			String m_SectionName;

		private:
			// Details of the preferences.
			static String s_AuthorName;
			static String s_ProgramName;
			static bool s_bInitialised;
	};

	extern Preferences& GetPreferences();
};

#endif // !vos_Preferences_h
