//
// String.cpp - platform independent String class.
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "String.h"

#include "StringParser.h"
#include "StringSupport.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	// This is the empty string that we refer when people want access to empty strings
	const char* String::m_EmptyString = "";


	/******************************************************************************

	>	String::String()

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Constructor

	******************************************************************************/

	String::String() : m_pData(NULL),
					   m_BufferLength(0),
					   m_StringLength(0)
	{
	}


	/******************************************************************************

	>	String::String(const String& Src)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Construct a new string from the given string.
		Inputs:		Src - String to copy

	******************************************************************************/

	String::String(const String& Src)
	{
		// Copy this string
		m_StringLength = Src.GetLength();

		if (m_StringLength > 0)
		{
			m_BufferLength = m_StringLength + 1;
			m_pData = new tchar[m_BufferLength];
			vos_strcpy(m_pData, Src.m_pData);
		}
		else
		{
			m_BufferLength = 0;
			m_pData = NULL;
		}
	}

	/******************************************************************************

	>	String::String(tchar Ch, uint nRepeat = 1)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Construct a new string, consisting of one or more of the
					same character.
		Inputs:		Ch - the character to repeat
					nRepeat - how many times to repeat it.

	******************************************************************************/

	String::String(tchar Ch, uint nRepeat)
	{
		// Make repetitions of this character.
		m_StringLength = nRepeat;
		m_BufferLength = m_StringLength + 1;
		m_pData = new tchar[m_BufferLength];

		for (int i = 0; i < m_StringLength; i++)
		{
			m_pData[i] = Ch;
		}

		// Terminate the string
		m_pData[m_StringLength] = EOS;
	}

	/******************************************************************************

	>	String::String(int n)
							 ,
		Author:		Sean Solle
		Created:	10/4/97
		Purpose:	Construct an ascii decimal version of the integer passed.
		Inputs:		An integer

	******************************************************************************/

	String::String(int n)
	{
		// Somewhere to store the string
		m_BufferLength = 32;
		m_pData = new tchar[m_BufferLength];

		// Convert the number to ascii decimal.
		vos_sprintf((tchar*) m_pData, _T("%i"), n);
		
		// Use this string
		m_StringLength = vos_strlen((tchar*)m_pData);
	}

	/******************************************************************************

	>	String::String(uint n)
							 ,
		Author:		Sean Solle
		Created:	10/4/97
		Purpose:	Construct an ascii decimal version of the integer passed.
		Inputs:		An integer

	******************************************************************************/

	String::String(uint n)
	{
		// Somewhere to store the string
		m_BufferLength = 32;
		m_pData = new tchar[m_BufferLength];

		// Convert the number to ascii decimal.
		vos_sprintf((tchar*) m_pData, _T("%i"), n);
		
		// Use this string
		m_StringLength = vos_strlen((tchar*)m_pData);
	}

	/******************************************************************************

	>	String::String(const tchar* pString)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Construct a new string from the given string.
		Inputs:		pString - null-terminated C-style string to copy.

	******************************************************************************/

	String::String(const tchar* pString)
	{
		// Copy this string
		m_StringLength = vos_strlen(pString);
		m_BufferLength = m_StringLength + 1;
		m_pData = new tchar[m_BufferLength];
		vos_strcpy(m_pData, pString);
	}


	/******************************************************************************

	>	String::String(const tchar* pString, uint nChars)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Construct a new string from the given string.
		Inputs:		pString - null-terminated C-style string to copy.
					nChars - the number of characters to copy from pString

	******************************************************************************/

	String::String(const tchar* pString, uint nChars) : m_pData(NULL),
													    m_BufferLength(0),
													    m_StringLength(0)
	{
		// Copy this string
		CopyChars(pString, 0, nChars);
	}


	/******************************************************************************

	>	String::String(const tchar* pString, uint nChars)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Construct a new string from a Pascal string
		Inputs:		pPStr - length-prefixed Pascal-style string to copy.

	******************************************************************************/
	

	String::String(uint8* pPStr)
	{
		// Create the buffer for the string
		// Byte 0 of the Pascal string is the length of the Pascal Data
		m_StringLength = pPStr[0];
		m_BufferLength = m_StringLength + 1;
		m_pData = new tchar[m_BufferLength];
		
		// Copy the Pascal string over
		for(int i = 0; i < m_StringLength; i++)
		{
			m_pData[i] = pPStr[i+1];
		}
		
		// Null terminate it
		m_pData[m_StringLength] = EOS;
	}


	/******************************************************************************

	>	void String::assign(const tchar *szString, uint nChars)

		Author:		Tim Browse
		Created:	2/9/00
		Purpose:	Assign a string from the first part of a C-string.
		Inputs:		szString - null-terminated C-style string to copy.
					nChars - the number of characters to copy from szString

	******************************************************************************/

	void String::assign(const tchar *szString, uint nChars)
	{
		CopyChars(szString, 0, nChars);
	}


	/******************************************************************************

	>	String::~String()

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Destructor

	******************************************************************************/

	String::~String()
	{
		delete [] m_pData;
	}


	/**************************************************************************
	
	>	bool String::Load(const StringResourceID& ID)
	
		Author:		Tim Browse
		Date:		28/8/2002
		Inputs:		ID - the ID of the string resource to load
		Returns:	true => string loaded successfully
					false => otherwise
	
		Purpose:	Load a string from a string resource.
	
	**************************************************************************/
	
	bool String::Load(const StringResourceID& ID)
	{
		return ID.Load(this);
	}
	
	/******************************************************************************

	>	void String::Clear()

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Set the string to be empty (i.e. it will be 0 characters long).

	******************************************************************************/

	void String::Clear()
	{
		if (m_pData != NULL)
		{
			m_pData[0] = 0;
		}
		m_StringLength = 0;
	}





	/******************************************************************************

	>	void String::SetAt(uint nIndex, tchar Ch)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Set the specified character to the specified value.
		Inputs:		nIndex - which character to set (0 based)
					Ch - the character value to set it to.

		Notes:		If you attempt to write past the end of the available buffer
					this call will currently be ignored.

					If you write beyond the current end of the string, the behaviour
					is undefined. In most cases this will result in a confused string
					(where the cached string length does not match the terminating
					character position)

	******************************************************************************/

	void String::SetAt(uint nIndex, tchar Ch)
	{
		if (nIndex < m_BufferLength)
		{
			// Set the character
			m_pData[nIndex] = Ch;

			// Was it a terminator character?
			if (Ch == 0)
			{
				// Yes, so update string length
				m_StringLength = nIndex;
			}
		}
	}


	/******************************************************************************

	>	String::operator const tchar *() const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Cast a string object to a (read-only/const) C-style string.

	******************************************************************************/

	String::operator const tchar *() const
	{
		return c_str();
	}


	/******************************************************************************

	>	const tchar* String::c_str() const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Cast a string object to a (read-only/const) C-style string.

	******************************************************************************/

	const tchar* String::c_str() const
	{
		// If the string is empty, then use the Empty null terminated string we 
		// happen to have
		if (m_pData==NULL)
			return m_EmptyString;

		// otherwise we can use the real string
		return (const tchar *) m_pData;
	}



	/******************************************************************************

	>	void String::ReserveCapacity(uint32 MinSizeChars)

		Author:		Tim
		Created:	14/4/2002

		Inputs:		MinSizeChars	- Number of *characters* of buffer-space required,
									  (not including the terminating NUL)

		Purpose:	Ensure the specified number of characters is available in
					the string buffer.

	******************************************************************************/

	void String::ReserveCapacity(uint32 MinSizeChars)
	{
		if (m_pData == NULL || m_BufferLength <= MinSizeChars)
		{
			// Allocate a new buffer that is large enough, and move our string into it
			tchar *pNewBuffer = new tchar[MinSizeChars+1];
			if (m_pData == NULL)
			{
				pNewBuffer[0]	= 0;
				m_StringLength	= 0;
			}
			else
			{
				vos_strcpy(pNewBuffer, m_pData);
				delete [] m_pData;
			}

			m_pData = pNewBuffer;
			m_BufferLength = MinSizeChars+1;
		}
	}

	/******************************************************************************

	>	tchar *String::LockBuffer(uint32 MinSizeChars)

		Author:		Jason
		Created:	6/11/2000

		Inputs:		MinSizeChars	- Number of *characters* of buffer-space required,
									  (not including the terminating NUL)

		Returns:	A pointer to the buffer, or NULL if it fails

		Purpose:	Extremely low-level string manipulation.
		
					Retrieves a pointer to a tchar array of not less than MinSizeChars
					*characters* (not bytes!) in length. This allows you to ensure
					that a certain amount of buffer space is available to work in.

					When you are finished with the Buffer, call UnlockBuffer().
					This allows the String to recalculate it s string length.

		Notes
					This implementation does not retain a lock flag/count and relies
					entirely on the user using (and Unlocking) it properly.

					Also note that if you change the length of the string, the
					internal cache of the length will not be updated until
					you call UnlockBuffer(), and hence some String functions will
					not work correctly on such a String. i.e. Only keep the buffer
					locked for a minimal time, and avoid doing any other string
					operations while it's locked.

					Also note that the buffer is an array of tchars - it is entirely
					up to the caller to determine what a tchar is (ASCII/Unicode)
					and take appropriate action.

	******************************************************************************/

	tchar *String::LockBuffer(uint32 MinSizeChars)
	{
		ReserveCapacity(MinSizeChars);
		return(m_pData);
	}



	/******************************************************************************

	>	void String::UnlockBuffer(void)

		Author:		Jason
		Created:	6/11/2000

		Purpose:	Releases the lock on a string (see LockBuffer() above).

					This allows the String to recalculate its string length and
					(optionally) garbage-collect any unused portion of the string
					buffer memory.

		Notes:		The string MUST be valid (NUL terminated) when you call this
					method or it may run off through memory looking for a NUL.

	******************************************************************************/

	void String::UnlockBuffer(void)
	{
		// Re-calculate our string length fom the new contents of the buffer
		m_StringLength = vos_strlen(m_pData);

		// TODO_NICE: We could return unneeded memory to the heap here...
	}


	
	/******************************************************************************

	>	const uint8* String::p_str( uint8 ) const

		Author:		Rob Myers
		Created:	14/7/00
		Inputs:		pPstr	-	A Mac Str255 pascal string ( uint8[256] ). 
								The C string is copied into this.
		Returns:	pPstr.
		Purpose:	Copy the string into a pascal-style string. We copy to avoid 
					allocating space for 256 bytes that may never be used, 
					allocating new space the user must dispose of, or mangling 
					cross-platform code to allow an extra byte before the C string.

	******************************************************************************/

	const uint8* String::p_str(uint8* pPStr) const
	{
#ifdef VOS_UNICODE
#error String::p_str needs porting to Unicode
#else
		// If the string is empty, then use the Empty null terminated string we 
		// happen to have
		if ((m_pData==NULL)||(m_StringLength==0))
		{
			*pPStr = 0;
		}
		else
		{
			// Set the length prefix byte
			int length = m_StringLength > 255 ? 255 : m_StringLength ;
			pPStr[ 0 ] = length;

			// Copy up to the first 255 bytes of the c string
			for ( int i = 0; i < length; i++)
			{
				pPStr[ i + 1 ] = m_pData[ i ];
			}
		}
		return pPStr;
#endif	
	}


	/******************************************************************************

	>	void String::Copy(const tchar *pSrc)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Generic copy function.
		Inputs:		pSrc - string to copy

	******************************************************************************/

	void String::Copy(const tchar *pSrc)
	{
		// Look for the easy way out
		if (pSrc == NULL)
		{
			// We need to copy an empty string
			Empty();
			return;
		}

		// Do we have enough space to hold this string?
		int SrcLength = vos_strlen(pSrc);
		if (m_BufferLength <= SrcLength)
		{
			// No - discard our buffer and get one big enough.
			m_BufferLength = SrcLength + 1;
			delete [] m_pData;
			m_pData = new tchar[m_BufferLength];
		}

		// Copy the string, as we now have a big enough buffer
		ASSERT(m_pData != NULL);
		vos_strcpy(m_pData, pSrc);
		m_StringLength = SrcLength;
	}

	/******************************************************************************

	>	const String& String::operator=(const String& Src)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Assignment operator.
		Inputs:		Src - string to copy
		Returns:	Reference to this string.

	******************************************************************************/

	const String& String::operator=(const String& Src)
	{
		Copy(Src.m_pData);
		return *this;
	}




	/******************************************************************************

	>	const String& String::operator=(tchar *pString)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Assignment operator.
		Inputs:		pString - string to copy.
		Returns:	Reference to this string.

	******************************************************************************/

	const String& String::operator=(tchar *pString)
	{
		Copy(pString);
		return *this;
	}




	/******************************************************************************

	>	const String& String::operator=(tchar Ch)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Assignment operator.  
					Makes the string one character long.
		Inputs:		Ch - the character to copy.
		Returns:	Reference to this string.

	******************************************************************************/

	const String& String::operator=(tchar Ch)
	{
		tchar Temp[] = _T("?");
		Temp[0] = Ch;
		Copy(Temp);
		return *this;
	}




	/******************************************************************************

	>	const String& String::operator=(const tchar *pString)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Assignment operator.
		Inputs:		pString - string to copy.
		Returns:	Reference to this string.

	******************************************************************************/

	const String& String::operator=(const tchar *pString)
	{
		Copy(pString);
		return *this;
	}


	/******************************************************************************

	>	bool String::operator==(const String& Other) const

		Author:		Tim Browse
		Created:	12/2/97
		Purpose:	Equality operator.
		Inputs:		Other - string to check against.
		Returns:	non-zero if equal;
					0 if not.

	******************************************************************************/

	bool String::operator==(const String& Other) const
	{
		// Avoid accessing m_pData pointer of either string if it could be NULL.
		if (Other.m_StringLength == 0)
		{
			// They're equal if they're both zero length.
			return (m_StringLength == 0);
		}
		else if (m_StringLength == 0)
		{
			// This one's zero length, but the other one isn't => not equal
			return false;
		}

		// Use strcmp to compare
		return (vos_strcmp(m_pData, Other.m_pData) == 0);
	}

	/******************************************************************************

	>	bool String::operator==(const tchar *pString) const

		Author:		Tim Browse
		Created:	12/2/97
		Purpose:	Equality operator.
		Inputs:		pString - string to check against.
		Returns:	non-zero if equal;
					0 if not.

	******************************************************************************/

	bool String::operator==(const tchar *pString) const
	{
		// Avoid accessing m_pData pointer of this string if it could be NULL.
		if (m_StringLength == 0)
		{
			// They're equal if they're both zero length.
			return (vos_strlen(pString) == 0);
		}

		// Use strcmp to compare
		return (vos_strcmp(m_pData, pString) == 0);
	}

	/******************************************************************************

	>	bool String::operator!=(const String& Other) const

		Author:		Tim Browse
		Created:	12/2/97
		Purpose:	Inequality operator.
		Inputs:		Other - string to check against.
		Returns:	non-zero if unequal;
					0 if not.

	******************************************************************************/

	bool String::operator!=(const String& Other) const
	{
		// Avoid accessing m_pData pointer of either string if it could be NULL.
		if (Other.m_StringLength == 0)
		{
			// They're equal if they're both zero length.
			return (m_StringLength != 0);
		}
		else if (m_StringLength == 0)
		{
			// This one's zero length, but the other one isn't => not equal
			return true;
		}

		// Use strcmp to compare
		return (vos_strcmp(m_pData, Other.m_pData) != 0);
	}

	/******************************************************************************

	>	bool String::operator!=(const tchar *pString) const

		Author:		Tim Browse
		Created:	12/2/97
		Purpose:	Inequality operator.
		Inputs:		pString - string to check against.
		Returns:	non-zero if unequal;
					0 if not.

	******************************************************************************/

	bool String::operator!=(const tchar *pString) const
	{
		// Avoid accessing m_pData pointer of this string if it could be NULL.
		if (m_StringLength == 0)
		{
			// They're equal if they're both zero length.
			return (strlen(pString) != 0);
		}

		// Use strcmp to compare
		return (vos_strcmp(m_pData, pString) != 0);
	}

	
	/**************************************************************************

	>	bool String::operator<(const String& Other) const

		Author:		Tim Browse
		Date:		28/04/2001
		Inputs:		Other - string to check against.
		Returns:	true if this object is 'less than' Other;
					false otherwise

		Purpose:	Relative comparison operator.

					Allows String objects to be used in STL maps/sets etc.

	**************************************************************************/

	bool String::operator<(const String& Other) const
	{
		// Avoid accessing m_pData pointer of either string if it could be NULL.
		if (Other.m_StringLength == 0)
		{
			// If both are zero length, then they're equal
			return (m_StringLength == 0);
		}
		else if (m_StringLength == 0)
		{
			// This one's zero length, but the other one isn't => less_than
			return true;
		}

		// Use strcmp to compare
		return (vos_strcmp(m_pData, Other.m_pData) < 0);
	}


	/******************************************************************************

	>	bool String::CompareText(const String &Other) const

		Author:		Tim Browse
		Created:	24/6/97
		Purpose:	Case-insensitive equality operator.
		Inputs:		Other - string to check against.
		Returns:	TRUE if equal;
					FALSE if not.

	******************************************************************************/

	bool String::CompareText(const String &Other) const
	{
		// Avoid accessing m_pData pointer of either string if it could be NULL.
		if (Other.m_StringLength == 0)
		{
			// They're equal if they're both zero length.
			return (m_StringLength == 0);
		}
		else if (m_StringLength == 0)
		{
			// This one's zero length, but the other one isn't => not equal
			return false;
		}

		// Not all platforms support stricmp
		bool AreTheSame = true;
		if( Other.m_StringLength != m_StringLength )
		{
			AreTheSame = false;
		}
		else
		{
			for(int i = 0; i < m_StringLength; i++)
			{
				if(vos_tolower(m_pData[ i ]) != vos_tolower(Other.m_pData[ i ]))
				{
					AreTheSame = false;
					break;
				}
			}
		}
		return (AreTheSame);
	}

	/******************************************************************************

	>	bool String::CompareText(const tchar *pOther) const

		Author:		Tim Browse
		Created:	12/2/97
		Purpose:	Case-insensitive equality operator.
		Inputs:		pOther - string to check against.
		Returns:	TRUE if equal;
					FALSE if not.

	******************************************************************************/

	bool String::CompareText(const tchar *pOther) const
	{
		// Avoid accessing m_pData pointer of this string if it could be NULL.
		if (m_StringLength == 0)
		{
			// They're equal if they're both zero length.
			return (vos_strlen(pOther) == 0);
		}

		// Not all platforms support stricmp
		bool AreTheSame = true;
		if( vos_strlen(pOther) != m_StringLength )
		{
			AreTheSame = false;
		}
		else
		{
			for(int i = 0; i < m_StringLength; i++)
			{
				if(vos_tolower(m_pData[ i ]) != vos_tolower(pOther[ i ]))
				{
					AreTheSame = false;
					break;
				}
			}
		}
		return (AreTheSame);
	}

	/******************************************************************************

	>	bool String::StartsWith(const String &Prefix) const

		Author:		Tim Browse
		Created:	9/10/99
		Purpose:	Check to see if the string begins with the specified prefix.
					The function is case-insensitive.
		Inputs:		Prefix - the prefix to check for
		Returns:	true => this string starts with Prefix
					false => otherwise

	******************************************************************************/

	bool String::StartsWith(const String &Prefix) const
	{
		if (m_pData == NULL)
			return false;
			
		// Make sure we're long enough to compare
		const char * PrefixString = Prefix.c_str();
		uint StringLength = Prefix.GetLength();

		if (m_StringLength < StringLength)
		{
			return false;
		}
		
		// Compare
		
		for (int i = 0; i < StringLength; i++)
		{
			if ( vos_tolower( PrefixString[ i ] ) != vos_tolower( m_pData[ i ] ) )
			{
				return false;
			}
		}

		return true;
	}

	/******************************************************************************

	>	bool String::StartsWith(const tchar *szPrefix) const

		Author:		Tim Browse
		Created:	9/10/99
		Purpose:	Check to see if the string begins with the specified prefix.
					The function is case-insensitive.
		Inputs:		szPrefix - the prefix to check for
		Returns:	true => this string starts with szPrefix
					false => otherwise

	******************************************************************************/

	bool String::StartsWith(const tchar *szPrefix) const
	{
		if (m_pData == NULL)
			return false;
			
		// Make sure we're long enough to compare
			
		uint StringLength 	= strlen(szPrefix);
		if (m_StringLength < StringLength)
		{
			return false;
		}
		
		// Compare
		
		for( int i = 0; i < StringLength; i++ )
		{
			if( vos_tolower( szPrefix[ i ] ) != vos_tolower( m_pData[ i ] ) )
			{
				return false;
			}
		}

		return true;
	}



	/******************************************************************************

	>	bool String::EndsWith(const String &Suffix) const
							 ,
		Author:		Sean Solle
		Created:	2002/01/17
		Purpose:	Check to see if the string ends with the specified suffix.
					The function is case-insensitive.
		Inputs:		Prefix - the prefix to check for
		Returns:	true => this string ends with Suffix
					false => otherwise

	******************************************************************************/

	bool String::EndsWith(const String &Suffix) const
	{
		if (m_pData == NULL)
			return false;
			
		// Make sure we're long enough to compare
		const char * SuffixString = Suffix.c_str();
		uint StringLength = Suffix.GetLength();

		if (m_StringLength < StringLength)
		{
			return false;
		}
		
		// Find where suffix should start.
		int offset = m_StringLength-StringLength;
		
		// Compare
		for (int i = 0; i < StringLength; i++)
		{
			if ( vos_tolower( SuffixString[ i ] ) != vos_tolower( m_pData[ i + offset ] ) )
			{
				return false;
			}
		}

		return true;
	}

	/******************************************************************************

	>	bool String::EndsWith(const tchar *szSuffix) const
							 ,
		Author:		Sean Solle
		Created:	2002/01/17
		Purpose:	Check to see if the string ends with the specified suffix.
					The function is case-insensitive.
		Inputs:		Prefix - the prefix to check for
		Returns:	true => this string ends with Suffix
					false => otherwise

	******************************************************************************/

	bool String::EndsWith(const tchar *szSuffix) const
	{
		if (m_pData == NULL)
			return false;
			
		// Make sure we're long enough to compare
		uint StringLength = strlen(szSuffix);

		if (m_StringLength < StringLength)
		{
			return false;
		}
		
		// Find where suffix should start.
		int offset = m_StringLength-StringLength;
		
		// Compare
		for (int i = 0; i < StringLength; i++)
		{
			if ( vos_tolower( szSuffix[ i ] ) != vos_tolower( m_pData[ i + offset ] ) )
			{
				return false;
			}
		}

		return true;
	}


	/******************************************************************************

	>	void String::Concat(const tchar *pSrc)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Generic string concatenation function.
		Inputs:		pSrc - string to concatenate with this one

	******************************************************************************/

	void String::Concat(const tchar *pSrc)
	{
		// Check for simple cases
		if (pSrc == NULL)
		{
			// Nothing to do!
			return;
		}

		if (m_StringLength == 0)
		{
			Copy(pSrc);
			return;
		}

		// Do we have enough space to hold this string?
		int RequiredLength = m_StringLength + strlen(pSrc);
		if (m_BufferLength <= RequiredLength)
		{
			// No - get a new buffer, copy our data into it, and discard 
			// the old buffer.
			m_BufferLength = RequiredLength + 1;
			tchar *pNewBuffer = new tchar[m_BufferLength];
			vos_strcpy(pNewBuffer, m_pData);
			delete [] m_pData;
			m_pData = pNewBuffer;

			// m_StringLength doesn't change yet...
		}

		// Concatentate the string, as we now have a big enough buffer
		ASSERT(m_pData != NULL);
		vos_strcat(m_pData, pSrc);
		m_StringLength = RequiredLength;
	}


	/******************************************************************************

	>	void String::AppendChars(const tchar *pSrc, int nCharsToCopy)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Copy a slice out of a C-style string, and append it to this one.
		Inputs:		pSrc - string to concatenate with this one
					nCharsToCopy - the number of characters to copy.

	******************************************************************************/

	void String::AppendChars(const tchar *pSrc, int nCharsToCopy)
	{
		// Check for simple cases
		if (pSrc == NULL)
		{
			// Nothing to do!
			return;
		}

		// Do we have enough space to hold this string?
		int RequiredLength = m_StringLength + nCharsToCopy;
		if (m_BufferLength <= RequiredLength)
		{
			// No - get a new buffer, copy our data into it, and discard 
			// the old buffer.
			m_BufferLength = RequiredLength + 1;
			tchar *pNewBuffer = new tchar[m_BufferLength];

			if (m_pData != NULL)
			{
				vos_strcpy(pNewBuffer, m_pData);
				delete [] m_pData;
			}
			else
			{
				// Empty string - just terminate the new string buffer.
				*pNewBuffer = EOS;
			}

			m_pData = pNewBuffer;

			// m_StringLength doesn't change yet...
		}

		// Concatentate the string, as we now have a big enough buffer
		ASSERT(m_pData != NULL);
		memcpy(m_pData + m_StringLength, pSrc, nCharsToCopy);
		m_StringLength = RequiredLength;

		// Terminate the string
		m_pData[m_StringLength] = EOS;

		// Ensure we have a valid length for the string - the data we just
		// copied may have contained 0 bytes.
		m_StringLength = strlen(m_pData);
	}



	/******************************************************************************

	>	void String::Truncate(int NewLength)

		Author:		Mike Kenny
		Created:	11/12/98
		Inputs:		NewLength = truncate the string to this number of characters
		Purpose:	Truncate operator.

	******************************************************************************/

	void String::Truncate(int NewLength)
	{
		if (NewLength>=m_StringLength || NewLength<0)
			return;

		// set the new string length
		m_StringLength = NewLength;

		// Terminate the string
		m_pData[m_StringLength] = EOS;
	}


	/******************************************************************************

	>	const String& String::operator+=(const String& String)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Concatenation operator.
		Inputs:		String - the string to append to this one.
		Returns:	Reference to this string.

	******************************************************************************/

	const String& String::operator+=(const String& String)
	{
		Concat(String.m_pData);
		return *this;
	}




	/******************************************************************************

	>	const String& String::operator+=(tchar Ch)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Concatenation operator.
		Inputs:		Ch - the character to append to this string.
		Returns:	Reference to this string.

	******************************************************************************/

	const String& String::operator+=(tchar Ch)
	{
		static tchar Temp[2];
		Temp[0] = Ch;
		Temp[1] = EOS;

		Concat(Temp);
		return *this;
	}




	/******************************************************************************

		const String& String::operator+=(const tchar *pString)

	>	Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Concatenation operator.
		Inputs:		String - the string to append to this one.
		Returns:	Reference to this string.

	******************************************************************************/

	const String& String::operator+=(const tchar *pString)
	{
		Concat(pString);
		return *this;
	}




	/******************************************************************************

	>	void String::CopyChars(const tchar *pSrc, int nFirst, int nCharsToCopy)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Copies the substring specified.
		Inputs:		pSrc - the string to copy from.
					nFirst - the character to start from (0 based)
					nCount - number of characters to get from the string.

	******************************************************************************/

	void String::CopyChars(const tchar *pSrc, int nFirst, int nCharsToCopy)
	{
		// Don't bother if we don't have a source string
		if (pSrc == NULL)
			return;

		// Ignore null requests
		if (nCharsToCopy <= 0)
			return;

		// Do we have enough space to hold this string?
		if (m_BufferLength <= nCharsToCopy)
		{
			// No - discard our buffer and get one big enough.
			m_BufferLength = nCharsToCopy + 1;
			delete [] m_pData;
			m_pData = new tchar[m_BufferLength];
		}

		// Copy the string, as we now have a big enough buffer
		ASSERT(m_pData != NULL);
		memcpy(m_pData, pSrc + nFirst, nCharsToCopy * sizeof(tchar));
		m_StringLength = nCharsToCopy;

		// Terminate the string.
		m_pData[m_StringLength] = EOS;
	}


	/******************************************************************************

	>	void String::RemoveTrailingChar(tchar Ch)

		Author:		Tim Browse
		Created:	19/5/97
		Purpose:	Remove the last character of a string if it matches the one
					the caller specified.
		Inputs:		Ch - character to remove.

	******************************************************************************/

	void String::RemoveTrailingChar(tchar Ch)
	{
		if (m_StringLength > 0)
		{
			tchar LastCh = GetAt(m_StringLength - 1);
			if (LastCh == Ch)
			{
				// Remove the last character
				m_StringLength--;

				// Terminate the string.
				m_pData[m_StringLength] = EOS;
			}
		}
	}


	/******************************************************************************

	>	String String::Mid(int nFirst, int nCount) const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Return a string object that contains the substring specified.
		Inputs:		nFirst - the character to start from (0 based)
					nCount - number of characters to get from the string.
		Returns:	The substring specified.

	******************************************************************************/

	String String::Mid(int nFirst, int nCount) const
	{
		String Result;

		if (nFirst >= m_StringLength)
		{
			// No characters to copy
			return Result;
		}

		// Use helper function
		int nCharsAvailable = m_StringLength - nFirst;
		if (nCount > nCharsAvailable)
			nCount = nCharsAvailable;

		Result.CopyChars(m_pData, nFirst, nCount);
		return Result;
	}




	/******************************************************************************

	>	String String::Mid(int nFirst) const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Return a string object that contains the substring specified.
					The substring is from the character specified up to the end of 
					the string.
		Inputs:		nFirst - the character to start from (0 based)
		Returns:	The substring specified.

	******************************************************************************/

	String String::Mid(int nFirst) const
	{
		String Result;

		if (nFirst >= m_StringLength)
		{
			// No characters to copy
			return Result;
		}

		// Use helper function
		Result.CopyChars(m_pData, nFirst, m_StringLength - nFirst);
		return Result;
	}





	/******************************************************************************

	>	String String::Left(int nCount) const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Return a substring of the first n characters of the string.
		Inputs:		nCount - the number of characters to copy.
		Returns:	The substring specified

	******************************************************************************/

	String String::Left(int nCount) const
	{
		// Use helper function
		String Result;
		Result.CopyChars(m_pData, 0, nCount);
		return Result;
	}




	/******************************************************************************

	>	String String::Right(int nCount) const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Return a substring of the last n characters of the string.
		Inputs:		nCount - the number of characters to copy.
		Returns:	The substring specified

	******************************************************************************/

	String String::Right(int nCount) const
	{
		// Use helper function
		String Result;

		// Limit substring to max size of source string
		if (nCount > m_StringLength)
			nCount = m_StringLength;

		Result.CopyChars(m_pData, m_StringLength - nCount, nCount);
		return Result;
	}




	/******************************************************************************

	>	void String::MakeUpper()

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Convert the whole string to upper case.

	******************************************************************************/

	void String::MakeUpper()
	{
		if ((m_pData != NULL) && (m_StringLength > 0))
		{
			for (int i = 0; i < m_StringLength; i++)
			{
				m_pData[i] = vos_toupper(m_pData[i]);
			}
		}
	}




	/******************************************************************************

	>	void String::MakeLower()

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Convert the whole string to lower case.

	******************************************************************************/

	void String::MakeLower()
	{
		if ((m_pData != NULL) && (m_StringLength > 0))
		{
			for (int i = 0; i < m_StringLength; i++)
			{
				m_pData[i] = vos_tolower(m_pData[i]);
			}
		}
	}




	/******************************************************************************

	>	void String::MakeReverse()

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Reverse the string on a character basis, 
						i.e. "hello" becomes "olleh".

	******************************************************************************/

	void String::MakeReverse()
	{
		if ((m_pData != NULL) && (m_StringLength > 1))
		{
			int i = 0;
			int j = m_StringLength - 1;
			while (i < j)
			{
				// Swap characters
				tchar Temp = m_pData[i];
				m_pData[i] = m_pData[j];
				m_pData[j] = Temp;

				// Move along
				i++;
				j--;
			}
		}
	}




	/******************************************************************************

	>	bool String::Find(tchar ch, uint *pPosition = NULL) const

		Author:		Tim Browse
		Created:	16/9/99
		Purpose:	Find the first occurence of the character within the string.
		Inputs:		ch - the character to look for.
		Outputs:	pPosition - the index of the character that was found, if any.
		Returns:	true => character was found in string
					false=> character not found

	******************************************************************************/

	bool String::Find(tchar ch, uint *pPosition) const
	{
		if ((m_pData != NULL) && (m_StringLength > 0))
		{
			// Look for first occurrence of character.
			for (uint i = 0; i < m_StringLength; i++)
			{
				if (m_pData[i] == ch)
				{
					// Found it!
					if (pPosition != NULL)
						*pPosition = i;

					return true;
				}
			}
		}

		// Not found
		return false;
	}




	/******************************************************************************

	>	bool String::ReverseFind(tchar ch, uint *pPosition = NULL) const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Find the last occurence of the character within the string.
		Inputs:		ch - the character to look for.
		Outputs:	pPosition - the index of the character that was found, if any.
		Returns:	true => character was found in string
					false=> character not found

	******************************************************************************/

	bool String::ReverseFind(tchar ch, uint *pPosition) const
	{
		if ((m_pData != NULL) && (m_StringLength > 0))
		{
			// Look for first occurence of character.
			for (int i = (m_StringLength - 1); i >= 0; i--)
			{
				if (m_pData[i] == ch)
				{
					// Found it!
					if (pPosition != NULL)
						*pPosition = i;

					return true;
				}
			}
		}

		// Not found
		return false;
	}



	/******************************************************************************

	>	bool String::Find(const tchar *pSubString, uint *pPosition = NULL) const

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Find the first occurence of the substring within this string.
		Inputs:		pSubString - the substring to search for.
		Returns:	0 or greater => index of the character
					negative => character not found

	******************************************************************************/

	bool String::Find(const tchar *pSubString, uint *pPosition) const
	{
		if ((m_pData != NULL) && (m_StringLength > 0))
		{
			char *Result = strstr(m_pData, pSubString);
			if (Result != NULL)
			{
				// Found it!
				if (pPosition != NULL)
					*pPosition = (uint) (Result - m_pData);

				return true;
			}
		}
		
		// Not found
		return false;
	}


	/******************************************************************************

	>	void String::Replace(tchar Old, tchar New)

		Author:		Tim Browse
		Created:	29/09/99
		Purpose:	Replace all occurrences of the character 'Old' with the
					character 'New'.

	******************************************************************************/

	void String::Replace(tchar Old, tchar New)
	{
		if ((m_pData != NULL) && (m_StringLength > 0))
		{
			for (int i = 0; i < m_StringLength; i++)
			{
				if (m_pData[i] == Old)
				{
					// Replace this character
					m_pData[i] = New;
				}
			}
		}
	}

	/**************************************************************************
	
	>	void String::Replace(const tchar* szOld, const tchar* szNew)
	
		Author:		Tim Browse
		Date:		14/4/2002
		Inputs:		szOld - sequence to replace (must be >0 chars long)
					szNew - sequence to overwrite the old sequence with
	
		Purpose:	Replace a literal sequence of characters with a different
					sequence (sequences do not need to be of equal length).

  					Not currently the most efficient way of doing it perhaps,
					but it works.

	**************************************************************************/
	
	void String::Replace(const tchar* szOld, const tchar* szNew)
	{
		// Nothing to do on empty strings
		if (m_pData == NULL)
			return;

		// Work out how much we have to change the string length if
		// we find a match.
		int PatternLength = strlen(szOld);
		int ReplacementLength = strlen(szNew);
		int SizeDelta = ReplacementLength - PatternLength;

		// Keep looking for strings to replace...
		char* szSubStr;
		char* szStringToSearch = m_pData;

		while ((szSubStr = strstr(szStringToSearch, szOld)) != NULL)
		{
			// We've found a string to replace - alter the
			// destination string, so we have the right number
			// of characters to replace with

			int PatternOffset = szSubStr - m_pData;

			char* szPatternStart = m_pData + PatternOffset;

			if (SizeDelta != 0)
			{

				// Make sure we have enough room
				ReserveCapacity(m_StringLength + SizeDelta);

				// Shift the remainder of the string (if needed)
				int RemainderLength = m_StringLength - (PatternOffset + PatternLength);

				// We use memmove, not memcpy, as regions are likely to overlap
				szPatternStart = m_pData + PatternOffset;

				memmove(szPatternStart + PatternLength + SizeDelta, 
						szPatternStart + PatternLength,
						RemainderLength);

				// Make sure string is correctly formed
				m_StringLength += SizeDelta;
				m_pData[m_StringLength] = EOS;
			}

			// Replace old pattern with new one
			memcpy(szPatternStart, szNew, ReplacementLength);

			// Start looking after the sequence we just inserted
			szStringToSearch = szPatternStart + ReplacementLength;
		}
	}

	/**************************************************************************
	
	>	void String::Replace(const String& Old, const String& New)
	
		Author:		Tim Browse
		Date:		14/4/2002
		Inputs:		Old - sequence to replace (must be >0 chars long)
					New - sequence to overwrite the old sequence with
	
		Purpose:	Replace a literal sequence of characters with a different
					sequence (sequences do not need to be of equal length).

  					Not currently the most efficient way of doing it perhaps,
					but it works.

	**************************************************************************/
	
	void String::Replace(const String& Old, const String& New)
	{
		Replace(Old.c_str(), New.c_str());
	}

	/**************************************************************************
	
	>	void String::Replace(const tchar* szOld, const String& New)
	
		Author:		Tim Browse
		Date:		28/4/2002
		Inputs:		szOld - sequence to replace (must be >0 chars long)
					New - sequence to overwrite the old sequence with
	
		Purpose:	Replace a literal sequence of characters with a different
					sequence (sequences do not need to be of equal length).

	
	**************************************************************************/
	
	void String::Replace(const tchar* szOld, const String& New)
	{
		Replace(szOld, New.c_str());
	}

	/******************************************************************************

	>	String operator+(const String& String1, const String& String2)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Join two strings together to form a new one.
		Inputs:		The strings to join together.
		Returns:	The result of concatenating String2 onto String1.

	******************************************************************************/

	String operator+(const String& String1, const String& String2)
	{
		String Result(String1);
		Result += String2;
		return Result;
	}



	/******************************************************************************

	>	String operator+(const String& TheString, tchar Ch)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Join a character onto the end of a string.
		Inputs:		The string and character to join together.
		Returns:	The result of concatenating Ch onto String.

	******************************************************************************/

	String operator+(const String& TheString, tchar Ch)
	{
		String Result(TheString);
		String ChString(Ch);
		Result += ChString;
		return Result;
	}




	/******************************************************************************

	>	String operator+(tchar Ch, const String& TheString)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Join a string onto the end of a single character string.
		Inputs:		The string and character to join together.
		Returns:	The result of concatenating String onto Ch.

	******************************************************************************/

	String operator+(tchar Ch, const String& TheString)
	{
		String Result(Ch);
		Result += TheString;
		return Result;
	}




	/******************************************************************************

	>	String operator+(const String& String1, const tchar *pString2)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Join two strings together to form a new one.
		Inputs:		The strings to join together.
		Returns:	The result of concatenating pString2 onto String1.

	******************************************************************************/

	String operator+(const String& String1, const tchar *pString2)
	{
		String Result(String1);
		Result += pString2;
		return Result;
	}




	/******************************************************************************

	>	String operator+(const tchar *pString1, const String& String2)

		Author:		Tim Browse
		Created:	20/3/97
		Purpose:	Join two strings together to form a new one.
		Inputs:		The strings to join together.
		Returns:	The result of concatenating String2 onto pString1.

	******************************************************************************/

	String operator+(const tchar *pString1, const String& String2)
	{
		String Result(pString1);
		Result += String2;
		return Result;
	}

	/******************************************************************************

	>	String String::TruncateRight(int nCount) const
							 ,
		Author:		Sean Solle
		Created:	3/6/97
		Purpose:	Return a substring with the right-most n characters stripped.
		Inputs:		nCount - the number of characters to strip.
		Returns:	The substring specified

	******************************************************************************/

	String String::TruncateRight(int nCount) const
	{
		// Use helper function
		String Result;

		if (nCount < m_StringLength)
		{
			Result.CopyChars(m_pData, 0, m_StringLength-nCount);
		}
		else
		{
			Result.Empty();
		}

		return Result;
	}



	/**************************************************************************
	
	>	void String::Printf(const char* szFormat, ...)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		szFormat - printf format string
	
		Purpose:	Use printf style format strings to set a String object's
					contents.

					For example:

					\code

						String S;
						S.Printf("%d is a %s number: %d", 42, "magic");

					\endcode

	**************************************************************************/
	
	void String::Printf(const char* szFormat, ...)
	{
		va_list ArgList;
		va_start(ArgList, szFormat);

		Printf_va(szFormat, ArgList);

		va_end(ArgList);
	}

	/**************************************************************************
	
	>	void String::Printf_va(const char* szFormat, va_list ArgList)
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Helper implementation function for Printf().
	
	**************************************************************************/
	
	void String::Printf_va(const char* szFormat, va_list ArgList)
	{
		char szTemp[512];
		vos_vsprintf(szTemp, szFormat, ArgList);

		assign(szTemp);
	}

	/**************************************************************************
	
	>	String String::QuoteString(vchar QuoteChar = '"')
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		QuoteChar - the character to use as a quote character
		Returns:	String - the quoted string
	
		Purpose:	Add quotes to a string, intelligently escaping any
					embedded quote characters.

					For example:

					\code

						String S1("Hello, world!");
						String S2("Jim said, 'Hi!'");

						// S1 will become "'Hello, world!'"
						S1 = S1.QuoteString(''');

						// S2 will become "'Jim said, \'Hi!\''"
						S2 = S2.QuoteString(''');

						// Quote character can be arbitray, e.g. 
						String S3("It's full of stars!");
						
						// S3 wil become "*It's full of stars!*"
						S3 = S3.QuoteString('*');

					\endcode

	**************************************************************************/
	
	String String::QuoteString(vchar QuoteChar) const
	{
		vchar QuoteSequence[] = _T("\\x");

		QuoteSequence[1] = QuoteChar;

		String QuotedString;

		// Whole string is quoted, so start with a quote.
		QuotedString += QuoteChar;
		
		int i = 0;
		int Length = GetLength();

		while (i < Length)
		{
			tchar ch = GetAt(i);

			if (ch == QuoteChar)
			{
				// Found a quote, so escape it
				QuotedString += QuoteSequence;
			}
			else
			{
				// just add the character
				QuotedString += ch;
			}

			// Try next character
			i++;
		}

		// All done, so finish string quoting
		QuotedString += QuoteChar;

		// Give caller the result
		return QuotedString;
	}

	

	/**************************************************************************
	
	>	void String::Trim()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Remove leading and trailing whitespace from the String.

					e.g.

					\code

						// A string with white space
						String S("  \t  Hello there!  \t           ");

						// S will become "Hello there!");
						S.Trim();

					\endcode

	**************************************************************************/
	
	void String::Trim()
	{
		// Get indices to either end of the string
		int Length = GetLength();
		int iStart = 0;
		int iEnd = Length - 1;

		// Move indices inwards to remove whitespace
		while ((iStart < Length) && isspace(at(iStart)))
			iStart++;

		while ((iEnd >= iStart) && isspace(at(iEnd)))
			iEnd--;

		Length = (iEnd - iStart) + 1;

		if (Length == 0)
		{
			// String consists entirely of whitespace, so just blank it.
			Clear();
		}
		else
		{
			// Remove trailing and preceding characters
			CopyChars(m_pData, iStart, (iEnd - iStart) + 1);
		}
	}


	/**************************************************************************
	
	>	int String::CompareOrder(const String& Other) const
	
		Author:		Tim Browse
		Date:		1/4/2002
		Inputs:		Other - the string to compare this one against.
		Returns:	-1 => this precedes Other in sort order
					 0 => this and Other are equivalent in sort order
					 1 => Other precedes this in sort order
	
		Purpose:	Compare two strings to find out the relative ordering - 
					similar to strcmp().
	
	**************************************************************************/
	
	int String::CompareOrder(const String& Other) const
	{
		// Avoid accessing m_pData pointer of either string if it could be NULL.
		if (Other.m_StringLength == 0)
		{
			// If both are zero length, then they're equal
			if (m_StringLength == 0)
				return 0;
			else
				return 1;
		}
		else if (m_StringLength == 0)
		{
			// This one's zero length, but the other one isn't => less_than
			return -1;
		}

		// Use strcmp to compare
		return vos_strcmp(m_pData, Other.m_pData);
	}

} // namespace vos

