//
// StringParser.cpp - class to parse String objects in a simple way
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "StringParser.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/**************************************************************************
	
	>	StringParser::StringParser()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	StringParser::StringParser() : m_ParsePos(0)
	{
	}

	/**************************************************************************
	
	>	StringParser::~StringParser()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	StringParser::~StringParser()
	{
	}

	/**************************************************************************
	
	>	StringParser::StringParser(const String& Other)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		Other - StringParser object to copy into this one.
	
		Purpose:	Copy constructor
	
	**************************************************************************/
	
	StringParser::StringParser(const String& Other) : String(Other),
													  m_ParsePos(0)
	{
	}


	/**************************************************************************
	
	>	void StringParser::SetContents(const String& Other)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		Other - the new string to parse
	
		Purpose:	Set the contents of the string parser to a new string,
					and reset the parsing state.
	
	**************************************************************************/
	
	void StringParser::SetContents(const String& Other)
	{
		// Copy this string
		*this = Other;

		ResetParsing();
	}

	/**************************************************************************
	
	>	void StringParser::ResetParsing()
	
		Author:		Tim Browse
		Date:		15/4/2002
	
		Purpose:	Clear the parsing state, ready to start parsing again from
					the start of the string.
	
	**************************************************************************/
	
	void StringParser::ResetParsing()
	{
		m_ParsePos = 0;
	}

	/**************************************************************************
	
	>	bool StringParser::ReadInt(int &outValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Outputs:	outValue - the integer value that was parsed.
		Returns:	true => valid integer found
					false => otherwise
	
		Purpose:	Read an integer value (with optional leading + or -
					symbol being handled correctly).
	
	**************************************************************************/
	
	bool StringParser::ReadInt(int &outValue)
	{
		SkipWS();

		// Look for sign characters...
		char Ch = NextChar();
		if (Ch == 0)
		{
			return false;
		}

		bool bNegative = false;
		if ((Ch == '-') || (Ch == '+'))
		{
			// Found a sign digit...
			bNegative = (Ch == '-');
			ReadChar();
			SkipWS();
		}
		else if (!isdigit(Ch))
		{
			return false;
		}

		// Read digits
		if (!ReadCardinal(outValue))
			return false;

		// Negate if necessary
		if (bNegative)
			outValue = -outValue;

		// We found a number
		return true;
	}

	/**************************************************************************
	
	>	bool StringParser::ReadDouble(double &outValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Outputs:	outValue - the floating point number that was decoded
							   from the string.
		Returns:	true => valid floating point number parsed ok
					false => otherwise
	
		Purpose:	Read a floating point number from the string.
					Scientific notation (e.g. 10E9) is not supported.
	
	**************************************************************************/
	
	bool StringParser::ReadDouble(double &outValue)
	{
		SkipWS();

		// Look for sign characters...
		tchar Ch = NextChar();
		if (Ch == 0)
		{
			// run out of string to read
			return false;
		}

		String Result;

		if ((Ch == '-') || (Ch == '+'))
		{
			// Found a sign digit...
			Result += ReadChar();
			SkipWS();
		}
		else if (!isdigit(Ch))
		{
			// Not a number or sign - error
			return false;
		}

		// Read digits before decimal point
		String Number;
		if (!ReadCardinalString(Number))
			return false;

		Result += Number;
		
		// If next character is decimal point, read fractional part
		if (NextChar() == '.')
		{
			Result += ReadChar();

			if (!ReadCardinalString(Number))
				return false;

			Result += Number;
		}

		// Compute whole number
		outValue = atof(Result.c_str());

		// We found a number
		return true;
	}

	/**************************************************************************
	
	>	bool StringParser::ReadCardinal(int &outValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Outputs:	outValue - the cardinal value that was read from the string
		Returns:	true => valid cardinal value was decoded ok
					false => otherwise
	
		Purpose:	Read a cardinal (unsigned integer) number from the string.
	
	**************************************************************************/
	
	bool StringParser::ReadCardinal(int &outValue)
	{
		String Number;
		if (!ReadCardinalString(Number))
			return false;

		// Convert the number
		outValue = atoi(Number.c_str());
		
		// We found a number
		return true;
	}

	/**************************************************************************
	
	>	bool StringParser::ReadCardinalString(String &outValue)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Outputs:	outValue - the string of digits read from the string
		Returns:	true => non-zero sequence of decimal digits was found
					false => otherwise
	
		Purpose:	Helper function for parsing functions - reads a string
					of decimal digits from the string.
	
	**************************************************************************/
	
	bool StringParser::ReadCardinalString(String &outValue)
	{
		// Keep collecting digits until they stop
		String Number;
		char Ch = NextChar();
		while (isdigit(Ch))
		{
			// Add to number string
			ReadChar();
			Number += Ch;
			Ch = NextChar();
		}

		// If we have a number, convert it and return it.
		if (Number.GetLength() == 0)
		{
			// No number to speak of...
			return false;
		}
		
		outValue = Number;	
		// We found a number
		return true;
	}
	

	/**********************************************************

	>	bool StringParser::ReadHex(int &outValue)

		Author:		Louise Bowden
		Date:		2/11/00

		Outputs:	outValue - the decimal value of the hex string
							   parsed from the string. 

		Returns:	true => hex number parsed ok
					false => otherwise

		Purpose:	Parse a hex number from the string.

					TODO: make this not crap

	***********************************************************/

	bool StringParser::ReadHex(int &outValue)
	{
		// Keep collecting digits until they stop
		String Number;
		char Ch = NextChar();

		// Add to number string
		while(m_ParsePos < GetLength())
		{
			ReadChar();
			Number += Ch;
			Ch = NextChar();
		}

		// If we have a number, convert it and return it.
		if (Number.GetLength() == 0)
		{
			// No number to speak of...
			return false;
		}

		
		// and convert hex to int - the long way...
		int i;
		int Result = 0;
		int Base = 16;
		int Unit = Number.GetLength();

		for(i=0; i < Number.GetLength(); i++)
		{
			int Amount = 1;
			int j = Unit - 1;
			while(j > 0)
			{
				Amount = Amount * Base;
				j--;
			}
			if(isdigit(Number.GetAt(i)))
			{
				Result = Result + (atoi(Number.Mid(i,1)) * Amount);
			}
			if(Number.GetAt(i) == 'F')
				Result = Result + (15 * Amount);
			if(Number.GetAt(i) == 'E')
				Result = Result + (14 * Amount);
			if(Number.GetAt(i) == 'D')
				Result = Result + (13 * Amount);
			if(Number.GetAt(i) == 'C')
				Result = Result + (12 * Amount);
			if(Number.GetAt(i) == 'B')
				Result = Result + (11 * Amount);
			if(Number.GetAt(i) == 'A')
				Result = Result + (10 * Amount);
			Unit--;
		}

		outValue = Result;
		
		// We found a number
		return true;
	}

	
	/**************************************************************************

	>	bool StringParser::ReadString(String &outValue, String Delimiters,
									  bool bUseQuotes)

		Author:		Tim Browse
		Date:		15/10/2000
		Inputs:		Delimiters - list of delimiters - reading stops when one
								 of these is found
					bUseQuotes - if true, then it handles quoted strings in the
								 input - that is, delimiters are not recognised
								 if in quoted string.
								 NB. If true, you should not use the double quote
									 character (") as a delimiter.
		Outputs:	outValue - the string that was read.
		Returns:	true  => non-empty string was read.
					false => otherwise

		Purpose:	Read a string from the string parser object, stopping at
					the specified delimiters.

	**************************************************************************/

	bool StringParser::ReadString(String &outValue, String Delimiters,
								  bool bUseQuotes)
	{
		if (m_ParsePos >= GetLength())
		{
			// Out of chars...
			return false;
		}

		// Check for empty string
		if (Delimiters.Find(at(m_ParsePos)))
		{
			// No characters before delimiters are found => no string
			return false;
		}

		outValue.Clear();

		bool bInQuotes = false;

		while ((m_ParsePos < GetLength()) &&
			   (bInQuotes || !Delimiters.Find(at(m_ParsePos))))
		{
			// If client asked us to, cope with strings enclosed in quotes
			if (bUseQuotes)
			{
				if (bInQuotes)
				{
					// Are we ending a string?
					if (at(m_ParsePos) == '"')
					{
						// Yes - all done
						m_ParsePos++;
						return true;
					}
					else
					{
						// No - fall through to normal behaviour
					}
				}
				else
				{
					// Are we starting a string?
					if (at(m_ParsePos) == '"')
					{
						// Yes - so remember this
						bInQuotes = true;
						m_ParsePos++;

						// And don't copy the quote character.
						continue;
					}
					else
					{
						// No - fall through to normal behaviour
						// (and we don't look for quotes anymore)
						bUseQuotes = false;
					}
				}
			}

			// Haven't reached a delimiter or end of string, so carry on
			outValue += at(m_ParsePos);
			m_ParsePos++;
		}

		// We got a string of some description => success
		return true;
	}

	
	/**************************************************************************

	>	bool StringParser::ReadQuotedString(String& outValue, tchar chEscape)

		Author:		Tim Browse
		Date:		15/10/2000
		Inputs:		chEscape - the character to use as escape character
		Outputs:	outValue - the String that was read from the input
		Returns:	true => well formed string read from input
					false => otherwise

		Purpose:	Read a 'quoted' string from the input.  The string should
					start at the next available character in the input - the
					first character is assumed to be the quote character used in
					the string.

					The function copes with escaped quote characters that
					are escaped with the specified escape character.

					Example usage:

						String S("'A quoted string, using \\' as the quote'");

						StringParser Parser(S);

						String Result
						Parser.ReadQuotedString(Result, '\');


						Result will contain: 

							A quoted string, using ' as the quote

	**************************************************************************/

	bool StringParser::ReadQuotedString(String& outValue, tchar chEscape)
	{
		// Ignore requests to read from null input
		if (EOS())
			return false;

		outValue.Clear();

		// The string starts with the quote character
		tchar chQuote = ReadChar();

		// Ok, keep reading characters until we find the closing quote
		while (!EOS())
		{
			// The next character may be:
			//
			//		* a quote character
			//		* an escape character
			//		* a normal character
			//
			// if it is a quote character, then we have finished with the string.
			//
			// if it is an escape character, then discard the escape and add the
			// next character to the string.
			//
			// otherwise, we add the character to the string.

			tchar ch = ReadChar();

			if (ch == chQuote)
			{
				// End of string
				return true;
			}
			else if (ch == chEscape)
			{
				// Escaped character
				if (EOS())
				{
					// input ends with a trailing escape character - illegal
					// format for a quoted string
					return false;
				}

				// Discard the escape character, and add the next one to the string
				outValue += ReadChar();
			}
			else
			{
				// Normal character - just add it
				outValue += ch;
			}
		}

		// If we get here, then we've reached the end of input without
		// finding a correctly terminated string
		return false;
	}

	/**************************************************************************
	
	>	tchar StringParser::ReadChar()
	
		Author:		Tim Browse
		Date:		15/4/2002
		Returns:	tchar - the next character from the string, or NUL
							if there are no more characters.
	
		Purpose:	Read the next character from the string being parsed
					(i.e. from the current parse point)
	
	**************************************************************************/
	
	tchar StringParser::ReadChar()
	{
		if (m_ParsePos >= GetLength())
		{
			// Out of chars...
			return 0;
		}
		else
		{
			// Should not have NULL data if we get to this clause
			tchar Ch = at(m_ParsePos);
			m_ParsePos++;
			return Ch;
		}
	}

	/**************************************************************************
	
	>	tchar StringParser::NextChar()
	
		Author:		Tim Browse
		Date:		15/4/2002
		Returns:	The next character (or 0 if there are no more available)
	
		Purpose:	Look at the next character from the string being parsed,
					but do not remove it or change parse point.

					This allows one-character lookahead.
	
	**************************************************************************/
	
	tchar StringParser::NextChar()
	{
		if (m_ParsePos >= GetLength())
		{
			// Out of chars...
			return 0;
		}
		else
		{
			// Should not have NULL data if we get to this clause
			tchar Ch = at(m_ParsePos);
			return Ch;
		}
	}

	/**************************************************************************
	
	>	bool StringParser::SkipWS()
	
		Author:		Tim Browse
		Date:		15/4/2002
		Returns:	always true - this makes it easy to slip SkipWS() into
								  a compound expression like:

									if (Parser.ReadInt(X) &&
										Parser.SkipWS() &&
										Parser.ReadInt(Y))
									{
										// Do stuff...
									}

		Purpose:	Skip past any white space - i.e. spaces, tabs etc.

					The whitespace is defined by the Standard C Library
					function isspace().
	
	**************************************************************************/
	
	bool StringParser::SkipWS()
	{
		while ((m_ParsePos < GetLength()) && isspace(at(m_ParsePos)))
		{
			m_ParsePos++;
		}
		return true;
	}

	/**************************************************************************
	
	>	bool StringParser::SkipDelim(String Delimiters)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		Delimiters - set of delimiters, e.g. ",.:;#"
		Returns:	true => delimiter was found (and skipped)
					false => no delimiter from the set was found.
	
		Purpose:	Skip one delimiter from the specified set of delimiters.

					If the next character in the string is one of the
					specified delimiters, then it is skipped.
	
	**************************************************************************/
	
	bool StringParser::SkipDelim(String Delimiters)
	{
		bool bFoundDelimiter = false;

		if ((m_ParsePos < GetLength()) && 
			(Delimiters.Find(at(m_ParsePos))))
		{
			m_ParsePos++;
			bFoundDelimiter = true;
		}

		return bFoundDelimiter;
	}

	/**************************************************************************
	
	>	bool StringParser::SkipDelims(String Delimiters)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		Delimiters - set of delimiters, e.g. ",.:;#"
		Returns:	true => one or more delimiters were found (and skipped)
					false => no delimiter from the set was found.
	
		Purpose:	Skips any characters in the string that belong to the
					specified set of delimiters.  Skipping is stopped as soon
					as a non-delimiter character is found.
	
	**************************************************************************/
	
	bool StringParser::SkipDelims(String Delimiters)
	{
		bool bFoundDelimiter = false;

		while ((m_ParsePos < GetLength()) && 
			   (Delimiters.Find(at(m_ParsePos))))
		{
			m_ParsePos++;
			bFoundDelimiter = true;
		}

		return bFoundDelimiter;
	}

	/**************************************************************************
	
	>	bool StringParser::SkipDelim(const tchar* szDelimiters)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		szDelimiters - set of delimiters, e.g. ",.:;#"
		Returns:	true => delimiter was found (and skipped)
					false => no delimiter from the set was found.
	
		Purpose:	Skip one delimiter from the specified set of delimiters.

					If the next character in the string is one of the
					specified delimiters, then it is skipped.
	
	**************************************************************************/
	
	bool StringParser::SkipDelim(const tchar* szDelimiters)
	{
		bool bFoundDelimiter = false;

		if ((m_ParsePos < GetLength()) && 
			(strchr(szDelimiters, at(m_ParsePos)) != NULL))
		{
			m_ParsePos++;
			bFoundDelimiter = true;
		}

		return bFoundDelimiter;
	}

	/**************************************************************************
	
	>	bool StringParser::SkipDelims(const tchar* szDelimiters)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		szDelimiters - set of delimiters, e.g. ",.:;#"
		Returns:	true => one or more delimiters were found (and skipped)
					false => no delimiter from the set was found.
	
		Purpose:	Skips any characters in the string that belong to the
					specified set of delimiters.  Skipping is stopped as soon
					as a non-delimiter character is found.
	
	**************************************************************************/
	
	bool StringParser::SkipDelims(const tchar* szDelimiters)
	{
		bool bFoundDelimiter = false;

		while ((m_ParsePos < GetLength()) && 
			   (strchr(szDelimiters, at(m_ParsePos)) != NULL))
		{
			m_ParsePos++;
			bFoundDelimiter = true;
		}

		return bFoundDelimiter;
	}


	/**************************************************************************
	
	>	void StringParser::SkipNChars(int nChars)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		nChars - the number of characters to skip
	
		Purpose:	Skip a fixed number of characters in the string.
	
	**************************************************************************/
	
	void StringParser::SkipNChars(int nChars)
	{
		while ((m_ParsePos < GetLength()) && (nChars > 0))
		{
			m_ParsePos++;
			nChars--;
		}
	}

	/**************************************************************************
	
	>	bool StringParser::EOS()
	
		Author:		Tim Browse
		Date:		15/4/2002
		Returns:	true => EOS reached; there are no more characters to parse
					false => otherwise
	
		Purpose:	Find out if the end of the string (EOS) has been reached.
	
	**************************************************************************/
	
	bool StringParser::EOS()
	{
		return (m_ParsePos >= GetLength());
	}
};

