//
// StringUtil.cpp
//

#include "StdHdr.h"

#include "StringUtil.h"

#include "StringSupport.h"

namespace vos
{
	/**************************************************************************
	
	>	String StringUtil::GetStorageSize(int SizeInBytes)
	
		Author:		Tim Browse
		Date:		15/4/2002
		Inputs:		SizeInBytes - the number of bytes to represent
		Returns:	A string representing the amount of storage.
	
		Purpose:	Create a string object that describes an amount of
					storage.

					e.g.

					\code

						// S1 will be "500 bytes"
						String S1 = StringUtil::GetStorageSize(500);
	
						// S2 will be "1 KB"
						String S2 = StringUtil::GetStorageSize(1024);
	
						// S3 will be "2 KB"
						String S3 = StringUtil::GetStorageSize(1800);
	
						// S4 will be "1.9 MB"
						String S4 = StringUtil::GetStorageSize(1800000);

					\endcode
	
	**************************************************************************/
	
	String StringUtil::GetStorageSize(uint64 SizeInBytes)
	{
		static const tchar* Units[] = 
		{
			_T("bytes"),
			_T("KB"),
			_T("MB"),
			_T("GB"),
			_T("TB"),
			NULL
		};

		int iUnits = 0;

		// We cast to int64 first, because MSVC (and possibly others) cannot
		// convert from unsigned int64 to a double directly.
		double fSize = (int64) SizeInBytes;
		while ((fSize > 1024.0) && (Units[iUnits] != NULL))
		{
			// Shift units
			iUnits++;
			fSize /= 1024;
		}

		// Construct string - use intelligent number of decimal places.
		String Desc;

		char szNumber[100];

		switch (iUnits)
		{
			case 0:
				// Bytes - use no decimal places
				vos_sprintf(szNumber, _T("%d"), (int) fSize);
				break;

			case 1:
				// Kilobytes - round up to nearest Kb
				vos_sprintf(szNumber, _T("%d"), (int) (fSize + 0.5));
				break;

			case 2:
			{
				// Megabytes - show to nearest 1dp, unless no decimal place 
				// is needed.
				double fRounded = fSize + 0.05;
				if ((fRounded - (int) fRounded) >= 0.1)
					vos_sprintf(szNumber, _T("%.1f"), fSize);
				else
					vos_sprintf(szNumber, _T("%d"), (int) fRounded);
				break;
			}

			case 3:
			default:
			{
				// Gigabytes (and larger) - show to nearest 2dp, unless no 
				// decimal places are needed.
				double fRounded = fSize + 0.005;
				if ((fRounded - (int) fRounded) >= 0.01)
					vos_sprintf(szNumber, _T("%.2f"), fSize);
				else
					vos_sprintf(szNumber, _T("%d"), (int) fRounded);
				break;
			}
		}

		Desc = szNumber;

		Desc += _T(" ");

		Desc += Units[iUnits];

		return Desc;
	}

	/**************************************************************************
	
	>	String StringUtil::FormatNumber(int32 Number)
	
		Author:		Tim Browse
		Date:		30/5/2002
		Inputs:		Number - the number to format
		Returns:	The formatted number string, e.g. "13,342,244"
	
		Purpose:	Format the number as a string, using the appropriate
					grouping and separator characters.
	
	**************************************************************************/
	
	String StringUtil::FormatNumber(int32 Number)
	{
		//
		// Simple but dumb algorithm to start with...
		//
		vos::String strNumber = vos::String(Number);

		int iChar = strNumber.GetLength() - 1;
		int Grouping = 0;

		vos::String FormattedNumber;

		// Build up string in reverse
		while (iChar >= 0)
		{
			FormattedNumber += strNumber[iChar--];

			Grouping++;
			if ((iChar >= 0) && ((Grouping % 3) == 0))
				FormattedNumber += ',';
		}

		// Reverse string and send to caller
		FormattedNumber.MakeReverse();
		return FormattedNumber;
	}

};
