//
// TreeNode.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "TreeNode.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif

namespace vos
{
	/******************************************************************************

	>	TreeNode::TreeNode()

		Author:		Tim Browse
		Created:	7/1/97
		Purpose:	Initialises this class

	******************************************************************************/

	TreeNode::TreeNode()
	{
		// set all the navigation pointers to NULL
		m_pParent = NULL;
		m_pNext = NULL;
		m_pPrevious = NULL;
		m_pFirstChild = NULL;
	}

	/******************************************************************************

	>	TreeNode::~TreeNode()

		Author:		Tim Browse
		Created:	9/5/99
		Purpose:	Deletes the node.

	******************************************************************************/

	TreeNode::~TreeNode()
	{
		// Check we have no linkage - otherwise the tree we are in will be
		// corrupted.
		ASSERT(m_pParent == NULL);
		ASSERT(m_pNext == NULL);
		ASSERT(m_pPrevious == NULL);
		ASSERT(m_pFirstChild == NULL);
	}



	/******************************************************************************

	>	TreeNode* TreeNode::GetNext() const

		Author:		Tim Browse
		Created:	7/1/97
		Returns:	A pointer to the next sibling in the tree after this one,
					or NULL if there is no next sibling.
		Purpose:	To find the next node in the tree.

	******************************************************************************/

	TreeNode* TreeNode::GetNext() const
	{
		return m_pNext;
	}




	/******************************************************************************

	>	TreeNode* TreeNode::GetPrevious() const

		Author:		Tim Browse
		Created:	7/1/97
		Returns:	A pointer to the previous sibling in the tree before this one,
					or NULL if this is the first node in the sibling list
		Purpose:	Find the previous node in the tree.

	******************************************************************************/

	TreeNode* TreeNode::GetPrevious() const
	{
		return m_pPrevious;
	}





	/******************************************************************************

	>	TreeNode* TreeNode::GetLastSibling()

		Author:		Tim Browse
		Created:	7/1/97
		Returns:	A pointer to the last sibling to the right of this one.
		Purpose:	Find the last sibling node in the tree.

	******************************************************************************/

	TreeNode* TreeNode::GetLastSibling()
	{
		// Need to find the last child
		TreeNode* pNode = this;
		TreeNode* pLastNode = NULL;

		// keep walking through the nodes until we reach the end of the sibling list
		while (pNode!=NULL)
		{
			// remember this node
			pLastNode = pNode;

			// Get the next node in sibling list
			pNode = pNode->GetNext();
		}

		// return the last node in this sibling list
		return pLastNode;
	}




	/******************************************************************************

	>	TreeNode* TreeNode::GetFirstChild() const

		Author:		Tim Browse
		Created:	7/1/97
		Returns:	A pointer to the first child node of this node, or NULL if this
					node has no children.
		Purpose:	Find the first child from this node in the tree

	******************************************************************************/

	TreeNode* TreeNode::GetFirstChild() const
	{
		return m_pFirstChild;
	}





	/******************************************************************************

	>	TreeNode* TreeNode::GetLastChild() const

		Author:		Tim Browse
		Created:	7/1/97
		Returns:	A pointer to the last child node of this node, or NULL if
					this node has no children
		Purpose:	Finds the last child from this node in the tree

	******************************************************************************/

	TreeNode* TreeNode::GetLastChild() const
	{
		// if there are no children, then there is no last child now is there
		if (m_pFirstChild==NULL)
			return NULL;

		// return the last sibling of the child
		return m_pFirstChild->GetLastSibling();
	}





	/******************************************************************************

	>	TreeNode* TreeNode::GetParent() const

		Author:		Tim Browse
		Created:	7/1/97
		Returns:	A pointer to this nodes parent, or NULL if it has no parent
					(ie, it is at the root of the tree)
		Purpose:	To find the parent of this node in the tree

	******************************************************************************/

	TreeNode* TreeNode::GetParent() const
	{
		// return the parent node
		return m_pParent;
	}



	/******************************************************************************

	>	TreeNode* TreeNode::InsertAfter(TreeNode* pContext)

		Author:		Tim Browse
		Created:	7/1/97
		Inputs:		pContext - a pointer to the node that we are to insert relative
					to.
		Returns:	A pointer to the this node after it has been inserted
		Purpose:	Inserts this node into the tree, after the context
					node passed in.

	******************************************************************************/

	TreeNode* TreeNode::InsertAfter(TreeNode* pContext)
	{
		// Make sure this node is not already inserted in any trees
		ASSERT(this!=NULL);
		ASSERT(pContext!=NULL);
		ASSERT(m_pNext==NULL);
		ASSERT(m_pPrevious==NULL);
		ASSERT(m_pParent==NULL);
		
		// Attempt to insert after pContext
		// Need to fix up the pointers of:
		// prev, next, parent, pContext->next->Prev, pContext->next

		// First get this node to point at all the things it needs to
		m_pNext		= pContext->m_pNext;
		m_pPrevious = pContext;
		m_pParent	= pContext->m_pParent;

		// Fix up the node after us, if there is one
		if (pContext->m_pNext!=NULL)
			pContext->m_pNext->m_pPrevious = this;

		// and repair the context node as well
		pContext->m_pNext = this;

		// return a pointer to myself
		return this;
	}





	/******************************************************************************

	>	TreeNode* TreeNode::InsertBefore(TreeNode* pContext)

		Author:		Tim Browse
		Created:	7/1/97
		Inputs:		pContext - a pointer to the node that we are to insert relative
					to.
		Returns:	A pointer to this node, after it has been inserted.
		Purpose:	Inserts this node into the tree, before the context node.

	******************************************************************************/

	TreeNode* TreeNode::InsertBefore(TreeNode* pContext)
	{
		// Make sure this node is not already inserted in any trees
		ASSERT(this!=NULL);
		ASSERT(pContext!=NULL);
		ASSERT(m_pNext==NULL);
		ASSERT(m_pPrevious==NULL);
		ASSERT(m_pParent==NULL);
		
		// Attempt to insert before pContext
		// Need to fix up the pointers of:
		// prev, next, parent, pContext->Prev->next, pContext->prev

		// First get this node to point at all the things it needs to
		m_pNext		= pContext;
		m_pPrevious = pContext->m_pPrevious;
		m_pParent	= pContext->m_pParent;

		// Fix up the node before us, if there is one
		if (pContext->m_pPrevious!=NULL)
			pContext->m_pPrevious->m_pNext = this;

		// and repair the context node as well
		pContext->m_pPrevious = this;

		// see if the context node has a parent
		if (pContext->m_pParent!=NULL)
		{
			// it does, so if the context node is the first child
			// we need to make a change
			if (pContext->m_pParent->m_pFirstChild == pContext)
				pContext->m_pParent->m_pFirstChild = this;
		}

		// return a pointer to myself
		return this;
	}




	/******************************************************************************

	>	TreeNode* TreeNode::InsertLast(TreeNode* pContext)

		Author:		Tim Browse
		Created:	7/1/97
		Inputs:		pContext - a pointer to the node that we are to insert relative
					to.
		Returns:	A pointer to this node after it has been inserted into the tree
		Purpose:	To insert this node as the very last sibling relative to the
					context node.

	******************************************************************************/

	TreeNode* TreeNode::InsertLast(TreeNode* pContext)
	{
		// Make sure this node is not already inserted in any trees
		ASSERT(this!=NULL);
		ASSERT(pContext!=NULL);
		ASSERT(m_pNext==NULL);
		ASSERT(m_pPrevious==NULL);
		ASSERT(m_pParent==NULL);

		// First we need to find the last node
		TreeNode* pNode = pContext->GetLastSibling();

		// Insert the node after the last node
		return InsertAfter(pNode);
	}




	/******************************************************************************

	>	TreeNode* TreeNode::InsertFirstChild(TreeNode* pContext)

		Author:		Tim Browse
		Created:	7/1/97
		Inputs:		pContext - a pointer to the node that we are to insert relative
					to.
		Returns:	A pointer to this node after it has been inserted into the tree
		Purpose:	To insert this node into the tree, as the first child of the
					context node.

	******************************************************************************/

	TreeNode* TreeNode::InsertFirstChild(TreeNode* pContext)
	{
		// Make sure this node is not already inserted in any trees
		ASSERT(this!=NULL);
		ASSERT(pContext!=NULL);
		ASSERT(m_pNext==NULL);
		ASSERT(m_pPrevious==NULL);
		ASSERT(m_pParent==NULL);

		// links effected:
		// pContext->child, pContext->Child->prev
		// prev, next, parent

		// Fix up the pointers in this node first
		m_pParent = pContext;
		m_pPrevious = NULL;
		m_pNext = pContext->m_pFirstChild;

		// and fix the pointers of the existing child node
		if (pContext->m_pFirstChild!=NULL)
			pContext->m_pFirstChild->m_pPrevious = this;

		// as well as the context node
		pContext->m_pFirstChild = this;

		// return a pointer to myself
		return this;
	}





	/******************************************************************************

	>	TreeNode* TreeNode::InsertLastChild(TreeNode* pContext)

		Author:		Tim Browse
		Created:	7/1/97
		Inputs:		pContext - a pointer to the node that we are to insert relative
					to.
		Returns:	A pointer to this node after it has been inserted into the tree
		Purpose:	To insert this node as the last child of the context node.

	******************************************************************************/

	TreeNode* TreeNode::InsertLastChild(TreeNode* pContext)
	{
		// Make sure this node is not already inserted in any trees
		ASSERT(this!=NULL);
		ASSERT(pContext!=NULL);
		ASSERT(m_pNext==NULL);
		ASSERT(m_pPrevious==NULL);
		ASSERT(m_pParent==NULL);

		// set up a return value to a default NULL pointer
		TreeNode* pReturnNode = NULL;

		// See if there is a child node yet
		if (pContext->m_pFirstChild==NULL)
		{
			// there is no child, so add this node as the first (and only) child
			pReturnNode = InsertFirstChild(pContext);
		}
		else
		{
			// There are children of the context node, so find the last one
			TreeNode* pNode = pContext->m_pFirstChild->GetLastSibling();
			ASSERT(pNode!=NULL);
			
			// Insert after this last child
			pReturnNode = InsertAfter(pNode);
		}

		// return whatever we got back
		return pReturnNode;
	}



	/******************************************************************************

	>	TreeNode* TreeNode::InsertAt(TreeNode* pContext, TTreeContext Context)

		Author:		Tim Browse
		Created:	16/5/99
		Inputs:		pContext, Context - where to insert the node.
		Returns:	A pointer to this node after it has been added to the tree
		Purpose:	To insert a node at a particular node and context psotion.

	******************************************************************************/

	TreeNode* TreeNode::InsertAt(TreeNode* pContext, TTreeContext Context)
	{
		// Simple mapping onto auxiliary functions.
		switch (Context)
		{
			case TC_InsertBefore:
				return InsertBefore(pContext);

			case TC_InsertAfter:
				return InsertAfter(pContext);

			case TC_InsertAsFirstChild:
				return InsertFirstChild(pContext);

			case TC_InsertAsLastChild:
				return InsertLastChild(pContext);
		}

		// If we get this far, it's gone wrong
		vosERROR();
		return NULL;
	}


	/******************************************************************************

	>	TreeNode* TreeNode::RemoveSubTree()

		Author:		Tim Browse
		Created:	7/1/97
		Returns:	A pointer to this node after it has been removed from the tree
		Purpose:	To remove a node and its children from the main tree. None
					of the nodes in the sub tree are deleted, they are simply
					no longer part of the main tree.

	******************************************************************************/

	TreeNode* TreeNode::RemoveSubTree()
	{
		// Make sure that this node is valid
		ASSERT(this!=NULL);

		// If there is a previous node, then fix it up
		if (m_pPrevious!=NULL)
			m_pPrevious->m_pNext = m_pNext;
		
		// If there is a next node, fix its previous pointer
		if (m_pNext!=NULL)
			m_pNext->m_pPrevious = m_pPrevious;
		
		// See if this node has a parent
		if (m_pParent!=NULL)
		{
			// it does, so check to if this is the first child of the parent
			if (m_pParent->m_pFirstChild == this)
			{
				// yes it is, so get it to point at this node next
				m_pParent->m_pFirstChild = m_pNext;
			}
		}

		// By here, all the nodes that surround us are fixed up.
		// We just need to reset all the pointer in this node
		m_pNext = NULL;
		m_pPrevious = NULL;
		m_pParent = NULL;

		// returns a pointer to the now removed node
		return this;
	}





	/******************************************************************************

	>	void TreeNode::DeleteSubTree()

		Author:		Tim Browse
		Created:	7/1/97
		Purpose:	Deletes a sub tree. The sub tree is first removed, and then
					deleted. This ultimatly performs a 'delete this' on this node

	******************************************************************************/

	void TreeNode::DeleteSubTree()
	{
		// Make sure that the subtree has been removed from the main tree
		TreeNode* pNode = RemoveSubTree();

		// keep deleting the children until there are none left
		while (pNode->m_pFirstChild!=NULL)
			pNode->m_pFirstChild->DeleteSubTree();

		// delete the now floating node
		delete pNode;		
	}





	/******************************************************************************

	>	TreeNode* TreeNode::ReplaceWith(TreeNode* pReplacementObject)

		Author:		Tim Browse
		Created:	7/1/97
		Inputs:		pReplacementObject - The object to replace this item with
		Returns:	A pointer to this item, after it has been pulled out the tree,
					or NULL if the object could not be replaced.
		Purpose:	This function replaces 'this' item with the item passed in.
					All the links around this item are repaired.

	******************************************************************************/

	TreeNode* TreeNode::ReplaceWith(TreeNode* pReplacementObject)
	{
		// See if we were passed rubbish
		if (pReplacementObject==NULL)
			return NULL;

		// Get the replacement object to point at all the things that this object points at
		pReplacementObject->m_pParent = m_pParent;
		pReplacementObject->m_pNext = m_pNext;
		pReplacementObject->m_pPrevious = m_pPrevious;
		pReplacementObject->m_pFirstChild = m_pFirstChild;

		// Fix all the objects around this item, so that they point at the replacement
		if ((m_pParent!=NULL) && (m_pParent->m_pFirstChild==this))
			m_pParent->m_pFirstChild = pReplacementObject;

		if (m_pNext!=NULL)
			m_pNext->m_pPrevious = pReplacementObject;

		if (m_pPrevious!=NULL)
			m_pPrevious->m_pNext = pReplacementObject;

		// If this has children we have to change the parent pointer for all of them
		TreeNode* pItem = m_pFirstChild;
		while (pItem!=NULL)
		{
			pItem->m_pParent = pReplacementObject;
			pItem = pItem->m_pNext;
		}

		// Fix this item so that it points at nothing
		m_pParent = NULL;
		m_pNext = NULL;
		m_pPrevious = NULL;
		m_pFirstChild = NULL;

		// return the, now seperated, object
		return this;
	}






	/******************************************************************************

	>	void TreeNode::DumpTree()

		Author:		Tim Browse
		Created:	7/1/97
		Purpose:	This function does nothing in release builds. In debug builds
					it dumps out a copy of the tree to the debugging terminal.

	******************************************************************************/

	void TreeNode::DumpTree(int Nesting)
	{
		// go into a loop for all the siblings from this point
		TreeNode* pNode = this;

		// loop through all the siblings of this node
		while (pNode!=NULL)
		{
			// indent the line according to the tree depth
			for (int i=0; i<Nesting; i++)
			{
			//	TRACE("  ");
			}

			// Display a description of this one
	//		pNode->Dump();

			// If this has a child, go and display that subtree
			if (pNode->m_pFirstChild!=NULL)
				pNode->m_pFirstChild->DumpTree(Nesting+1);

			// move on to the next sibling
			pNode = pNode->GetNext();
		}
	}




	/******************************************************************************

	>	TreeNode *TreeNode::FindRoot()

		Author:		Tim Browse
		Created:	24/1/97
		Purpose:	Find the root node of the tree, which should be a 
					CProjectItem object.

	******************************************************************************/

	TreeNode *TreeNode::FindRoot()
	{
		TreeNode *pRoot = this;

		while (pRoot->GetParent() != NULL)
		{
			pRoot = pRoot->GetParent();
		}

		// Sanity check
		ASSERT(pRoot != NULL);
		return pRoot;
	}





	/******************************************************************************

	>	TreeNode* TreeNode::FindNextDepthFirst(TreeNode* pRootItem)

		Author:		Tim Browse
		Created:	10/2/97
		Inputs:		pRootItem - The item that is the root of the sub tree that
					we are searching
		Purpose:	Finds the next item in the tree in depth first order.
					This function assumes that you have started the search from
					pRootItem. If you keep called FindNextDepthFirst() it will
					return all the nodes that are children of pRootItem and then
					NULL. Here is how you would use it
						TreeNode* pRootItem = ...
						TreeNode* pItem = pRootItem->GetFirstChild();
						while (pItem!=NULL)
						{
							// Do something

							pItem = pItem->FindNextDepthFirst(pRootItem);
						}

	******************************************************************************/

	TreeNode* TreeNode::FindNextDepthFirst(TreeNode* pRootItem)
	{
		// sanity checks
		ASSERT(pRootItem!=NULL);
		ASSERT(this!=NULL);

		// If it has any children, return the first one
		if (m_pFirstChild!=NULL)
			return m_pFirstChild;

		// No children, so see if there is a node that comes after it
		TreeNode* pItem = this;
		while (pItem!=NULL)
		{
			// check that we have not made it back to where we started
			if (pItem == pRootItem)
				return NULL;

			// See if it has something after it.
			// if it does, return that
			if (pItem->m_pNext!=NULL)
				return pItem->m_pNext;

			// Nope, so go back up to its parent
			pItem = pItem->m_pParent;
		}

		// We fell out the top of the tree without ever reaching the rootitem
		vosERROR();
		return NULL;
	}


	/******************************************************************************

	>	TreeNode* TreeNode::FindNextDepthFirstReverse(TreeNode* pRootItem)

		Author:		Tim Browse
		Created:	10/2/97
		Inputs:		pRootItem - The item that is the root of the sub tree that
								we are searching.
		Purpose:	Finds the next item in the tree in *reverse* depth first order.

					This function assumes that you have started the search from
					the rightmost child of pRootItem. If you keep calling 
					FindNextDepthFirst() it will return all the nodes that are 
					children of pRootItem, then pRootItem itself, and then finally
					NULL.
					
					  Here is how you would use it:

						TreeNode* pRootItem = ...
						TreeNode* pItem;
						pRootItem->GetRightmostChild(pItem);

						while (pItem!=NULL)
						{
							// Do something

							pItem = pItem->FindNextDepthFirstReverse(pRootItem);
						}

	******************************************************************************/

	TreeNode* TreeNode::FindNextDepthFirstReverse(TreeNode* pRootItem)
	{
		// sanity checks
		ASSERT(pRootItem!=NULL);
		ASSERT(this!=NULL);

		// Have we finished?
		if (this == pRootItem)
		{
			// Yes - we've reached the root.
			return NULL;
		}

		// If it has a previous sibling, return the rightmost child of that sibling
		if (m_pPrevious != NULL)
		{
			TreeNode *pItem = m_pPrevious;

			// Find rightmost child - if there is none, we just use this previous sibling.
			pItem->GetRightmostChild(pItem);
			return pItem;
		}

		// Otherwise, return the parent.
		return m_pParent;
	}

	/**************************************************************************
	
	>	bool TreeNode::GetRightmostChild(TreeNode*& pRightmostChild)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Outputs:	pRightmostChild - the rightmost child of the node
		Returns:	true => rightmost child found
					false => node has no children
	
		Purpose:	Get the last sibling (rightmost child) of the node
	
	**************************************************************************/
	
	bool TreeNode::GetRightmostChild(TreeNode*& pRightmostChild)
	{
		TreeNode *pItem = this;

		TreeNode* pLastChild;

		do
		{
			// Go down the rightmost branch one level, if we can.
			pLastChild = pItem->GetLastChild();

			if (pLastChild != NULL)
			{
				pItem = pLastChild;
			}

			// Terminate when we reach the rightmost child.
		}
		while (pLastChild != NULL);

		if (pItem == this)
		{
			// No child found
			return false;
		}
		else
		{
			// Found a child
			pRightmostChild = pItem;
			return true;
		}
	}


	/**************************************************************************
	
	>	bool TreeNode::IsDescendedFrom(TreeNode *pAncestor)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		pAncestor - the possible ancestor to check for
		Returns:	true => this node is descended from pAncestor
					false => otherwise
	
		Purpose:	Determine if this node is descended from another.
	
	**************************************************************************/
	
	bool TreeNode::IsDescendedFrom(TreeNode *pAncestor)
	{
		// Simple check
		if (pAncestor == NULL)
			return false;

		// Go up the parent links, searching for ancestor
		TreeNode *pParent = GetParent();
		while (pParent != NULL)
		{
			if (pParent == pAncestor)
				return true;

			pParent = pParent->GetParent();
		}

		// It was not found
		return false;
	}
};

