//
// VolumeSpec.h
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#ifndef vos_VolumeSpec_h
#define vos_VolumeSpec_h

#include "vos.h"

#include "String.h"
#include "ErrorManager.h"

//
// Lose annoying Win32 macro
//
#ifdef VOS_MSWINDOWS
#undef GetFreeSpace
#endif

namespace vos
{
	/**************************************************************************

		class VolumeSpec

		Author:		Tim Browse

		Date:		21/11/99

		Purpose:	The VolumeSpec class is used to provide a platform 
					independent way of specifying a disc volume object.

					On most platforms this is simply a string.

					Strings and VolumeSpecs
					~~~~~~~~~~~~~~~~~~~~~~~

					All platforms must support the flattening of a VolumeSpec
					object into a String - and the reversal of this, which
					is to convert a String into a VolumeSpec object.

					If you convert VolumeSpec V into a string object S, and
					then initialise another VolumeSpec object, V' using the
					string S, then the two VolumeSpec objects V and V' *must*
					refer to the same disc volume object.

					When converting between VolumeSpecs and String objects, 
					you use the SetFromString() and GetString() functions.

					The most common way of creating a VolumeSpec object is to
					initialise it from a FileSpec object - the FileSpec's
					volume root is extracted (if the FileSpec is relative, 
					the volume spec is initialised to a null value).

	**************************************************************************/

	class VolumeSpec
	{
		public:
			VolumeSpec() {}
			VolumeSpec(const String& Pathname);
			VolumeSpec(const FileSpec& aFileSpec);

			// Standard comparison operators.
			bool operator==(const VolumeSpec& rhs) const;
			bool operator!=(const VolumeSpec& rhs) const;

			// Is the VolumeSpec empty (null) ?
			bool IsEmpty() const;

			// Convert between VolumeSpec objects and String objects.
			String GetString() const;
			void SetFromString(const String& Filename);

			// Disc free space functions
			Error GetFreeSpace(uint *pBytesFree);
			Error GetAllocationUnitSize(uint *pClusterSize);

			// User-friendly form of volume name.
			void GetDisplayName(String* pDisplayName);

		private:
			bool DecodeVolumeName(const String& Pathname);

		private:
#ifdef VOS_MACOS
			int32 m_VRefNum;
#endif
			// Voulume name string
			String m_VolumeName;


		private:
#ifdef VOS_MSWINDOWS
			bool IsUNC() const;
			Error GetWin32FreeDiskSpace(uint *pBytesFree, uint *pClusterSize);
#endif
	};
};


#endif // !vos_VolumeSpec_h
