//
// ZipCatalogueReader.cpp
//

#include "StdHdr.h"

#include "ZipCatalogueReader.h"

#include "ZipLayout.h"

#include "FileSpec.h"
#include "Errors.h"

namespace vos
{
	
	/**************************************************************************

	>	class ZipFileInfo

		Author:		Tim Browse

		Date:		07/06/2001

		Purpose:	Simple data structure used to store information about a 
					file in the ZIP catalogue.

	**************************************************************************/

	class ZipFileInfo
	{
		public:
			// Constructor
			ZipFileInfo() : m_cbFileHeaderOffset(0), m_cbFileSize(0), m_CRC32(0) {}

			// Name of this file
			vos::FileSpec m_FS;

			// Offset into the ZIP file where this file's header can be found.
			uint m_cbFileHeaderOffset;

			// The size/position of the file's data is not always available
			// immediately - if less than 0, these values are not available yet.
			int m_cbFileSize;
			int m_cbFileDataPos;

			// CRC32 value for the file
			uint32 m_CRC32;
	};

	
	/**************************************************************************

	>	ZipCatalogueReader::ZipCatalogueReader()

		Author:		Tim Browse
		Date:		07/06/2001

		Purpose:	Constructor

	**************************************************************************/

	ZipCatalogueReader::ZipCatalogueReader() : m_cMaxFiles(0),
											   m_cFiles(0), 
											   m_pFiles(NULL)
	{
		// TCILB
	}

	
	/**************************************************************************

	>	ZipCatalogueReader::~ZipCatalogueReader()

		Author:		Tim Browse
		Date:		07/06/2001

		Purpose:	Destructor

	**************************************************************************/

	ZipCatalogueReader::~ZipCatalogueReader()
	{
		delete [] m_pFiles;
	}

	/**************************************************************************

	>	bool ZipCatalogueReader::AllocateCatalogue(int cFiles)

		Author:		Tim Browse
		Date:		07/06/2001
		Inputs:		cFiles - the maximum number of files that will be
							 placed in this catalogue.

		Returns:	true => catalogue created successfully
					false => ran out of memory

		Purpose:	Allocate buffer for the catalogue.

					NB. You can do this only once - you can't extend the
						catalogue with more objects - it's a static data 
						structure.

	**************************************************************************/

	bool ZipCatalogueReader::AllocateCatalogue(int cFiles)
	{
		// Allocate enough file info objects
		m_cMaxFiles = cFiles;
		m_pFiles = new ZipFileInfo[m_cMaxFiles];

		// TODO: check for memory failures
		return true;
	}

	
	/**************************************************************************

	>	void ZipCatalogueReader::AddFile(char *Filename,
									     const RawZipFileHeader& FileHeader)

		Author:		Tim Browse
		Date:		07/06/2001
		Inputs:		Filename - filename of the file to add to the
							   catalogue (does not need to be zero-terminated).
					FileHeader - header info for this file.

		Purpose:	Add a file's details to the ZIP catalogue.

					The filename is not specified as a vos::FileSpec, as this
					class is used to load data direct from a ZIP catalogue,
					and filenames are stored as non-0-terminated character
					arrays with a length parameter.

	**************************************************************************/

	void ZipCatalogueReader::AddFile(char *Filename,
								     const RawZipFileHeader& FileHeader)
	{
		// Check we haven't hit the file limit
		if (m_cFiles >= m_cMaxFiles)
		{
			COMPLAIN("Added too many files to zip catalogue");
			return;
		}

		//
		// Copy details into array
		//

		// First, set up the filename field.
		//
		// NB. ZIP files *always* use '/' as folder separator, 
		//	   regardless of platform (or at least, that's the spec!)
		//
		vos::String strFilename(Filename, FileHeader.m_FilenameLength);
		m_pFiles[m_cFiles].m_FS.SetFromString(strFilename, '/');

		// Add position info for file
		m_pFiles[m_cFiles].m_cbFileHeaderOffset = FileHeader.m_LocalHeaderRelativeOffset;

		// Fill in size of file, if available.
		if (FileHeader.m_GPBits & BIT(3))
		{
			// File size info not available in directory.
			m_pFiles[m_cFiles].m_cbFileSize = -1;
		}
		else
		{
			// File size info is present in this header, so use it.
			m_pFiles[m_cFiles].m_cbFileSize = FileHeader.m_UncompressedSize;
		}

		//
		// Ok, work out where the actual data for the file starts.
		//

		// This gives us the position of the local file header.
		uint cbDataOffset = FileHeader.m_LocalHeaderRelativeOffset;

		// Now move past the standard local file header
		cbDataOffset += sizeof(RawZipLocalFileHeader);

		// And take into account the variable length fields stored after
		// the local file header.
		cbDataOffset += FileHeader.m_FilenameLength;
		cbDataOffset += FileHeader.m_ExtraFieldLength;

		// So now cbDataOffset is the offset to the first byte of data in this
		// file...
		m_pFiles[m_cFiles].m_cbFileDataPos = cbDataOffset;

		// Copy CRC32 value
		m_pFiles[m_cFiles].m_CRC32 = FileHeader.m_CRC32;

		// This file's info is complete
		m_cFiles++;
	}


	/**************************************************************************

	>	bool ZipCatalogueReader::FindFile(const vos::FileSpec& FS,
										  uint *pFileOffset,
										  uint *pFileLength,
										  uint32 *pCRC32 = NULL)

		Author:		Tim Browse
		Date:		07/06/2001
		Inputs:		FS - name of the file to find.
		Outputs:	pFileOffset - position of the file's data within the
								  ZIP file (as a byte offset).
					pFileLength - length of the file's data (in bytes).
					pCRC32 - if non-NULL, will be filled in with the file's
							 CRC-32 value from the ZIP catalogue.
		Returns:	true => File was found; offset and length are filled in.
					false => File not found.

		Purpose:	Find the specified file in this ZIP catalogue.

	**************************************************************************/

	bool ZipCatalogueReader::FindFile(const vos::FileSpec& FS,
									  uint *pFileOffset,
									  uint *pFileLength,
									  uint32 *pCRC32)
	{
		for (int iFile = 0; iFile < m_cMaxFiles; iFile++)
		{
			if (m_pFiles[iFile].m_FS == FS)
			{
				// Found the file
				*pFileOffset = m_pFiles[iFile].m_cbFileDataPos;
				*pFileLength = m_pFiles[iFile].m_cbFileSize;

				if (pCRC32 != NULL)
					*pCRC32 = m_pFiles[iFile].m_CRC32;

				return true;
			}
		}

		// File not found
		return false;
	}

	/**************************************************************************

	>	void ZipCatalogueReader::GetFilename(uint iFile, vos::FileSpec* pFileSpec)

		Author:		Tim Browse
		Date:		07/06/2001
		Inputs:		iFile - 0-based index of the file in question
		Outputs:	pFileSpec - filename of the specified file.

		Purpose:	Obtain the filename of the n'th file in the catalogue.

	**************************************************************************/

	void ZipCatalogueReader::GetFilename(uint iFile, vos::FileSpec* pFileSpec)
	{
		if (iFile < m_cMaxFiles)
		{
			*pFileSpec = m_pFiles[iFile].m_FS;
		}
		else
		{
			COMPLAIN("Illegal file index in ZipCatalogueReader::GetFilename()");
		}
	}

	/**************************************************************************

	>	bool ZipCatalogueReader::GetFile(uint iFile,
										 uint *pFileOffset,
										 uint *pFileLength,
										 uint32 *pCRC32)

		Author:		Tim Browse
		Date:		07/06/2001
		Outputs:	pFileOffset - position of the file's data within the
								  ZIP file (as a byte offset).
					pFileLength - length of the file's data (in bytes).
					pCRC32 - if non-NULL, will be filled in with the file's
							 CRC-32 value from the ZIP catalogue.
		Returns:	true => File was found; offset and length are filled in.
					false => File not found.

		Purpose:	Get catalogue information for the n'th file in the
					catalogue.

	**************************************************************************/

	bool ZipCatalogueReader::GetFile(uint iFile,
									 uint *pFileOffset,
									 uint *pFileLength,
									 uint32 *pCRC32)
	{
		if (iFile < m_cMaxFiles)
		{
			*pFileOffset = m_pFiles[iFile].m_cbFileDataPos;
			*pFileLength = m_pFiles[iFile].m_cbFileSize;

			if (pCRC32 != NULL)
				*pCRC32 = m_pFiles[iFile].m_CRC32;

			return true;
		}
		else
		{
			COMPLAIN("Illegal file index in ZipCatalogueReader::GetFilename()");
			return false;
		}
	}

	/**************************************************************************

	>	bool ZipCatalogueReader::FindFile(const vos::FileSpec& FS, uint* piFile)

		Author:		Tim Browse
		Date:		07/06/2001
		Inputs:		FS - name of the file to look for.
		Outputs:	piFile - the 0-based index of the file that was found.
		Returns:	true => file found;
					false => file not found.

		Purpose:	Search for a named file in the catalogue, and obtain the
					(0-based) index of the file's position in the catalogue.

	**************************************************************************/

	bool ZipCatalogueReader::FindFile(const vos::FileSpec& FS, uint* piFile)
	{
		for (int iFile = 0; iFile < m_cMaxFiles; iFile++)
		{
			if (m_pFiles[iFile].m_FS == FS)
			{
				// Found the file
				*piFile = iFile;
				return true;
			}
		}

		// File not found
		return false;
	}

}

