//
// ZipLayout.cpp
//

#include "StdHdr.h"

#include "ZipLayout.h"

#include "Endian.h"
#include "File.h"

namespace vos
{
	/**************************************************************************

	>	void RawZipDataDescriptor::EndianNormalise()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Convert the little-endian data loaded from a ZIP file
					to the host's byte order.

	**************************************************************************/

	void RawZipDataDescriptor::EndianNormalise()
	{
		//
		// Swap data fields from little-endian to native-endian...
		//
		m_CRC32				= vos::EndianSwap_LtoN_uint32(m_CRC32);
		m_CompressedSize	= vos::EndianSwap_LtoN_uint32(m_CompressedSize);
		m_UncompressedSize	= vos::EndianSwap_LtoN_uint32(m_UncompressedSize);
	}

	/**************************************************************************

	>	vos::Error ZipDataDescriptor::Load(vos::File* pFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFile - the (ZIP) file to load the data from.
		Returns:	Error

		Purpose:	Load the data for this structure from the ZIP file.

	**************************************************************************/

	vos::Error ZipDataDescriptor::Load(vos::File* pFile)
	{
		if (pFile->Read(this, sizeof(RawZipDataDescriptor)))
		{
			EndianNormalise();
			return VNOERROR;
		}
		else
		{
			// Failed to load data descriptor
			return LASTERROR;
		}
	}

	/**************************************************************************

	>	void RawZipLocalFileHeader::EndianNormalise()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Convert the little-endian data loaded from a ZIP file
					to the host's byte order.

	**************************************************************************/

	void RawZipLocalFileHeader::EndianNormalise()
	{
		//
		// Swap data fields from little-endian to native-endian...
		//
		m_Signature					= vos::EndianSwap_LtoN_uint32(m_Signature);
		m_VersionNeededToExtract	= vos::EndianSwap_LtoN_uint16(m_VersionNeededToExtract);
		m_GPBits					= vos::EndianSwap_LtoN_uint16(m_GPBits);
		m_CompressionMethod			= vos::EndianSwap_LtoN_uint16(m_CompressionMethod);
		m_ModifiedFileTime			= vos::EndianSwap_LtoN_uint16(m_ModifiedFileTime);
		m_ModifiedFileDate			= vos::EndianSwap_LtoN_uint16(m_ModifiedFileDate);
		m_CRC32						= vos::EndianSwap_LtoN_uint32(m_CRC32);
		m_CompressedSize			= vos::EndianSwap_LtoN_uint32(m_CompressedSize);
		m_UncompressedSize			= vos::EndianSwap_LtoN_uint32(m_UncompressedSize);
		m_FilenameLength			= vos::EndianSwap_LtoN_uint16(m_FilenameLength);
		m_ExtraFieldLength			= vos::EndianSwap_LtoN_uint16(m_ExtraFieldLength);
	}

	/**************************************************************************

	>	ZipLocalFileHeader::ZipLocalFileHeader()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Constructor

	**************************************************************************/

	ZipLocalFileHeader::ZipLocalFileHeader() : m_pFilename(NULL),
											   m_pExtraField(NULL)
	{
		m_Signature					= kLocalFileHeaderSignature;
		m_VersionNeededToExtract	= 0;
		m_GPBits					= 0;
		m_CompressionMethod			= 0;
		m_ModifiedFileTime			= 0;
		m_ModifiedFileDate			= 0;
		m_CRC32						= 0;
		m_CompressedSize			= 0;
		m_UncompressedSize			= 0;
		m_FilenameLength			= 0;
		m_ExtraFieldLength			= 0;
	}

	/**************************************************************************

	>	ZipLocalFileHeader::~ZipLocalFileHeader()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Destructor

	**************************************************************************/

	ZipLocalFileHeader::~ZipLocalFileHeader()
	{
		CleanUp();
	}

	/**************************************************************************

	>	void ZipLocalFileHeader::SetFilename(const vos::String& Filename)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		Filename - the new filename to set for the local file header.

		Purpose:	Set the filename for the local file header.

	**************************************************************************/

	void ZipLocalFileHeader::SetFilename(const vos::String& Filename)
	{
		m_FilenameLength = Filename.GetLength();
		m_pFilename = new char[m_FilenameLength];
		memcpy(m_pFilename, Filename.c_str(), m_FilenameLength);
	}


	/**************************************************************************

	>	void ZipLocalFileHeader::CleanUp()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Clean up dynamic data for this object.

	**************************************************************************/

	void ZipLocalFileHeader::CleanUp()
	{
		delete [] m_pFilename;
		m_pFilename = NULL;
		delete [] m_pExtraField;
		m_pExtraField = NULL;
	}

	/**************************************************************************

	>	vos::Error ZipLocalFileHeader::Load(vos::File* pFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFile - the (ZIP) file to load the data from.
		Returns:	Error

		Purpose:	Load the data for this structure from the ZIP file.

	**************************************************************************/

	vos::Error ZipLocalFileHeader::Load(vos::File* pFile)
	{
		// Clear up any dynamic data before we start
		CleanUp();

		if (pFile->Read(this, sizeof(RawZipLocalFileHeader)))
		{
			EndianNormalise();

			// Load the filename (which follows the local file header)
			m_pFilename = new char[m_FilenameLength];
			if (!pFile->Read(m_pFilename, m_FilenameLength))
				return LASTERROR;

			// Load the extra field.
			if (m_ExtraFieldLength > 0)
			{
				m_pExtraField = new byte[m_ExtraFieldLength];
				if (!pFile->Read(m_pExtraField, m_ExtraFieldLength))
					return LASTERROR;
			}

			return VNOERROR;
		}
		else
		{
			// Failed to load data descriptor
			return LASTERROR;
		}
	}


	/**************************************************************************

	>	vos::Error ZipLocalFileHeader::Write(vos::File* pFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFile - the ZIP file to write the data to.
		Returns:	Error

		Purpose:	Write the local file header to the specified ZIP file, 
					according to the ZIP file format.

					NB! Only works on little-endian hosts at the moment.
					~~

	**************************************************************************/

	vos::Error ZipLocalFileHeader::Write(vos::File* pFile)
	{
		// TODO: this will not work on big-endian machines - revise vos::File
		//       to handle endian file access properly, so we can use Write_uint16()
		//		 et. al.
		if (pFile->Write(this, sizeof(RawZipLocalFileHeader)))
		{
			// Write out the filename and extra fields, if they exist.
			if (m_FilenameLength > 0)
			{
				if (!pFile->Write(m_pFilename, m_FilenameLength))
					return LASTERROR;
			}

			if (m_ExtraFieldLength > 0)
			{
				if (!pFile->Write(m_pExtraField, m_ExtraFieldLength))
					return LASTERROR;
			}

			return VNOERROR;
		}
		else
		{
			return LASTERROR;
		}
	}

	/**************************************************************************

	>	void RawZipFileHeader::EndianNormalise()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Convert the little-endian data loaded from a ZIP file
					to the host's byte order.

	**************************************************************************/

	void RawZipFileHeader::EndianNormalise()
	{
		//
		// Swap data fields from little-endian to native-endian...
		//
		m_Signature					= vos::EndianSwap_LtoN_uint32(m_Signature);
		m_VersionMadeBy				= vos::EndianSwap_LtoN_uint16(m_VersionMadeBy);
		m_VersionNeededToExtract	= vos::EndianSwap_LtoN_uint16(m_VersionNeededToExtract);
		m_GPBits					= vos::EndianSwap_LtoN_uint16(m_GPBits);
		m_CompressionMethod			= vos::EndianSwap_LtoN_uint16(m_CompressionMethod);
		m_ModifiedFileTime			= vos::EndianSwap_LtoN_uint16(m_ModifiedFileTime);
		m_ModifiedFileDate			= vos::EndianSwap_LtoN_uint16(m_ModifiedFileDate);
		m_CRC32						= vos::EndianSwap_LtoN_uint32(m_CRC32);
		m_CompressedSize			= vos::EndianSwap_LtoN_uint32(m_CompressedSize);
		m_UncompressedSize			= vos::EndianSwap_LtoN_uint32(m_UncompressedSize);
		m_FilenameLength			= vos::EndianSwap_LtoN_uint16(m_FilenameLength);
		m_ExtraFieldLength			= vos::EndianSwap_LtoN_uint16(m_ExtraFieldLength);
		m_FileCommentLength			= vos::EndianSwap_LtoN_uint16(m_FileCommentLength);
		m_DiskNumberStart			= vos::EndianSwap_LtoN_uint16(m_DiskNumberStart);
		m_InternalFileAttributes	= vos::EndianSwap_LtoN_uint16(m_InternalFileAttributes);
		m_ExternalFileAttributes	= vos::EndianSwap_LtoN_uint32(m_ExternalFileAttributes);
		m_LocalHeaderRelativeOffset	= vos::EndianSwap_LtoN_uint32(m_LocalHeaderRelativeOffset);
	}

	/**************************************************************************

	>	ZipFileHeader::ZipFileHeader()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Constructor

	**************************************************************************/

	ZipFileHeader::ZipFileHeader() : m_pFilename(NULL),
								     m_pExtraField(NULL),
									 m_pFileComment(NULL)
	{
		// TCILB
	}

	/**************************************************************************

	>	ZipFileHeader::~ZipFileHeader()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Destructor

	**************************************************************************/

	ZipFileHeader::~ZipFileHeader()
	{
		CleanUp();
	}

	/**************************************************************************

	>	void ZipFileHeader::Init(const ZipLocalFileHeader&rhs)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		rhs - the local file header to use as initialisation data.

		Purpose:	Initialise a File Header from a Local File Header structure.

	**************************************************************************/

	void ZipFileHeader::Init(const ZipLocalFileHeader&rhs)
	{
		m_Signature					= kCentralFileHeaderSignature;

		// Double up.
		m_VersionNeededToExtract	= rhs.m_VersionNeededToExtract;
		m_VersionMadeBy				= rhs.m_VersionNeededToExtract;

		m_GPBits					= rhs.m_GPBits;
		m_CompressionMethod			= rhs.m_CompressionMethod;
		m_ModifiedFileTime			= rhs.m_ModifiedFileTime;
		m_ModifiedFileDate			= rhs.m_ModifiedFileDate;
		m_CRC32						= rhs.m_CRC32;
		m_CompressedSize			= rhs.m_CompressedSize;
		m_UncompressedSize			= rhs.m_UncompressedSize;
		m_FilenameLength			= rhs.m_FilenameLength;
		m_ExtraFieldLength			= rhs.m_ExtraFieldLength;

		m_FileCommentLength = 0;
		m_DiskNumberStart = 0;
		m_InternalFileAttributes = 0;
		m_ExternalFileAttributes = 0;
		m_LocalHeaderRelativeOffset = 0;

		// Copy dynamic fields
		if (m_FilenameLength > 0)
		{
			m_pFilename = new char[m_FilenameLength];
			memcpy(m_pFilename, rhs.m_pFilename, m_FilenameLength);
		}

		if (m_ExtraFieldLength > 0)
		{
			m_pExtraField = new byte[m_ExtraFieldLength];
			memcpy(m_pExtraField, rhs.m_pExtraField, m_ExtraFieldLength);
		}
	}

	/**************************************************************************

	>	void ZipFileHeader::SetFilename(const vos::String& Filename)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		Filename - the filename to use.

		Purpose:	Set the dynamic filename field of the file header.

	**************************************************************************/

	void ZipFileHeader::SetFilename(const vos::String& Filename)
	{
		m_FilenameLength = Filename.GetLength();
		delete [] m_pFilename;
		m_pFilename = new char[m_FilenameLength];
		memcpy(m_pFilename, Filename.c_str(), m_FilenameLength);
	}

	/**************************************************************************

	>	void ZipFileHeader::CleanUp()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Deallocate the dynamic data fields used by this object.

	**************************************************************************/

	void ZipFileHeader::CleanUp()
	{
		delete [] m_pFilename;
		m_pFilename = NULL;
		delete [] m_pExtraField;
		m_pExtraField = NULL;
		delete [] m_pFileComment;
		m_pFileComment = NULL;
	}


	/**************************************************************************

	>	vos::Error ZipFileHeader::Load(vos::File* pFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFile - the (ZIP) file to load the data from.
		Returns:	Error

		Purpose:	Load the data for this structure from the ZIP file.

	**************************************************************************/

	vos::Error ZipFileHeader::Load(vos::File* pFile)
	{
		// Clear up any dynamic data before we start
		CleanUp();

		if (pFile->Read(this, sizeof(RawZipFileHeader)))
		{
			EndianNormalise();

			// Load the filename (which follows the local file header)
			m_pFilename = new char[m_FilenameLength];
			if (!pFile->Read(m_pFilename, m_FilenameLength))
				return LASTERROR;

			// Load the extra field.
			if (m_ExtraFieldLength > 0)
			{
				m_pExtraField = new byte[m_ExtraFieldLength];
				if (!pFile->Read(m_pExtraField, m_ExtraFieldLength))
					return LASTERROR;
			}

			// Load the file comment
			if (m_FileCommentLength > 0)
			{
				m_pFileComment = new char[m_FileCommentLength];
				if (!pFile->Read(m_pFileComment, m_FileCommentLength))
					return LASTERROR;
			}

			return VNOERROR;
		}
		else
		{
			// Failed to load zip file header
			return LASTERROR;
		}
	}

	/**************************************************************************

	>	vos::Error ZipFileHeader::Write(vos::File* pFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFile - the ZIP file to write the file header to.
		Returns:	Error

		Purpose:	Write the file header structure to the specified ZIP
					file (according to the standard ZIP file format).

					NB! Only works on little-endian hosts at the moment.
					~~

	**************************************************************************/

	vos::Error ZipFileHeader::Write(vos::File* pFile)
	{
		// TODO: this will not work on big-endian machines - revise vos::File
		//       to handle endian file access properly, so we can use Write_uint16()
		//		 et. al.
		if (pFile->Write(this, sizeof(RawZipFileHeader)))
		{
			// Write out the filename and extra fields, if they exist.
			if (m_FilenameLength > 0)
			{
				if (!pFile->Write(m_pFilename, m_FilenameLength))
					return LASTERROR;
			}

			if (m_ExtraFieldLength > 0)
			{
				if (!pFile->Write(m_pExtraField, m_ExtraFieldLength))
					return LASTERROR;
			}

			if (m_FileCommentLength > 0)
			{
				if (!pFile->Write(m_pFileComment, m_FileCommentLength))
					return LASTERROR;
			}

			return VNOERROR;
		}
		else
		{
			return LASTERROR;
		}
	}


	/**************************************************************************

	>	void RawZipEndOfCentralDir::EndianNormalise()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Convert the little-endian data loaded from a ZIP file
					to the host's byte order.

	**************************************************************************/

	void RawZipEndOfCentralDir::EndianNormalise()
	{
		//
		// Swap data fields from little-endian to native-endian...
		//
		m_Signature					= vos::EndianSwap_LtoN_uint32(m_Signature);
		m_DiskNumber				= vos::EndianSwap_LtoN_uint16(m_DiskNumber);
		m_CentralDirDiskNumber		= vos::EndianSwap_LtoN_uint16(m_CentralDirDiskNumber);
		m_NumCentralDirEntries		= vos::EndianSwap_LtoN_uint16(m_NumCentralDirEntries);
		m_TotalNumCentralDirEntries	= vos::EndianSwap_LtoN_uint16(m_TotalNumCentralDirEntries);
		m_CentralDirSize			= vos::EndianSwap_LtoN_uint32(m_CentralDirSize);
		m_CentralDirOffset			= vos::EndianSwap_LtoN_uint32(m_CentralDirOffset);
		m_ZipFileCommentLength		= vos::EndianSwap_LtoN_uint16(m_ZipFileCommentLength);
	}

	/**************************************************************************

	>	ZipEndOfCentralDir::ZipEndOfCentralDir()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Constructor

	**************************************************************************/

	ZipEndOfCentralDir::ZipEndOfCentralDir() : m_pZipFileComment(NULL)
	{
		m_Signature = kEndOfCentralDirSignature;

		m_DiskNumber = 0;
		m_CentralDirDiskNumber = 0;
		m_NumCentralDirEntries = 0;
		m_TotalNumCentralDirEntries = 0;
		m_CentralDirSize = 0;
		m_CentralDirOffset = 0;
		m_ZipFileCommentLength = 0;
	}

	/**************************************************************************

	>	ZipEndOfCentralDir::~ZipEndOfCentralDir()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Destructor

	**************************************************************************/

	ZipEndOfCentralDir::~ZipEndOfCentralDir()
	{
		CleanUp();
	}

	/**************************************************************************

	>	void ZipEndOfCentralDir::CleanUp()

		Author:		Tim Browse
		Date:		10/06/2001

		Purpose:	Deallocate any dynamic data used by this object.

	**************************************************************************/

	void ZipEndOfCentralDir::CleanUp()
	{
		delete [] m_pZipFileComment;
		m_pZipFileComment = NULL;
	}

	/**************************************************************************

	>	vos::Error ZipEndOfCentralDir::Load(vos::File* pFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFile - the (ZIP) file to load the data from.
		Returns:	Error

		Purpose:	Load the data for this structure from the ZIP file.

	**************************************************************************/

	vos::Error ZipEndOfCentralDir::Load(vos::File* pFile)
	{
		// Clear up any dynamic data before we start
		CleanUp();

		if (pFile->Read(this, sizeof(RawZipEndOfCentralDir)))
		{
			EndianNormalise();

			// Load the zip file comment
			if (m_ZipFileCommentLength > 0)
			{
				m_pZipFileComment = new char[m_ZipFileCommentLength];
				if (!pFile->Read(m_pZipFileComment, m_ZipFileCommentLength))
					return LASTERROR;
			}


			return VNOERROR;
		}
		else
		{
			// Failed to load end of central dir record
			return LASTERROR;
		}
	}

	/**************************************************************************

	>	vos::Error ZipEndOfCentralDir::Write(vos::File* pFile)

		Author:		Tim Browse
		Date:		10/06/2001
		Inputs:		pFile - the ZIP file to write the end-of-dir marker to.
		Returns:	Error

		Purpose:	Write the end of central directory marker structure to the 
					specified ZIP file (according to the standard ZIP file 
					format).

					NB! Only works on little-endian hosts at the moment.
					~~

	**************************************************************************/

	vos::Error ZipEndOfCentralDir::Write(vos::File* pFile)
	{
		// TODO: this will not work on big-endian machines - revise vos::File
		//       to handle endian file access properly, so we can use Write_uint16()
		//		 et. al.
		if (pFile->Write(this, sizeof(RawZipEndOfCentralDir)))
		{
			// Save the zip file comment
			if (m_ZipFileCommentLength > 0)
			{
				if (!pFile->Write(m_pZipFileComment, m_ZipFileCommentLength))
					return LASTERROR;
			}

			return VNOERROR;
		}
		else
		{
			// Failed to save end of central dir record
			return LASTERROR;
		}
	}
}
