//
// FileSpec.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "..\FileSpec.h"

#include "..\DiskFile.h"
#include "..\Time.h"

#include "windows.h"

// For the special registry key names.
#include "regstr.h"

// Win32 support functions.
#include "Win32Registry.h"
#include "Win32Error.h"

#include "..\Errors.h"
#include "..\FileMatches.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif


namespace vos
{
	/**************************************************************************

	>	void FileSpec::SetFromString(const String& Filename, 
									 char FolderSeparator = kStringFolderSeparator)

		Author:		Tim Browse
		Date:		20/08/99
		Inputs:		Filename - the string to convert to a FileSpec.
					FolderSeparator - the character used in 'Filename' to
									  denote separation of folders/pathname
									  components.

		Purpose:	Convert a String object into a FileSpec object.

					e.g.

						FileSpec F;
						F.SetFromString("C:\\windows");

						   - or -

						F.SetFromString("MacintoshHD:System:Extensions", ":");

						   - or -

						FileSpec::kStringFolderSeparator = '/';
						F.SetFromString("/usr/local/lib");

	**************************************************************************/

	void FileSpec::SetFromString(const String& Filename, char FolderSeparator)
	{
		// Copy string
		m_FileName = Filename;

		// Convert folder separators
		if (FolderSeparator != kStringFolderSeparator)
		{
			m_FileName.Replace(FolderSeparator, kStringFolderSeparator);
		}
	}

	/**************************************************************************

	>	void FileSpec::AppendFolderSeparator()

		Author:		Tim Browse
		Date:		20/08/99

		Purpose:	Add a folder separator onto the end of the string, unless
					there is already one there.

	**************************************************************************/

	void FileSpec::AppendFolderSeparator()
	{
		// If filespec is empty, don't add a separator
		if (m_FileName.GetLength() == 0)
			return;

		// If the filename does not end in a backslash, then add one.
		if (m_FileName[m_FileName.GetLength() - 1] != kStringFolderSeparator)
			m_FileName += kStringFolderSeparator;
	}

	/**************************************************************************

	>	void FileSpec::AppendFolder(const String& FolderName)

		Author:		Tim Browse
		Date:		20/08/99
		Inputs:		FolderName - the folder to append.

		Purpose:	Append a folder name to the FileSpec.

					e.g.

						FileSpec F;
						F.SetFromString("c:\games");

						F.AppendFolder("pong");

						String S;
						S = F.GetString();

						// S now equals "c:\games\pong"

	**************************************************************************/

	void FileSpec::AppendFolder(const String& FolderName)
	{
		// Ignore null folder names
		if (FolderName.IsEmpty())
			return;

		// Make sure we have the folder separator
		AppendFolderSeparator();

		String TheFolderName = FolderName;

		// Remove any preceding folder separators on the folder name.
		while ((TheFolderName.GetLength() > 0) &&
			   (TheFolderName[0] == kStringFolderSeparator))
		{
			TheFolderName = TheFolderName.Right(TheFolderName.GetLength() - 1);
		}

		m_FileName += TheFolderName;
	}

	/**************************************************************************

	>	void FileSpec::AppendFilename(const String& Filename)

		Author:		Tim Browse
		Date:		20/08/99
		Inputs:		Filename - the filename to append,

		Purpose:	Append a filename (leafname) to the FileSpec.

					e.g.

						FileSpec F;
						F.SetFromString("c:\data");

						F.AppendFolder("myfile.txt");

						String S;
						S = F.GetString();

						// S now equals "c:\data\myfile.txt"

	**************************************************************************/

	void FileSpec::AppendFilename(const String& Filename)
	{
		// Make sure we have the folder separator
		AppendFolderSeparator();
		m_FileName += Filename;
	}

	/**************************************************************************

	>	void FileSpec::AppendRelativeFileSpec(const FileSpec& RelativeFileSpec)

		Author:		Tim Browse
		Date:		20/08/99
		Inputs:		RelativeFileSpec - the path fragment to append.

		Purpose:	Append a path fragment to the FileSpec.

					e.g.

						FileSpec F;
						F.SetFromString("c:\myfiles");

						FileSpec RF;
						RF.SetFromString("apps\word");
						F.AppendRelativeFileSpec(RF);

						String S;
						S = F.GetString();

						// S now equals "c:\myfiles\apps\word"

	**************************************************************************/

	void FileSpec::AppendRelativeFileSpec(const FileSpec& RelativeFileSpec)
	{
		// Make sure we have the folder separator
		AppendFolderSeparator();
		m_FileName += RelativeFileSpec.m_FileName;
	}

	/**************************************************************************

	>	bool FileSpec::IsAbsolute() const

		Author:		Tim Browse
		Date:		8/10/99

		Purpose:	Test to see if the FileSpec is relative or absolute.

	**************************************************************************/

	bool FileSpec::IsAbsolute() const
	{

		// Heuristic for detecting absolute paths under Windows...
		//
		// If it starts with "\\", it's a UNC.
		// If it starts with "<letter>:\", it's a full drive path
		//
		// Otherwise it's a relative path.

		if (m_FileName.Left(2) == "\\\\")
		{
			return true;
		}

		if (isalpha(m_FileName.GetAt(0)))
		{
			// Starts with a letter - is it followed by ":\" ?
			if (m_FileName.Mid(1,2) == ":\\")
			{
				return true;
			}
		}

		// Can't find evidence of it being an absolute pathname
		return false;
	}
	
	/**************************************************************************

	>	bool FileSpec::GetParent(FileSpec* pParent) const

		Author:		Tim Browse
		Date:		20/08/99
		Outputs:	pParent - the parent FileSpec of this object.
		Returns:	true => parent was extracted successfully.
					false => THis object has no parent.

		Purpose:	Extract the parent FileSpec object of this FileSpec.

					e.g.

						FileSpec F;
						F.SetFromString("c:\apps\lotus");

						FileSpec Parent;
						F.GetParent(&Parent);

						String S = Parent.GetString();

						// S now equals "c:\apps"

	**************************************************************************/

	bool FileSpec::GetParent(FileSpec* pParent) const
	{
		// Copy pathname
		String PathName = m_FileName;

		// Make sure there is no separator at the end
		int Length = PathName.GetLength();

		while ((Length > 0) && (PathName[Length - 1] == kStringFolderSeparator))
		{
			// Have we found the drive specifier?
			if ((Length == 3) && isalpha(PathName[0]) && (PathName[1] == ':'))
			{
				// Yes, so no parent exists for this
				return false;
			}
			// Otherwise, remove trailing separator
			Length--;
			PathName.Truncate(Length);
		}

		uint Pos;
		if (PathName.ReverseFind(kStringFolderSeparator, &Pos))
		{
			// Found a folder separator - terminate string here.
			PathName.Truncate(Pos + 1);
			pParent->SetFromString(PathName);

			// Success
			return true;
		}

		// Cannot obtain parent folder
		return false;
	}

	/**************************************************************************

	>	FileType FileSpec::GetFileType() const

		Author:		Tim Browse
		Date:		20/08/99
		Returns:	FileType_Folder - the FileSpec describes a folder.
					FileType_File - the FileSpec describes a normal file.
					FileType_NotFound - the FileSpec names a non-existent file.

		Purpose:	Find out if this FileSpec describes a file or a folder, or
					to see if the file even exists.

	**************************************************************************/

	FileType FileSpec::GetFileType() const
	{
		// Assume it doesn't exist to start with.
		FileType Type = FileType_NotFound;

		// Get the object's file attributes
		DWORD dwAttrs = ::GetFileAttributes(m_FileName.c_str());

		if (dwAttrs != 0xffffffff)
		{
			// We have found a matching file - work out what sort it is.
			if (dwAttrs & FILE_ATTRIBUTE_DIRECTORY)
			{
				// It's a folder
				Type = FileType_Folder;
			}
			else
			{
				// It's a file
				Type = FileType_File;
			}
		}

		// Tell caller what we found out about the file
		return Type;
	}


	/**************************************************************************

	>	uint32 FileSpec::GetFileSize(void) const

		Author:		Jason Williams
		Date:		25/10/99

		Returns:	The size, in bytes, of the referenced file.
					If the file is not found, or the FileSpec refers to a
					non-file object (e.g. directory or device) then the 
					returned size will be 0.

		Purpose:	Determine the size of the referenced file.

		Notes:		Only returns the lower 32-bits of the file size, so
					cannot handle files of more than 4GB.

	**************************************************************************/

	uint32 FileSpec::GetFileSize(void) const
	{
		uint32 Size = 0;

		WIN32_FIND_DATA FindData;

		HANDLE hFile = ::FindFirstFile(m_FileName.c_str(), &FindData);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			// We have found a matching file - work out what sort it is.
			if ((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
				Size = FindData.nFileSizeLow;

			// Shut down the find object.
			::FindClose(hFile);
			hFile = NULL;
		}

		return(Size);
	}

	/**************************************************************************

	>	Error GetFileTime(Time* pTime) const;

		Author:		Tim Browse
		Date:		10/9/00
		Outputs:	pTime - the last-modified time of the file.
		Returns:	File errors

		Purpose:	Obtain the last-modified time of the file named by this
					FileSpec object.

	**************************************************************************/

	Error FileSpec::GetFileTime(Time* pTime) const
	{
		struct _stat FileStatus;
		if (_stat(m_FileName.c_str(), &FileStatus) == 0)
		{
			// Extract the last-modified time
			*pTime = Time(FileStatus.st_mtime);
			return VNOERROR;
		}
		else
		{
			// Failed to stat file
			return Win32Error::TranslateError();
		}
	}

	/**************************************************************************
	
	>	bool FileSpec::IsReadOnly() const
	
		Author:		Tim Browse
		Date:		10/5/2002
		Returns:	true => file is read-only
					false => file is not read-only, or file not found
	
		Purpose:	Find out if the file is read-only.
	
	**************************************************************************/
	
	bool FileSpec::IsReadOnly() const
	{
		const char* szFilename = m_FileName.c_str();

		// Check for existence
		if ((_access(szFilename, 0)) == -1)
		{
			// File not found
			return false;
		}

		// Check for write permission
		return ((_access(szFilename, 2)) == -1);
	}

	/**************************************************************************

	>	Error FileSpec::CreateFolder(CreateFolderHandler *pHandler = NULL) const

		Author:		Tim Browse
		Date:		20/08/99
		Returns:	Error

		Purpose:	Create a folder with the name specified by this object.

	**************************************************************************/

	Error FileSpec::CreateFolder(CreateFolderHandler *pHandler) const
	{
			// Try to create the folder.
			if (::CreateDirectory(m_FileName.c_str(), NULL))
			{
				// Directory created succesfully - notify handler if necessary
				if (pHandler != NULL)
				{
					pHandler->NotifyFolderCreated(*this);
				}

				return VNOERROR;
			}
			else
			{
				// Directory creation failed - find out why, and report it.
				return Win32Error::TranslateError();
			}
		
		return VNOERROR;
	}

	/**************************************************************************

	>	Error FileSpec::Delete() const

		Author:		Tim Browse
		Date:		20/08/99
		Returns:	Error

		Purpose:	Delete the file specified by this object.

	**************************************************************************/

	Error FileSpec::Delete() const
	{
		// Attempt to delete the file or folder - first, we need to know what it is.
		switch (GetFileType())
		{
			case FileType_File:
			{
				if (::DeleteFile(m_FileName.c_str()))
				{
					// Folder deleted successfully
					return VNOERROR;
				}
				else
				{
					// File deletion failed - find out why, and report it.
					return Win32Error::TranslateError();
				}
			}

			case FileType_Folder:
			{
				if (::RemoveDirectory(m_FileName.c_str()))
				{
					// Folder deleted successfully
					return VNOERROR;
				}
				else
				{
					// File deletion failed - find out why, and report it.
					return Win32Error::TranslateError();
				}
			}
		
			case FileType_NotFound:
			default:
			{
				return ERROR(Err_File_NotFound, "File not found.");
			}

		}

	}

	/**************************************************************************
	
	>	Error FileSpec::Rename(const FileSpec& NewName) const
	
		Author:		Tim Browse
		Date:		2001/10/27
		Inputs:		NewName - the new name to give the file.
		Returns:	Error - if rename failed (e.g. file already exists, file
							destination is on a different disk/volume).
	
		Purpose:	Rename the specified file/folder object.  The file may 
					move within the directory structure, but if the file needs
					to migrate from one volume to another for the rename to 
					work, then it will fail.
	
	**************************************************************************/
	
	Error FileSpec::Rename(const FileSpec& NewName) const
	{
		// Try to rename the file - we specifically use MoveFileEx() and
		// do not pass the MOVEFILE_COPY_ALLOWED flag, so that the OS
		// will not try to copy the file from volume to volume, because
		// this Rename() function is only supposed to work within a volume.
		if (::MoveFileEx(GetString(), NewName.GetString(), 0) != 0)
		{
			// File was renamed successfully
			return VNOERROR;
		}
		else
		{
			// Error occured - report to caller
			return Win32Error::TranslateError();
		}
	}

	/**************************************************************************
	
	>	void FileSpec::Clear()
	
		Author:		Tim Browse
		Date:		2001/10/28
	
		Purpose:	Empty a FileSpec object, so it contains a null filename.
	
	**************************************************************************/
	
	void FileSpec::Clear()
	{
		m_FileName.Clear();
	}

	/**************************************************************************
	
	>	bool FileSpec::operator<(const FileSpec& rhs) const
	
		Author:		Tim Browse
		Date:		10/5/2002
		Inputs:		rhs - the FileSpec to compare against
		Returns:	true => this < rhs
					false => this >= rhs
	
		Purpose:	operator< for FileSpecs (so we can sort them)
	
	**************************************************************************/
	
	bool FileSpec::operator<(const FileSpec& rhs) const
	{
		return (m_FileName.CompareOrder(rhs.m_FileName) == -1);
	}

};	// namespace vos

