//
// FileMatches.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "..\FileMatches.h"

#include "..\FileSpec.h"

#include "..\FileInfo.h"

//
// Enable memory tracking
//
#ifdef _DEBUG
#define new vosDEBUG_NEW
#endif


namespace vos
{
	/**************************************************************************
	
	>	FileMatches::FileMatches()
	
		Author:		Tim Browse
		Date:		30/4/2002
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	FileMatches::FileMatches() : m_hFile(INVALID_HANDLE_VALUE)
	{
		// TCILB
	}

	/**************************************************************************
	
	>	FileMatches::FileMatches(const FileSpec& TheFileSpec)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Inputs:		TheFileSpec - the file spec to find matches for.
	
		Purpose:	Constructor
	
	**************************************************************************/
	
	FileMatches::FileMatches(const FileSpec& TheFileSpec) : m_hFile(INVALID_HANDLE_VALUE)
	{
		StartScan(TheFileSpec);
	}

	/**************************************************************************
	
	>	FileMatches::~FileMatches()
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Destructor
	
	**************************************************************************/
	
	FileMatches::~FileMatches()
	{
		CleanUp();
	}

	/**************************************************************************
	
	>	void FileMatches::StartScan(const FileSpec& fsFileSpec)
	
		Author:		Tim Browse
		Date:		30/4/2002
		Inputs:		fsFileSpec - spec of files to match
	
		Purpose:	Initialise (or re-initialise) the FileMatches object, ready
					for scanning.

					Alternatively, you can pass the FileSpec to the constructor
					instead.
	
	**************************************************************************/
	
	void FileMatches::StartScan(const FileSpec& fsFileSpec)
	{
		// Clean up any existing match
		CleanUp();

		// Remember the parent folder
		fsFileSpec.GetParent(&m_ParentFolder);

		// Start the find op
		m_hFile = ::FindFirstFile(fsFileSpec.GetString().c_str(), &m_FindData);
	}

	/**************************************************************************
	
	>	void FileMatches::CleanUp()
	
		Author:		Tim Browse
		Date:		30/4/2002
	
		Purpose:	Clean up any existing search objects.
	
	**************************************************************************/
	
	void FileMatches::CleanUp()
	{
		// Clean up the search if one is still open/pending.
		if (m_hFile != INVALID_HANDLE_VALUE)
		{
			::FindClose(m_hFile);
			m_hFile = INVALID_HANDLE_VALUE;
		}
	}

	/**************************************************************************
	
	>	void FileMatches::SkipSysDirEntries()
	
		Author:		Tim Browse
		Date:		18/4/2002
	
		Purpose:	Skip past any system directory entries, ie. "." and ".."
	
	**************************************************************************/
	
	void FileMatches::SkipSysDirEntries()
	{
		if (m_hFile == INVALID_HANDLE_VALUE)
		{
			// Ignore - no files left
			return;
		}

		for(;;)
		{
			if ((strcmp(m_FindData.cFileName, ".")  != 0) &&
				(strcmp(m_FindData.cFileName, "..") != 0))
			{
				// We're done
				return;
			}

			// It's "." or "..", so skip it
			if (!::FindNextFile(m_hFile, &m_FindData))
			{
				// No more matches - clean up
				::FindClose(m_hFile);
				m_hFile = INVALID_HANDLE_VALUE;
				return;
			}

		}
	}

	/**************************************************************************
	
	>	bool FileMatches::GetNextFile(FileSpec* pFileSpec, FileInfo* pInfo)
	
		Author:		Tim Browse
		Date:		18/4/2002
		Outputs:	pFileSpec - the name of the file
					pInfo - extra information about the file.
							(pInfo may be NULL if you don't want the info)
		Returns:	true => another matching file was available
					false => no more files
	
		Purpose:	Get the next matching file from the FileMatches object.
	
	**************************************************************************/
	
	bool FileMatches::GetNextFile(FileSpec* pFileSpec, FileInfo* pInfo)
	{
		// ignore "." and ".."
		SkipSysDirEntries();

		// If no files found, return false
		if (m_hFile == INVALID_HANDLE_VALUE)
		{
			return false;
		}

		// Return what we have at the moment
		*pFileSpec = m_ParentFolder;
		pFileSpec->AppendFilename(m_FindData.cFileName);

		// Fill in the file info, if the caller wants it
		if (pInfo != NULL)
		{
			// At the moment, we only do files up to 4Gb in size, even though
			// Win32 supports 64-bit file sizes.
			pInfo->m_Size = (uint) m_FindData.nFileSizeLow;

			// Is it a folder?
			pInfo->m_bIsFolder = (m_FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? true : false;

			// Copy file modification time
			pInfo->m_Modified.SetFromFileTime(m_FindData.ftLastWriteTime);
		}

		// Find next file (for next time)
		if (!::FindNextFile(m_hFile, &m_FindData))
		{
			// No more matches - clean up
			::FindClose(m_hFile);
			m_hFile = INVALID_HANDLE_VALUE;

			// But we still have the last match to return, so we return true.
			return true;
		}

		// There is a file match
		return true;
	}


};

