//
// Win32_Folders.cpp
//

//
// This software is protected by the GNU Lesser General Public License,
// also known as the 'GNU library License', or the LGPL.
//
// See the file COPYING.txt for details.
//

#include "StdHdr.h"

#include "Win32_Headers.h"
#include <Shlobj.h>
// For the special registry key names.
#include "regstr.h"
// Win32 support functions.
#include "Win32Registry.h"
#include "Win32Error.h"

#include "../Errors.h"
#include "../File.h"
#include "../FileSpec.h"

#include "../Folders.h"

// local static methods

static vos::Error GetSpecialFolder(vos::FileSpec* pTarget, uint32 FolderType);

/**************************************************************************

>	vos::Error GetSpecialFolder(vos::FileSpec* pTarget, uint32 FolderType)

	Author:		Tim Browse
	Date:		18/4/2002
	Inputs:		FolderType - the folder to access, e.g. CSIDL_PERSONAL, etc.
	Outputs:	pTarget - the name of the folder
	Returns:	Error (if special folder name not found/recognised)

	Purpose:	Helper function for finding 'special' folders in the Win32
				Shell namespace.

**************************************************************************/

vos::Error GetSpecialFolder(vos::FileSpec* pTarget, uint32 FolderType)
{
	vos::Error Result = VNOERROR;

	// Get a pointer to an IDL for the path
	// We are responsible for the pointer
	LPITEMIDLIST pIDL;
	HRESULT hrGotIDL = ::SHGetSpecialFolderLocation(NULL, FolderType, &pIDL);

	if (SUCCEEDED(hrGotIDL))
	{
		// Convert the item ID list's binary
		// representation into a file system path
		char Path[_MAX_PATH];
		BOOL bGotPath = ::SHGetPathFromIDList(pIDL, Path);

		// Get our allocator
		LPMALLOC pAllocator;
		HRESULT hrGotAllocator = ::SHGetMalloc(&pAllocator);

		if (SUCCEEDED(hrGotAllocator))
		{
			// Free the IDL we're responsible for
			pAllocator->Free(pIDL);

			// Free our task allocator
			pAllocator->Release();
		}

		if (bGotPath)
		{
			// Assign the path
			*pTarget = vos::FileSpec(Path);
		}
		else
		{
			Result = ERROR(vos::Err_File_NotFound, "User Home directory not found.");
		}
	}
	else
	{
		Result = ERROR(vos::Err_File_NotFound, "User Home directory not found.");
	}
	
	return Result;
}

// class methods

namespace vos
{
	//
	// The override application folder spec
	//
	FileSpec Folders::m_AppFolderOverride;

	/**************************************************************************

	>	Error Folders::UserHome(FileSpec* pTarget)

		Author:		Rob Myers
		Date:		8/10/01
		Inputs:		pTarget - The Filespec to fill out.
		Returns:	Standard error

		Purpose:	Find the current user's home directory,
					or a global documents/permanent store directory if this
					isn't possible on this system.

	**************************************************************************/

	Error Folders::UserHome(FileSpec* pTarget)
	{
		Error Result = GetSpecialFolder(pTarget, CSIDL_PERSONAL);
		return Result;
	}
	
	
	/**************************************************************************

	>	Error Folders::UserDocuments(FileSpec* pTarget)

		Author:		Rob Myers
		Date:		8/10/01
		Inputs:		pTarget - The Filespec to fill out.
		Returns:	Standard error

		Purpose:	Find the current user's document directory,
					or their home directory if this isn't possible.

	**************************************************************************/

	Error Folders::UserDocuments(FileSpec* pTarget)
	{
		Error Result = GetSpecialFolder(pTarget, CSIDL_PERSONAL);
		return Result;
	}

	/**************************************************************************

	>	Error Folders::UserPreferences(FileSpec* pTarget)

		Author:		Rob Myers
		Date:		8/10/01
		Inputs:		pTarget - The Filespec to fill out.
		Returns:	Standard error

		Purpose:	Find the current user's preferences directory,
					or their home directory if this isn't possible.

	**************************************************************************/

	Error Folders::UserPreferences(FileSpec* pTarget)
	{
		Error Result = GetSpecialFolder(pTarget, CSIDL_PERSONAL);
		return Result;
	}
	
	/**************************************************************************

	>	Error Folders::SystemTemp(FileSpec* pTarget)

		Author:		Rob Myers
		Date:		8/10/01
		Inputs:		pTarget - The Filespec to fill out.
		Returns:	Standard error

		Purpose:	Find a temporary/scratch directory for the application.
					This should not be user-specific, and files in it
					*MUST* be deleted by the OS when no longer used.

	**************************************************************************/

	Error Folders::SystemTemp(FileSpec* pTarget)
	{
		char szTempPath[_MAX_PATH + 1];

		DWORD dwLength = ::GetTempPath(_MAX_PATH, szTempPath);
		
		// Check for out of bounds error
		if (dwLength > _MAX_PATH)
		{
			// Illegal path length
			return ERROR(Err_BadParam, "Illegal path length");
		}

		// Check for no path
		if (dwLength == 0)
		{
			
			return ERROR(Err_BadParam, "No path (zero length)");
		}

		// Got a temporary path - put it into the filespec
		pTarget->SetFromString(szTempPath);

		// All done
		return VNOERROR;
	}
	
	/**************************************************************************

	>	Error Folders::SystemApplications(FileSpec* pTarget)

		Author:		Rob Myers
		Date:		8/10/01
		Inputs:		pTarget - The Filespec to fill out.
		Returns:	Standard error

		Purpose:	Find the global new application install directory,
					or a writable disk if this cannot be found.

	**************************************************************************/

	Error Folders::SystemApplications(FileSpec* pFileSpec)
	{
		//Error Result = GetSpecialFolder(pTarget, CSIDL_PROGRAMS);
		//return Result;

		// We need to find the standard "Program Files" directory, which is lurking in 
		// the registry in a standard place.
		DWORD dwDisposition;
		HKEY hKey = NULL;

		if (::RegCreateKeyEx(HKEY_LOCAL_MACHINE,
							 REGSTR_PATH_SETUP,
							 0,
							 "",
							 REG_OPTION_NON_VOLATILE,
							 KEY_READ,
							 NULL,
							 &hKey,
							 &dwDisposition) != ERROR_SUCCESS)
		{
			// Failed to read the key
			return ERROR(Err_NotFound, "Couldn't get Program Files registry key");
		}

		// Read the ProgramFilesDir key
		String ProgramFilesDir;
		if (!Win32Registry::ReadStringValue(hKey, "ProgramFilesDir", &ProgramFilesDir))
		{
			// Unable to read folder name
			return ERROR(Err_NotFound, "Couldn't get Program Files registry entry");
		}

		// Make filespec from the folder name we just read from the registry.
		pFileSpec->SetFromString(ProgramFilesDir);
		
		return VNOERROR;
	}

	/**************************************************************************

	>	Error FileSpec::ApplicationHome(FileSpec* pFileSpec)

		Author:		Tim Browse
		Date:		20/08/99
		Outputs:	The FileSpec identifying the folder this application
					is being run from.
		Returns:	true => folder found ok
					false => the foldere this app is running from could not
							 be determined.

		Purpose:	Find out which folder this program is being run from.

	**************************************************************************/

	Error Folders::ApplicationHome(FileSpec* pFileSpec)
	{
		// If the user has overridden the application's working directory,
		// use the override instead
		if(! m_AppFolderOverride.IsEmpty())
		{
			*pFileSpec = m_AppFolderOverride;
			return VNOERROR;
		}

		char szPathName[_MAX_PATH+1];

		int cChars = ::GetModuleFileName(NULL, szPathName, _MAX_PATH);
		if (cChars == 0)
		{
			// Failed to get filename
			return ERROR(Err_NotFound, "Couldn't get application home folder");
		}

		// Check API call gave us the right length
		ASSERT(cChars == strlen(szPathName));

		// Strip module name off the end.
		int iIndex = cChars-1;

		while ((iIndex > 0) && (szPathName[iIndex] != FileSpec::kStringFolderSeparator))
		{
			// Move to previous character
			iIndex--;
		}

		// Chop leafname off the end
		// (If no pathname is present, then we end up with an empty
		//  pathname, which is the conceptually correct result).
		szPathName[iIndex] = 0;

		// Assign to filespec
		pFileSpec->SetFromString(szPathName);

		return VNOERROR;
	}
};